/**************************************************************

bias.c (C-Munipack project)
Bias correction
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "comfun.h"
#include "ccdfile.h"
#include "console.h"
#include "cmpack_common.h"
#include "cmpack_bias.h"

/**********************   PRIVATE DATA TYPES   ***************************/

/* Bias correction context */
struct _CmpackBiasCorr
{
	CmpackContext gc;					/**< Global context */
	CmpackConsole *con;					/**< Output console */
	CmpackBorder border;				/**< Border size */
	int nx;             				/**< Width of bias-frame */
	int ny;             				/**< Height of bias-frame */
	double *bdata;         				/**< Data of bias-frame */
};

/***********************   LOCAL FUNCTIONS   **************************/

/* Frees allocated memory */
static void bias_clean(void *ctx)
{
	CmpackBiasCorr *lc = (CmpackBiasCorr*)ctx;

	cmpack_free(lc->bdata);
	cmpack_unref(lc->con);
	gc_clean(ctx);
}

static int bias_bias(CmpackBiasCorr *lc, CmpackCcdFile *infile, CmpackCcdFile *outfile)
{
	int res, i, nx, ny, left, right, top, bottom, x, y, underflow, overflow;
	double *sdata, minvalue, maxvalue;
	char msg[MAXLINE];
	CmpackBitpix bitpix;

	/* Check image size */
	nx = ny = 0;
	cmpack_ccd_get_size(infile, &nx, &ny);
	if (nx<=0 || ny<=0 || nx>=65536 || ny>=65536)
		return CMPACK_ERR_INVALID_SIZE;
	if (lc->nx!=nx || lc->ny!=ny) 
		return CMPACK_ERR_DIFF_SIZE_BIAS;

	/* Get data format */
	bitpix = cmpack_ccd_get_bitpix(infile);
	if (bitpix==CMPACK_BITPIX_UNKNOWN) {
		printout(lc->con, 0, "Unsupported image data format.");
		return CMPACK_ERR_READ_ERROR;
	}
	cmpack_ccd_get_range(infile, &minvalue, &maxvalue);
	if (is_debug(lc->con)) {
		sprintf(msg, "Data format  : %s", pixformat(bitpix));
		printout(lc->con, 1, msg);
		sprintf(msg, "Min. value   : %.12g", minvalue); 
		printout(lc->con, 1, msg);
		sprintf(msg, "Max. value   : %.12g", maxvalue); 
		printout(lc->con, 1, msg);
		sprintf(msg, "Border     : %d %d %d %d", lc->border.left, lc->border.top,
			lc->border.right, lc->border.bottom);
		printout(lc->con, 1, msg);
	}

	/* Read exposure data */
	sdata = cmpack_malloc(nx*ny*sizeof(double));
	res = cmpack_ccd_get_image(infile, CMPACK_BITPIX_DOUBLE, 0, 0, sdata);
	if (res!=0) {
		cmpack_free(sdata);
		return res;
	}

	/* bias subtraction and range checking */
	underflow = overflow = 0;
	left = lc->border.left;
	top = lc->border.top;
	right = nx - lc->border.right;
	bottom = ny - lc->border.bottom;
	for (y=0;y<ny;y++) {
		for (x=0;x<nx;x++) {
			i = x + y*nx;
			if (x>=left && x<right && y>=top && y<bottom) {
				sdata[i] -= lc->bdata[i];
				if (sdata[i]<minvalue) {
					sdata[i] = minvalue;
					underflow = 1;
				}
				if (sdata[i]>maxvalue) {
					sdata[i] = maxvalue;
					overflow = 1;
				}
			} else {
				sdata[i] = 0.0;
			}
		}
	}

	/* Store image data */
	res = ccd_write_image(outfile, CMPACK_BITPIX_DOUBLE, sdata);
	cmpack_free(sdata);

	/* Update frame header */
	ccd_write_history(outfile, "Bias frame subtracted.");

	/* Normal return */
	if (overflow)
		printout(lc->con, 1, "An overflow has been occurred during computation.");
	if (underflow)
		printout(lc->con, 1, "An underflow has been occurred during computation.");
	return res;
}

/**********************   PUBLIC FUNCTIONS   **************************/

/* Initializes the context */
CmpackBiasCorr *cmpack_bias_init(void)
{
	CmpackBiasCorr *lc = gc_create(sizeof(CmpackBiasCorr));
	lc->gc.clean = &bias_clean;
	return lc;
}

/* Attach console */
void cmpack_bias_set_console(CmpackBiasCorr *ctx, CmpackConsole *con)
{
	if (con!=ctx->con) {
		cmpack_unref(ctx->con);
		ctx->con = cmpack_addref(con);
	}
}

/* Set image border */
void cmpack_bias_set_border(CmpackBiasCorr *lc, const CmpackBorder *border)
{
	if (border)
		lc->border = *border;
	else
		memset(&lc->border, 0, sizeof(CmpackBorder));
}

/* Set image border */
void cmpack_bias_get_border(CmpackBiasCorr *lc, CmpackBorder *border)
{
	if (border)
		*border = lc->border;
}

/* Read bias frame */
int cmpack_bias_rbias(CmpackBiasCorr *lc, CmpackCcdFile *bias)
{
	int res, nx, ny;
	double *data;

	/* Clear previous bias frame */
	cmpack_free(lc->bdata);
	lc->bdata = NULL;
	lc->nx = lc->ny = 0;

    /* Check file name */
    if (!bias) {
        printout(lc->con, 0, "Invalid bias frame context.");
 	    return CMPACK_ERR_INVALID_PAR;
    }

    /* Read and check dimensions */
	nx = ny = 0;
	cmpack_ccd_get_size(bias, &nx, &ny);
	if (nx<=0 || nx>=65536 || ny<=0 || ny>=65536) {
		printout(lc->con, 1, "Invalid dimensions of the bias frame.");
		return CMPACK_ERR_INVALID_SIZE;
	}
	lc->nx = nx;
	lc->ny = ny;

	/* Read bias frame data */
	data = cmpack_malloc(lc->nx*lc->ny*sizeof(double));
	res = cmpack_ccd_get_image(bias, CMPACK_BITPIX_DOUBLE, 0, 0, data);
	if (res!=0) {
		cmpack_free(data);
		return res;
	}
	lc->bdata = data;

	/* Print parameters */
	if (is_debug(lc->con)) {
		printout(lc->con, 1, "Bias correction frame:");
		printpari(lc->con, "Width", 1, lc->nx);
		printpari(lc->con, "Height", 1, lc->ny);
	}
    return 0;
}

/* Computes the output image from source image and bias image (out=sci-bias) */
int cmpack_bias(CmpackBiasCorr *lc, CmpackCcdFile *file)
{
	/* Check bias file */
	if (!lc->bdata) {
		printout(lc->con, 0, "Missing bias frame.");
		return CMPACK_ERR_NO_BIAS_FRAME;
	}

  	/* Check parameters */
	if (!file) {
		printout(lc->con, 0, "Invalid frame context.");
		return CMPACK_ERR_INVALID_PAR;
	}

	/* Bias correction */
	return bias_bias(lc, file, file);
}

/* Computes the output image from source image and bias image (out=sci-bias) */
int cmpack_bias_ex(CmpackBiasCorr *lc, CmpackCcdFile *infile, CmpackCcdFile *outfile)
{
	int res, nx, ny;
	CmpackBitpix bitpix;

	/* Check bias file */
	if (!lc->bdata) {
		printout(lc->con, 0, "Missing bias frame.");
		return CMPACK_ERR_NO_BIAS_FRAME;
	}

  	/* Check parameters */
	if (!infile) {
		printout(lc->con, 0, "Invalid input frame context.");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (!outfile) {
		printout(lc->con, 0, "Invalid output frame context.");
		return CMPACK_ERR_INVALID_PAR;
	}

	/* Check image size */
	nx = ny = 0;
	cmpack_ccd_get_size(infile, &nx, &ny);
	if (nx<=0 || nx>=65536 || ny<=0 || ny>=65536) {
		printout(lc->con, 0, "Invalid frame size.");
		return CMPACK_ERR_INVALID_SIZE;
	}

	/* Get data format */
	bitpix = cmpack_ccd_get_bitpix(infile);
	if (bitpix==CMPACK_BITPIX_UNKNOWN) {
		printout(lc->con, 0, "Unsupported image data format.");
		return CMPACK_ERR_READ_ERROR;
	}

	/* Init output image */
	res = ccd_prepare(outfile, nx, ny, bitpix);
	if (res!=0) 
		return res;

	/* Copy file header */
	res = ccd_copy_header(outfile, infile);
	if (res!=0) 
		return res;

	/* Bias correction */
	return bias_bias(lc, infile, outfile);
}

