/**************************************************************

cat_head.c (C-Munipack project)
Catalogue file header
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hash.h"
#include "comfun.h"
#include "console.h"
#include "cmpack_common.h"
#include "cathead.h"

/*************************   Constants  ***********************/

#define ALLOC_BY 	64

/***********************      Datatypes      ******************/

/* Single header item */
struct _CmpackCatHeadItem
{
	char *key;						/**< Keyword (case insensitive) */
	char *val;						/**< Value */
	char *com;						/**< Comment text */
};
	
/* File header */
struct _CmpackCatHeader
{
	int	count;						/**< Number of items */
	int	capacity;					/**< Number of allocated records */
	CmpackCatHeadItem *list;		/**< List of header items */
	CmpackHashTable hash;				/**< Hash table of items */
};

/********************   Local functions    ***********************/

/* Init head item */
static void headitem_init(CmpackCatHeadItem *item, const char *key, const char *val, const char *com)
{
	item->key = cmpack_strdup(key);
	item->val = cmpack_strdup(val);
	item->com = cmpack_strdup(com);
}

/* Makes copy of header item */
static void headitem_copy(CmpackCatHeadItem *dst, const CmpackCatHeadItem *src)
{
	dst->key = cmpack_strdup(src->key);
	dst->val = cmpack_strdup(src->val);
	dst->com = cmpack_strdup(src->com);
}

/* Change value of header item */
static void headitem_setval(CmpackCatHeadItem *item, const char *val, const char *com)
{
	if (val && strlen(val)>0) {
		item->val = (char*) cmpack_realloc(item->val, (strlen(val)+1)*sizeof(char));
		strcpy(item->val, val);
	} else {
		cmpack_free(item->val);
		item->val = NULL;
	}
	if (com && strlen(com)>0) {
		item->com = (char*) cmpack_realloc(item->com, (strlen(com)+1)*sizeof(char));
		strcpy(item->com, com);
	} else {
		cmpack_free(item->com);
		item->com = NULL;
	}
}

/* Destroy head item */
static void headitem_clear(CmpackCatHeadItem *item)
{
	cmpack_free(item->key);
	cmpack_free(item->val);
	cmpack_free(item->com);
	memset(item, 0, sizeof(CmpackCatHeadItem));
}

/* Dump head item to stdout */
static void headitem_dump(CmpackConsole *con, CmpackCatHeadItem *item)
{
	char aux[MAXLINE];

	if (item->key && strlen(item->key)>0 && item->val && strlen(item->val)>0) {
		strcat(aux, " = ");
		strcat(aux, item->val);
		if (item->com && strlen(item->com)>0) {
			strcat(aux, " ; ");
			strcat(aux, item->com);
		}
	}
	printout(con, 1, aux);
}

/* Write item to stdout */
static void headitem_write(CmpackCatHeadItem *item, FILE *to)
{
	if (item->key && strlen(item->key)>0 && item->val && strlen(item->val)>0) {
		if (item->com && strlen(item->com)>0) 
			fprintf(to, "<%s>%s <!-- %s --></%s>\n", item->key, item->val, item->com, item->key);
		else
			fprintf(to, "<%s>%s</%s>\n", item->key, item->val, item->key);
	}
}

/********************   Public functions    ***********************/

/* Create a new header */
CmpackCatHeader *cat_head_create(void)
{
	CmpackCatHeader *head = (CmpackCatHeader*)cmpack_malloc(sizeof(CmpackCatHeader));
	head->list = NULL;
	head->count = head->capacity = 0;
	hash_init(&head->hash);
	return head;
}

/* Free the header */
void cat_head_free(CmpackCatHeader *head)
{
	int j;
	
	for (j=0; j<head->count; j++)
		headitem_clear(&head->list[j]);
	cmpack_free(head->list);
	head->list = NULL;
	head->count = head->capacity = 0;
	hash_clear(&head->hash);
	cmpack_free(head);
}

/* Dump header content */
void cat_head_dump(CmpackCatHeader *head, CmpackConsole *con)
{
	int j;
	char aux[MAXLINE];
	
	sprintf(aux, "\tHeader: %d items\n", head->count);
	printout(con, 1, aux);
	for (j=0; j<head->count; j++)
		if (head->list[j].key)
			headitem_dump(con, &head->list[j]);
}	

/* Save header to file */
void cat_head_write(CmpackCatHeader *head, FILE *to)
{
	int j;
	
	fprintf(to, "<info>\n");
	for (j=0; j<head->count; j++)
		if (head->list[j].key)
			headitem_write(&head->list[j], to);
	fprintf(to, "</info>\n");
}	

/* Make copy of a header */
void cat_head_copy(CmpackCatHeader *dsthead, const CmpackCatHeader *srchead)
{
	int j;
	
	cat_head_clear(dsthead);
	if (srchead->count>0) {
		dsthead->capacity = dsthead->count = srchead->count;
		dsthead->list = (CmpackCatHeadItem*) cmpack_realloc(dsthead->list, dsthead->capacity*sizeof(CmpackCatHeadItem));
		for (j=0; j<srchead->count; j++) {
			headitem_copy(&dsthead->list[j], &srchead->list[j]);
			hash_insert(&dsthead->hash, dsthead->list[j].key, &dsthead->list[j]);
		}
	}
}	

/* Clear header */
void cat_head_clear(CmpackCatHeader *head)
{
	int j;
	
	for (j=0; j<head->count; j++)
		headitem_clear(&head->list[j]);
	cmpack_free(head->list);
	head->list = NULL;
	head->count = head->capacity = 0;
	hash_clear(&head->hash);
}	

/* Insert or update header item */
void cat_head_set(CmpackCatHeader *head, const char *key, const char *val, const char *com)
{
	CmpackCatHeadItem *item = (CmpackCatHeadItem*) hash_search(&head->hash, key);
	if (item) {
		/* Modify existing item */
		if (val) headitem_setval(item, val, com);
	} else {
		/* Add new item */
		if (head->count>=head->capacity) {
			head->capacity += ALLOC_BY;
			head->list = (CmpackCatHeadItem*) cmpack_realloc(head->list, head->capacity*sizeof(CmpackCatHeadItem));
		}
		headitem_init(&head->list[head->count], key, val, com);
		hash_insert(&head->hash, key, &head->list[head->count]);
		head->count++;
	}
}

/* Find header item by its name */
CmpackCatHeadItem *cat_head_find(CmpackCatHeader *head, const char *key)
{
	return (CmpackCatHeadItem*)hash_search(&head->hash, key);
}

/* Find header item by its name */
CmpackCatHeadItem *cat_head_get(CmpackCatHeader *head, int index)
{
	if (index>=0 && index<head->count)
		return &head->list[index];
	else
		return NULL;
}

/* Delete header item */
void cat_head_delete(CmpackCatHeader *head, CmpackCatHeadItem *item)
{
	/* Delete item */
	hash_delete(&head->hash, item->key, item);
	headitem_clear(item);
	head->count--;
	if (head->count==0) {
		/* If there is no item, clear the list */
		cmpack_free(head->list);
		head->list = NULL;
		head->count = head->capacity = 0;
	} else {
		/* Pack the list (move all following items to the left) */
		size_t len = head->count - (item - head->list);
		if (len>0) {
			CmpackCatHeadItem *aux = (CmpackCatHeadItem*) cmpack_calloc(len, sizeof(CmpackCatHeadItem));
			memcpy(aux, item+1, len*sizeof(CmpackCatHeadItem));
			memcpy(item, aux, len*sizeof(CmpackCatHeadItem));
			cmpack_free(aux);
		}
	}
}

/* Get number of header items */
int cat_head_count(CmpackCatHeader *head)
{
	return head->count;
}

/* Find header item by its name */
const char *cat_head_item_key(CmpackCatHeadItem *item)
{
	return item->key;
}

/* Find header item by its name */
const char *cat_head_item_val(CmpackCatHeadItem *item)
{
	return item->val;
}

/* Find header item by its name */
const char *cat_head_item_com(CmpackCatHeadItem *item)
{
	return item->com;
}
