/**************************************************************

konve.c (C-Munipack project)
Conversion tool for CCD frames
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "comfun.h"
#include "console.h"
#include "ccdfile.h"
#include "cmpack_common.h"
#include "cmpack_ccdfile.h"
#include "cmpack_konve.h"

/*******************   DATA TYPES   ***************************/

/* Conversion context */
struct _CmpackKonv
{
	CmpackContext gc;			/**< Global context	*/
	CmpackConsole *con;			/**< Console */
	CmpackBitpix bitpix;		/**< Output data format */
	int flags;					/**< Image flipping and rotation flags */
	CmpackBorder border;		/**< Border size */
};

/*********************   Local data and functions   *************************/

/* Free allocated memory */
static void konv_clean(void *ctx)
{
	CmpackKonv *lc = (CmpackKonv*)ctx;

	cmpack_unref(lc->con);
	gc_clean(ctx);
}

/***********************   Public functions   ****************************/

/* Create and init the context */
CmpackKonv *cmpack_konv_init(void)
{
	CmpackKonv *lc = gc_create(sizeof(CmpackKonv));
	lc->gc.clean = &konv_clean;
	lc->bitpix = CMPACK_BITPIX_AUTO;
	return lc;
}

/* Attach console */
void cmpack_konv_set_console(CmpackKonv *ctx, CmpackConsole *con)
{
	if (con!=ctx->con) {
		cmpack_unref(ctx->con);
		ctx->con = cmpack_addref(con);
	}
}

/* Set image flip flags */
void cmpack_konv_set_trafo(CmpackKonv *lc, int flags)
{
	lc->flags = flags;
}

/* Get image flip flags */
int cmpack_konv_get_trafo(CmpackKonv *lc)
{
	return lc->flags;
}

/* Set image flip flags */
void cmpack_konv_set_bitpix(CmpackKonv *lc, CmpackBitpix bitpix)
{
	lc->bitpix = bitpix;
}

/* Get image flip flags */
CmpackBitpix cmpack_konv_get_bitpix(CmpackKonv *lc)
{
	return lc->bitpix;
}

/* Set image border */
void cmpack_konv_set_border(CmpackKonv *lc, const CmpackBorder *border)
{
	if (border)
		lc->border = *border;
	else
		memset(&lc->border, 0, sizeof(CmpackBorder));
}

/* Set image border */
void cmpack_konv_get_border(CmpackKonv *lc, CmpackBorder *border)
{
	if (border)
		*border = lc->border;
}

/* Converts 'in' to 'out' */
int cmpack_konv(CmpackKonv *lc, CmpackCcdFile *infile, CmpackCcdFile *outfile)
{
	int flipx, flipy;
	char msg[MAXLINE];

  	/* Check parameters */
	if (!infile) {
		printout(lc->con, 0, "Invalid input frame context.");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (!outfile) {
		printout(lc->con, 0, "Invalid output frame context.");
		return CMPACK_ERR_INVALID_PAR;
	}

	/* Print configuration parameters */
	flipx = (lc->flags & CMPACK_FLIP_X);
	flipy = (lc->flags & CMPACK_FLIP_Y);
	if (is_debug(lc->con)) {
		sprintf(msg, "Format     : %s", pixformat(lc->bitpix));
		printout(lc->con, 1, msg);
		sprintf(msg, "Flip       : %s%s", (flipx ? "x" : ""), (flipy ? "y" : ""));
		printout(lc->con, 1, msg);
		sprintf(msg, "Border     : %d %d %d %d", lc->border.left, lc->border.top,
			lc->border.right, lc->border.bottom);
		printout(lc->con, 1, msg);
	}

	/* Convert data */
	return ccd_copy(infile, outfile, lc->bitpix, flipx, flipy, 
		&lc->border, lc->con);
}
