/**************************************************************

konvsbig.c (C-Munipack project)
Conversion functions for SBIG files
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "config.h"
#include "trajd.h"
#include "console.h"
#include "comfun.h"
#include "cmpack_common.h"
#include "iost.h"
#include "konvsbig.h"

/* Pixel value bias */
#define BIAS       	32768
/* Header length of the SBIG image file in bytes */
#define HEADSIZE  	2048             
/* Length of string buffers */
#define MAXLINE		1024

/* Uppercases the string */
static void strup(char *s)
{
  for( ; *s != '\0'; s++ )
    if( *s >= 'a' && *s <= 'z' )
      *s = *s - ('a' - 'A');
}

/* Returns 1 if the file is in SBIG format */
int sbig_test(const char *block, size_t length)
{
	return (length>HEADSIZE) && 
		(strstr(block, "ST-")==block || strstr(block, "SBIG")==block || strstr(block, "PixCel")==block);
}

/* Open the file and read the header into memory */
int sbig_open(tHandle *fs, const char *filename, CmpackOpenMode flags)
{
	return stopen((struct stfile**)fs, filename);
}

/* Close a file and free allocated memory */
void sbig_close(tHandle fs)
{
	stclos((struct stfile*)fs);
}

/* Reads string value from header */
int sbig_gets(tHandle fs, const char *key, char **buf)
{
	return stgkys((struct stfile*)fs, key, buf);
}

/* Get file format designation */
char *sbig_getmagic(tHandle fs)
{
	char *value = NULL;
	stgkyn((struct stfile*)fs, 0, &value, NULL);
	return value;
}

/* Get optical filter name */
char *sbig_getfilter(tHandle fs)
{
	char *value = NULL;
	stgkys((struct stfile*)fs, "Filter", &value);
	return value;
}

/* Get observer's name */
char *sbig_getobserver(tHandle fs)
{
	char *value = NULL;
	stgkys((struct stfile*)fs, "Observer", &value);
	return value;
}

/* Reads integer value from header */
int sbig_geti(tHandle fs, const char *key, int *value)
{
	return stgkyi((struct stfile*)fs, key, value);
}

int sbig_getsize(tHandle fs, int *width, int *height)
{
	stgkyi((struct stfile*)fs, "Width", width);
	stgkyi((struct stfile*)fs, "Height", height);
	return 0;
}

/* Reads floating point value from header */
int sbig_getd(tHandle fs, const char *key, double *value)
{
	return stgkyd((struct stfile*)fs, key, value);
}

/* Get date and time of observation */
int sbig_getdatetime(tHandle fs, CmpackDateTime *dt)
{
	memset(dt, 0, sizeof(CmpackDateTime));
	stgetdate((struct stfile*)fs, &dt->date.year, &dt->date.month, &dt->date.day);
	stgettime((struct stfile*)fs, &dt->time.hour, &dt->time.minute, &dt->time.second);
	return 0;
}

/* Get exposure duration */
int sbig_getexptime(tHandle fs, double *val)
{
	int res;
	res = stgkyd((struct stfile*)fs, "Exposure", val);
	if (res==0)
		*val = 0.01 * (*val);
	return res;
}

/* Get CCD temperature */
int sbig_getccdtemp(tHandle fs, double *val)
{
	return stgkyd((struct stfile*)fs, "Temperature", val);
}

/* Get key name and value by index. Comm is set always to "" */
int sbig_gkyn(tHandle fs, int nkey, char **key, char **val, char **com)
{
	if (com)
		*com = NULL;
	return stgkyn((struct stfile*)fs, nkey, key, val);
}

/* Get image data type */
CmpackBitpix sbig_getbitpix(tHandle fs)
{
	return CMPACK_BITPIX_DOUBLE;
}

/* Get image data type */
int sbig_getrange(tHandle fs, double *minvalue, double *maxvalue)
{
	if (minvalue)
		*minvalue = 0.0;
	if (maxvalue)
		*maxvalue = 65535.0;
	return 0;
}

/* Read image data */
int sbig_getimage(tHandle fs, CmpackBitpix bitpix, int flipx, int flipy, void *buf)
{
	int res;

	switch (bitpix)
	{
	case CMPACK_BITPIX_USHORT:
		res = stg2duj((struct stfile*)fs, flipx, flipy, buf);
		return (res==0 ? 0 : CMPACK_ERR_READ_ERROR);
	case CMPACK_BITPIX_SLONG:
		res = stg2dl((struct stfile*)fs, flipx, flipy, buf);
		return (res==0 ? 0 : CMPACK_ERR_READ_ERROR);
	case CMPACK_BITPIX_ULONG:
		res = stg2dul((struct stfile*)fs, flipx, flipy, buf);
		return (res==0 ? 0 : CMPACK_ERR_READ_ERROR);
	case CMPACK_BITPIX_FLOAT:
		res = stg2de((struct stfile*)fs, flipx, flipy, buf);
		return (res==0 ? 0 : CMPACK_ERR_READ_ERROR);
	case CMPACK_BITPIX_DOUBLE:
		res = stg2dd((struct stfile*)fs, flipx, flipy, buf);
		return (res==0 ? 0 : CMPACK_ERR_READ_ERROR);
	default:
		return CMPACK_ERR_READ_ERROR;
	}
}

static int copyheader(struct stfile *src, fitsfile *dst)
{
	double	f;
	char	msg[MAXLINE], *filter, *history, *observer;
	int		i, sstat, dstat=0;
	long	li;
	char	datestr[64], timestr[64];
	CmpackDateTime dt;

	/* Set date and time of observation */
	memset(&dt, 0, sizeof(CmpackDateTime));
	stgetdate(src, &dt.date.year, &dt.date.month, &dt.date.day);
	stgettime(src, &dt.time.hour, &dt.time.minute, &dt.time.second);
	sprintf(datestr, "%04d-%02d-%02d", dt.date.year, dt.date.month, dt.date.day);
	ffukys(dst,"DATE-OBS", (char*) datestr, "UT DATE OF START", &dstat);
	sprintf(timestr, "%02d:%02d:%02d.%03d", dt.time.hour, dt.time.minute, dt.time.second, dt.time.milisecond);
	ffukys(dst,"TIME-OBS", (char*) timestr, "UT TIME OF START", &dstat);

	/* Other parameters */
	sstat = stgkyi(src,"Exposure",&i);
	if (sstat==0) ffpkyg(dst,"EXPTIME",1.0*i/100,2,"EXPOSURE IN SECONDS",&dstat);
	sstat = stgkyi(src,"Number_exposures",&i);
	if (sstat==0) ffpkyj(dst,"SNAPSHOT",i,"NUMBER OF SNAPSHOT",&dstat);
	sstat = stgkyi(src,"Each_exposure",&i);
	if (sstat==0) ffpkyg(dst,"EACHSNAP",1.0*i/100,2,"SNAPSHOT IN SECONDS",&dstat);
	sstat = stgkyd(src,"X_pixel_size",&f);
	if (sstat==0) ffpkyd(dst,"PIXWIDTH",f,3,"PIXEL WIDTH IN MM",&dstat);
	sstat = stgkyd(src,"Y_pixel_size",&f);
	if (sstat==0) ffpkyd(dst,"PIXHEIGH",f,3,"PIXEL HEIGHT IN MM",&dstat);
	sstat = stgkyi(src,"Readout_mode",&i);
	if (sstat==0) ffpkyj(dst,"RESMODE",i,"RESOLUTION MODE",&dstat);
	sstat = stgkyi(src,"Exposure_state",&i);
	if (sstat==0) {
		li = i;
		sprintf(msg,"%lx",li);
		strup(msg);
		ffpkys(dst,"EXPSTATE",msg,"EXPOSURE STATE (HEX)",&dstat);
	}
	sstat = stgkyd(src,"Temperature",&f);
	if (sstat==0) ffpkyg(dst,"TEMPERAT",f,2,"TEMPERATURE IN DEGREES C",&dstat);
	sstat = stgkyd(src,"Response_factor",&f);
	if (sstat==0) ffpkyg(dst,"RESPONSE",f,3,"CCD RESPONSE FACTOR",&dstat);
	sstat = stgkyd(src,"E_Gain",&f);
	if (sstat==0) ffpkyg(dst,"EPERADU",f,2,"ELECTRONS PER ADU",&dstat);
	sstat = stgkyd(src,"Focal_length",&f);
	if (sstat==0) ffpkyd(dst,"FOCALLEN",f,3,"FOCAL LENGHT IN INCHES",&dstat);
	sstat = stgkyd(src,"Aperture",&f);
	if (sstat==0) ffpkyd(dst,"APERTURE",f,4,"APERTURE AREA IN SQ-INCHES",&dstat);
	sstat = stgkys(src,"Filter",&filter);
	if (sstat==0) ffpkys(dst,"FILTER",filter,"OPTICAL FILTER NAME",&dstat);
	cmpack_free(filter);
	sstat = stgkyi(src,"Background",&i);
	if (sstat==0) ffpkyj(dst,"BACKGRND",i,"BACKGROUND FOR DISPLAY",&dstat);
	sstat = stgkyi(src,"Range",&i);
	if (sstat==0) ffpkyj(dst,"RANGE",i,"RANGE FOR DISPLAY",&dstat);
	sstat = stgkys(src,"History",&history);
	if (sstat==0) ffphis(dst,history,&dstat);
	cmpack_free(history);
	/* Can fail if the string contains characters <32 or >126 */
	sstat = stgkys(src,"Observer",&observer);
	if (sstat==0) ffpkys(dst,"OBSERVER",observer,"OBSERVERS NAME",&dstat);
	cmpack_free(observer);
	dstat = 0;
	sstat = stgkyi(src,"Pedestal",&i);
	if (sstat==0) ffpkyj(dst,"PEDESTAL",i,"PEDESTAL OFFSET",&dstat);
	sstat = stgkyi(src,"Track_time",&i);
	if (sstat==0) ffpkyj(dst,"TRAKTIME",i,"TRACK PERIOD IN SECONDS",&dstat);
	sstat = stgkyi(src,"Sat_level",&i);
	if (sstat==0) ffpkyj(dst,"SATURATE",i,"SATURATION LEVEL",&dstat);	
	return dstat;
}

/* Add string to history */
int sbig_copyheader(tHandle fs, fitsfile *fits)
{
	int dstat = copyheader((struct stfile*)fs, fits);
	return (dstat!=0 ? CMPACK_ERR_WRITE_ERROR : 0);
}

/* Convert SBIG file to FITS file */
int sbig_copy(tHandle fs, fitsfile *fits, int bitpix, int flipx, int flipy, 
			const CmpackBorder *border, CmpackConsole *con)
{
	unsigned short *image;
	char	msg[MAXLINE], origin[128], actdate[128];
	int		y, len, sstat=0, dstat=0, nx, ny;
	long	naxes[2];
	time_t	tsec;
	struct tm *actim;

	/* Image size */
	sstat = stghpr((struct stfile*)fs, &nx, &ny);
	if (sstat!=0) {
		if (con)
			printerr(con, sstat);
		return CMPACK_ERR_READ_ERROR;
	}

	/* Output data format */
	if (bitpix==CMPACK_BITPIX_AUTO)
		bitpix = CMPACK_BITPIX_USHORT;
	if (bitpix > 0 && bitpix < CMPACK_BITPIX_USHORT) {
		printout(con, 0, "Please, use higher image data precision.");
		return CMPACK_ERR_INVALID_BITPIX;
	}

	/* Set output data format */
	naxes[0] = nx; 
	naxes[1] = ny;
	ffphps(fits,bitpix,2,naxes,&dstat);
	if (dstat!=0) {
		if (con) {
			ffgerr(dstat, msg);
			printout(con, 0, msg);
		}
		return CMPACK_ERR_WRITE_ERROR;
	}

	/* Copy file header */
	dstat = copyheader((struct stfile*)fs, fits);
	if (dstat!=0) {
		if (con) {
			ffgerr(dstat, msg);
			printout(con, 0, msg);
		}
		return CMPACK_ERR_WRITE_ERROR;
	}

	/* Origin and date of creation */
	tsec = time(&tsec);
	actim = localtime(&tsec);
	sprintf(origin, "%s %s", PACKAGE_NAME, VERSION);
	ffukys(fits, "ORIGIN", origin, "CREATOR NAME AND VERSION", &dstat);
	sprintf(actdate,"%04d-%02d-%02d",actim->tm_year+1900,actim->tm_mon+1,actim->tm_mday);
 	ffukys(fits, "DATE", actdate, "DATE OF CREATION", &dstat);

	/* Read image data */
	sstat = 0;
	image = (unsigned short*)cmpack_calloc(nx*ny, sizeof(unsigned short));
	stg2duj((struct stfile*)fs,flipx,flipy,image);
	if (sstat!=0) { 
		if (con)
			printerr(con, sstat);
		cmpack_free(image); 
		return CMPACK_ERR_READ_ERROR; 
	}

	/* Clear border */
	if (border->left>0) {
		len = (border->left < nx ? border->left : nx);
		for (y=0; y<ny; y++)
			memset(image+(y*nx), 0, len*sizeof(unsigned short));
	}
	if (border->right) {
		len = (border->right < nx ? border->right: nx);
		for (y=0; y<ny; y++)
			memset(image+((y+1)*nx-len), 0, len*sizeof(unsigned short));
	}
	if (border->top>0) {
		len = (border->top < ny ? border->top : ny);
		for (y=0; y<len; y++)
			memset(image+(y*nx), 0, nx*sizeof(unsigned short));
	}
	if (border->bottom>0) {
		len = (border->bottom < ny ? border->bottom : ny);
		for (y=0; y<len; y++)
			memset(image+((ny-y-1)*nx), 0, nx*sizeof(unsigned short));
	}

	/* Write image data */
	ffp2dui(fits,1,nx,nx,ny,image,&dstat);
	if (dstat!=0) { 
		if (con) {
			ffgerr(dstat, msg);
			printout(con, 0, msg);
		}
		cmpack_free(image); 
		return CMPACK_ERR_WRITE_ERROR; 
	}

	/* Normal return */
	cmpack_free(image);
	return 0;
}
