/**************************************************************

parser.c (C-Munipack project)
Photometry file parser
Copyright (C) 2004 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <expat.h>

#include "hash.h"
#include "comfun.h"
#include "cmpack_common.h"
#include "phtparser.h"
#include "phtstars.h"
#include "phtfile.h"

#define BUFFSIZE 4096

/* Context structure for catalogue file parser */
struct _CmpackPhtParser
{
	CmpackPhtFile *file;				/**< Photometry file context */
	int headonly;						/**< Load only file header */
	int state;							/**< Parser state */
	int result;							/**< Parser result code */
	int level;							/**< Parser level */
	CmpackPhtStar *actstar;				/**< Active star index */
	char *element;						/**< Element name */
	char *value;						/**< Character data */
	char *comment;						/**< Comment */
};
typedef struct _CmpackPhtParser CmpackPhtParser;

/*************************   HELPER FUNCTIONS   ***************/

/* Returns pointer to the buffer which holds the value of the XML attribute.
 * The 'key' parameter specifies the name of the attribute. Do not free the
 * memory.
 */
static const char *get_attr(const char **atlist, const char *key)
{
	int i;
	
  	for (i=0; atlist[i]; i+=2) {
	  	if (strcmp(atlist[i], key)==0)
		  	return atlist[i+1];
  	}
  	return NULL;
}

static char *set_string(char *dst, const char *src)
{
	if (src) {
		dst = (char*)cmpack_realloc(dst, (strlen(src)+1)*sizeof(char));
		strcpy(dst, src);
	} else {
		cmpack_free(dst);
		dst = NULL;
	}
	return dst;
}

static char *append_bytes(char *dst, const char *src, int srclen)
{
	int dstlen;

	if (srclen>0) {
		if (dst) {
			dstlen = (int)strlen(dst);
			dst = (char*)cmpack_realloc(dst, (dstlen+srclen+1)*sizeof(char));
			memcpy(dst + dstlen, src, srclen);
			dst[dstlen+srclen] = '\0';
		} else {
			dst = (char*)cmpack_malloc((srclen+1)*sizeof(char));
			memcpy(dst, src, srclen);
			dst[srclen] = '\0';
		}
	}
	return dst;
}

/* Strip leading and trailing white spaces */
static char *strip_string(char *str)
{
	char *ptr;

	if (str) {
		/* Remove leading spaces */
		for (ptr=str; *ptr!='\0'; ptr++) {
			if (*ptr>32 || *ptr<0)
				break;
		}
		if (*ptr=='\0') {
			str[0] = '\0';
			return str;
		}
		memmove(str, ptr, strlen(ptr)+1);
		/* Remove trailing spaces */
		for (ptr=str+strlen(str)-1; ptr>=str; ptr--) {
			if (*ptr>32 || *ptr<0)
				break;
		}
		*(ptr+1) = '\0';
	}
	return str;
}

/**********************   XML PARSER HANDLERS   *****************/

/* Obligatory callback from XML parser */
static int XMLCALL XMLUnknownEncodingHandler(void *data, const XML_Char *encoding, XML_Encoding *info)
{
 	int i;
 	
 	for (i=0; i<256; i++) 
 		info->map[i] = i;
 	info->data = NULL;
 	info->convert = NULL;
 	info->release = NULL;
 	return XML_STATUS_OK;
}

/* Start element callback. This is called whenever the XML parser finds
 * a start of the element. The 'aname' parameter is name of the element,
 * 'atlist' is the list of attributes (key+val) pairs.
 */
static void XMLCALL XMLStartElementHandler(void *adata, const char *aname, const char **atlist)
{
	int i, aperture, xok, yok;
	char *aux, *endptr;
  	CmpackPhtParser *f = (CmpackPhtParser*) adata;
	CmpackPhtMeasurement maginfo;  	

	switch (f->state) 
	{
	case 0:
		/* Root element */
		if (strcmp(aname, "phot")==0) {
			f->state = 1;
			f->level = 0;
		} else {
			f->result = CMPACK_ERR_UNKNOWN_FORMAT;
		}
		break;
		
	case 1:
		/* First level elements */
		if (strcmp(aname, "head")==0) {
			f->level++;
			f->state = 2;
			break;
		}
		if (strcmp(aname, "apertures")==0 && !f->headonly) {
			/* Apertures */
			f->level++;
			f->state = 4;
			break;
		}
		if (strcmp(aname, "body")==0 && !f->headonly) {
			/* Stars */
			f->level++;
			f->state = 5;
			break;
		}
		if (f->state==1) {
			/* Unknown section -> ignore it */
			f->level++;
			f->state = 3;
			break;
		}
		break;
		
	case 2:
		/* Read header <key>val</key> */
		if (f->level==1) {
			f->element = set_string(f->element, aname);
			f->comment = set_string(f->comment, NULL);
			f->state = 21;
		}
		f->level++;
		break;

	case 21:
		/* Ignore all subelements, but store value and comment */
		f->level++;
		break;
		
	case 3:
		/* Ignore unknown header */
		f->level++;
		break;
		
	case 4:
		/* Apertures: <aper id="n" key=val/> */
		if (f->level==1 && strcmp(aname, "aper")==0) {
			aux = (char*) get_attr(atlist, "id");
			i = strtol(aux, &endptr, 10);
			if (endptr!=aux && i>=0) {
				CmpackPhtAperture *aperture = pht_aper_new(i);
	  			for (i=0; atlist[i]; i+=2) {
					if (strcmp(atlist[i], "radius")==0) {
						aperture->mask |= CMPACK_MASK_RADIUS;
						aperture->radius = atof(atlist[i+1]);
						pht_aptab_add(&f->file->ap, aperture);
						break;
					}
	  			}
			}
		}
		f->level++;
		break;
		
	case 5:
		/* Stars: <st id="n" key1="val1" key2="val2" ...> */
		if (f->level==1 && strcmp(aname, "object")==0) {
			aux = (char*) get_attr(atlist, "id");
			i = strtol(aux, &endptr, 10);
			if (endptr!=aux && i>=0) {
				CmpackPhtStar *star = pht_star_new(i);
				xok = yok = 0;
				star->refid = -1;
  				for (i=0; atlist[i]; i+=2) {
					if (strcmp(atlist[i], "x")==0) {
						xok = 1;
						star->center[0] = atof(atlist[i+1]);
					} else 
					if (strcmp(atlist[i], "y")==0) {
						yok = 1;
						star->center[1] = atof(atlist[i+1]);
					} else
					if (strcmp(atlist[i], "x-ref")==0) {
						star->refid = atoi(atlist[i+1]);
					}
				}
				if (xok && yok)
					star->mask |= CMPACK_MASK_POS;
				pht_stars_add(&f->file->st, star);
				f->actstar = star;
				f->state = 6;
			}
		}
		f->level++;
		break;
		
	case 6:
		/* Magnitude data: <p a="aperture" c="code" m="mag" e="err"/>  */
		if (f->level==2 && f->actstar!=NULL) {
			if (strcmp(aname, "p")==0) {
				aperture = 0;
				memset(&maginfo, 0, sizeof(CmpackPhtMeasurement));
				aux = (char*) get_attr(atlist, "a");
				if (aux) {
					aperture = strtol(aux, &endptr, 10);
					if (aux==endptr) aperture = 0;
				}
				aux = (char*) get_attr(atlist, "m");
				if (aux) {
					maginfo.magnitude = strtod(aux, &endptr);
					if (aux!=endptr) maginfo.mask |= CMPACK_MASK_MAG;
				}
				aux = (char*) get_attr(atlist, "e");
				if (aux) {
					maginfo.error = strtod(aux, &endptr);
					if (aux!=endptr) maginfo.mask |= CMPACK_MASK_ERR;
				}
				aux = (char*) get_attr(atlist, "c");
				if (aux) {
					maginfo.code = strtol(aux, &endptr, 10);
					if (aux!=endptr) maginfo.mask |= CMPACK_MASK_CODE;
				}
				if (maginfo.mask!=0 && aperture>0) {
					pht_star_set_data(f->actstar, aperture, &maginfo);
				}
			}
		}
		f->level++;
		break;
		
	default:
		f->result = CMPACK_ERR_READ_ERROR;
	}
}

/* End element handler. XML parser calls this function whenever the end of
 * an element is found.
 */
static void XMLCALL XMLEndElementHandler(void *adata, const char *aname)
{
  	CmpackPhtParser *f = (CmpackPhtParser*) adata;
  	
	switch (f->state) 
	{
	case 2:
	case 3:
	case 4:
	case 5:
		f->level--;
		if (f->level == 0)
			f->state = 1;
		break;
		
	case 21:
		f->level--;
		if (f->level == 1) {
			if (f->element && f->value) {
				f->value = strip_string(f->value);
				f->comment = strip_string(f->comment);
				pht_head_set(&f->file->hd, f->element, f->value, f->comment);
			}
			cmpack_free(f->element);
			f->element = NULL;
			cmpack_free(f->value);
			f->value = NULL;
			cmpack_free(f->comment);
			f->comment = NULL;
			f->state = 2;
		}
		break;

	case 6:
		f->level--;
		if (f->level == 1)
			f->state = 5;
		break;
	}
}

/* Character data handler */
static void XMLCALL XMLCharacterDataHandler(void *adata, const char *data, int len)
{
  	CmpackPhtParser *f = (CmpackPhtParser*) adata;

	switch (f->state)
	{
	case 21:
		f->value = append_bytes(f->value, data, len);
		break;
	}
}

/* Comment handler */
static void XMLCALL XMLCommentHandler(void *adata, const char *data)
{
  	CmpackPhtParser *f = (CmpackPhtParser*) adata;

	switch (f->state)
	{
	case 21:
		f->comment = set_string(f->comment, data);
		break;
	}
}

/****************   PUBLIC FUNCTIONS   ****************************/

/* Load a XML photometry file. Returns zero on success or one of MPK_xxx error codes.
 * Uses Expat library. Parsing context is held in internal file structure,
 * passed as user data (xml_xxx fields).
 */
int pht_loadxml(CmpackPhtFile *f, FILE *from, CmpackLoadMode flags)
{
	size_t len;
	int done, res;
	char buf[BUFFSIZE];
	CmpackPhtParser p;
	XML_Parser parser;

	memset(&p, 0, sizeof(CmpackPhtParser));
	p.file = f;
	p.headonly = flags & CMPACK_LOAD_HEADONLY;
	parser = XML_ParserCreate(NULL);
	XML_SetUserData(parser, &p);
	XML_SetElementHandler(parser, XMLStartElementHandler, XMLEndElementHandler);
	XML_SetUnknownEncodingHandler(parser, XMLUnknownEncodingHandler, NULL);
	XML_SetCharacterDataHandler(parser, XMLCharacterDataHandler);
	XML_SetCommentHandler(parser, XMLCommentHandler);
	do {
		len = fread(buf, 1, BUFFSIZE, from);
		if (ferror(from)) 
			return CMPACK_ERR_READ_ERROR;
		done = feof(from);
		if (XML_Parse(parser, buf, (int)len, done)==XML_STATUS_ERROR)
   			return CMPACK_ERR_READ_ERROR;
	} while (!done);
	res = p.result;
	XML_ParserFree(parser);
	cmpack_free(p.value);
	cmpack_free(p.comment);
	cmpack_free(p.element);
	return res;
}
