/**************************************************************

phot_stars.c (C-Munipack project)
Table of stars
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "comfun.h"
#include "console.h"
#include "quicksort.h"
#include "hash.h"
#include "cmpack_common.h"
#include "cmpack_phtfile.h"
#include "phtstars.h"

#define ALLOC_BY 	64

/********************   Local functions    ***********************/

/* Dump measurement */
static void mag_dump(CmpackConsole *con, CmpackMag *mag)
{
	char aux[MAXLINE], aux1[MAXLINE];

	sprintf(aux, "\t\t\t\tAp=%d", mag->aperture);
	if (mag->data.mask & CMPACK_MASK_MAG) {
		sprintf(aux1, "; Mag=%.*f", MAG_PRECISION, mag->data.magnitude);
		strcat(aux, aux1);
	}
	if (mag->data.mask & CMPACK_MASK_ERR) {
		sprintf(aux1, "; Err=%.*f", MAG_PRECISION, mag->data.error);
		strcat(aux, aux1);
	}
	if (mag->data.mask & CMPACK_MASK_CODE) {
		sprintf(aux1, "; Code=%d", mag->data.code);
		strcat(aux, aux1);
	}
	printout(con, 1, aux);
}

/* Write measurement */
static void mag_write(CmpackMag *mag, FILE *to)
{	
	fprintf(to, "\t\t<p a=\"%d\"", mag->aperture);
	if (mag->data.mask & CMPACK_MASK_MAG)
		fprintf(to, " m=\"%.*f\"", MAG_PRECISION, mag->data.magnitude);
	if (mag->data.mask & CMPACK_MASK_ERR)
		fprintf(to, " e=\"%.*f\"", MAG_PRECISION, mag->data.error);
	if ((mag->data.mask & CMPACK_MASK_CODE) && mag->data.code!=0)
		fprintf(to, " c=\"%d\"", mag->data.code);
	fprintf(to, "/>\n");
}

/* Copy measurement */
static void mag_copy(CmpackMag *to, const CmpackMag *from)
{
	to->aperture = from->aperture;
	to->data = from->data;
}

/* Clear all measurements */
static void pht_mag_clear(CmpackMagList *mag)
{
	cmpack_free(mag->list);
	mag->list = NULL;
	mag->count = mag->capacity = 0;
}

/* Copy parameters and measurements */
static void pht_mag_copy(CmpackMagList *dst, const CmpackMagList *src)
{
	int j;

	pht_mag_clear(dst);

	if (src->count>0) {
		dst->capacity = dst->count = src->count;
		dst->list = (CmpackMag*) cmpack_calloc(dst->capacity, sizeof(CmpackMag));
		for (j=0; j<dst->count; j++) 
			mag_copy(&dst->list[j], &src->list[j]);
	}
}

/* Dump measurements */
static void pht_mag_dump(void *gc, CmpackMagList *tab)
{
	int j;
	
	for (j=0; j<tab->count; j++) 
		mag_dump(gc, &tab->list[j]);
}

/* Write measurements to file */
static void pht_mag_write(CmpackMagList *tab, FILE *to)
{
	int j;
	
	for (j=0; j<tab->count; j++) {
		if (tab->list[j].data.mask!=0)
			mag_write(&tab->list[j], to);
	}
}

/* Add magnitude */
static int pht_mag_add(CmpackMagList *tab, int apid)
{
	int index = tab->count;
	if (tab->count>=tab->capacity) {
		tab->capacity += ALLOC_BY;
		tab->list = (CmpackMag*)cmpack_realloc(tab->list, tab->capacity*sizeof(CmpackMag));
	}
	memset(&tab->list[index], 0, sizeof(CmpackMag));
	tab->list[index].aperture = apid;
	tab->count++;
	return index;
}

/* Search magnitude */
static int pht_mag_search(CmpackMagList *mag, int aperture)
{
	int i;

	for (i=0; i<mag->count; i++) {
		if (mag->list[i].aperture == aperture)
			return i;
	}
	return -1;
}

/* Search magnitude */
static CmpackPhtMeasurement *pht_mag_get(CmpackMagList *mag, int aperture)
{
	int i = pht_mag_search(mag, aperture);
	if (i>=0)
		return &mag->list[i].data;
	else
		return NULL;
}

/* Clear all measurements */
static void star_clear(CmpackPhtStar *st)
{
	st->mask = 0;
	pht_mag_clear(&st->mag);
}

/* Clear all measurements */
static void star_clean(void *ctx)
{
	CmpackPhtStar *st = (CmpackPhtStar*)ctx;
	star_clear(st);
	gc_clean(ctx);
}

/* Make new star context */
static CmpackPhtStar *star_create(int starid)
{
	CmpackPhtStar *st = gc_create(sizeof(CmpackPhtStar));
	st->gc.clean = star_clean;
	st->id = starid;
	st->refid = -1;
	return st;
}

/* Copy parameters and measurements */
static void star_copy(CmpackPhtStar *dst, const CmpackPhtStar *src)
{
	dst->mask = src->mask;
	memcpy(dst->center, src->center, 2*sizeof(double));
	dst->refid = src->refid;
	pht_mag_copy(&dst->mag, &src->mag);
}

/********************   Public functions    ***********************/

/* Initialize table of stars */
void pht_stars_init(CmpackStarTab *tab)
{
	memset(tab, 0, sizeof(CmpackStarTab));
}

/* Dump stars */
void pht_stars_dump(CmpackConsole *con, CmpackStarTab *tab)
{
	int i;
	char aux[MAXLINE];
	
	sprintf(aux, "\tStars: %d items", tab->count);
	printout(con, 1, aux);
	for (i=0; i<tab->count; i++) {
		CmpackPhtStar *st = tab->list[i];
		sprintf(aux, "\t\tStar ID   : %d", st->id);
		printout(con, 1, aux);
		/* Dump parameters */
		if (st->mask & CMPACK_MASK_POS) {
			sprintf(aux, "\t\t\tPosition: %.3f, %.3f", st->center[0], st->center[1]);
			printout(con, 1, aux);
		}
		if (st->refid>=0) {
			sprintf(aux, "\t\t\tRef-id  : %d", st->refid);
			printout(con, 1, aux);
		}
		/* Dump magnitudes */
		printout(con, 1, "\t\t\tMagnitudes:");
		pht_mag_dump(con, &st->mag);
	}
}

/* Write all stars to file */
void pht_stars_write(CmpackStarTab *tab, FILE *to)
{
	int i;
	
	for (i=0; i<tab->count; i++) {
		CmpackPhtStar *st = tab->list[i];
		fprintf(to, "\t<object id=\"%d\"", st->id); 
		/* Write attributes */
		if (st->mask & CMPACK_MASK_POS) 
			fprintf(to, " x=\"%.3f\" y=\"%.3f\"", st->center[0], st->center[1]);
		if (st->refid>=0) 
			fprintf(to, " x-ref=\"%d\"", st->refid);
		if (st->mag.count==0) {
			/* Table of magnitudes is empty */
			fprintf(to, "/>\n");
		} else {
			fprintf(to, ">\n");
			/* Write measurements */
			pht_mag_write(&st->mag, to);
			fprintf(to, "\t</object>\n");
		}
	}
}

/* Make copy of stars */
void pht_stars_copy(CmpackStarTab *dsttab, const CmpackStarTab *srctab)
{
	int i;
	
	pht_stars_clear(dsttab);

	if (srctab->list && srctab->count>0) {
		dsttab->count = dsttab->capacity = srctab->count;
		dsttab->list = (CmpackPhtStar**) cmpack_malloc(dsttab->capacity*sizeof(CmpackPhtStar*));
		for (i=0; i<dsttab->count; i++) {
			dsttab->list[i] = star_create(srctab->list[i]->id);
			star_copy(dsttab->list[i], srctab->list[i]);
		}
	}
}

/* Delete all stars */
void pht_stars_clear(CmpackStarTab *tab)
{
	int i;
	
	if (tab->list) {
		for (i=0; i<tab->count; i++)
			cmpack_unref(tab->list[i]);
		cmpack_free(tab->list);
		tab->list = NULL;
	}
	tab->count = tab->capacity = 0;
}

/* Insert a star */
int pht_stars_add(CmpackStarTab *tab, CmpackPhtStar *star)
{
	int i = tab->count;
	if (tab->count>=tab->capacity) {
		tab->capacity += ALLOC_BY;
		tab->list = (CmpackPhtStar**) cmpack_realloc(tab->list, tab->capacity*sizeof(CmpackPhtStar*));
	}
	tab->list[i] = star;
	tab->count++;
	return i;	
}

/* Create a new star */
CmpackPhtStar *pht_star_new(int id)
{
	return star_create(id);
}

/* Find star by identifier, return its index */
int pht_stars_search(CmpackStarTab *tab, int starid)
{
	int i;

	for (i=0; i<tab->count; i++) {
		if (tab->list[i]->id == starid)
			return i;
	}
	return -1;
}

/* Find star by reference identifier, return its index */
int pht_stars_search_by_refid(CmpackStarTab *tab, int refid)
{
	int i;

	if (refid>=0) {
		for (i=0; i<tab->count; i++) {
			if (tab->list[i]->refid == refid)
				return i;
		}
	}
	return -1;
}

/* Delete a star */
void pht_stars_delete(CmpackStarTab *tab, int index)
{
	CmpackPhtStar *item = tab->list[index];
	/* Delete item */
	cmpack_unref(item);
	tab->count--;
	if (tab->count==0) {
		/* If there is no item, clear the list */
		cmpack_free(tab->list);
		tab->list = NULL;
		tab->count = tab->capacity = 0;
	} else {
		/* Pack the list (move all following items to the left) */
		size_t len = tab->count - index;
		if (len>0) {
			CmpackPhtStar **aux = (CmpackPhtStar**) cmpack_malloc(len*sizeof(CmpackPhtStar*));
			memcpy(aux, item+1, len*sizeof(CmpackPhtStar*));
			memcpy(item, aux, len*sizeof(CmpackPhtStar*));
			cmpack_free(aux);
		}
	}
}

/* Clear measurements */
void pht_star_clear(CmpackPhtStar *star)
{
	pht_mag_clear(&star->mag);
}

/* Find measurement */
CmpackPhtMeasurement *pht_star_get_data(CmpackPhtStar *star, int aperture)
{
	return pht_mag_get(&star->mag, aperture);
}

/* Add new measurement */
void pht_star_set_data(CmpackPhtStar *star, int aperture, const CmpackPhtMeasurement *data)
{
	int index = pht_mag_search(&star->mag, aperture);
	if (index>=0) {
		star->mag.list[index].data = *data;
	} else {
		index = pht_mag_add(&star->mag, aperture);
		if (index>=0) 
			star->mag.list[index].data = *data;
	}
}

/****************************   PUBLIC INTERFACE   ****************************/

/* Find aperture / create new aperture */
CmpackPhtStar *cmpack_pht_star_new(int starid)
{
	if (starid>=0)
		return pht_star_new(starid);
	else
		return NULL;
}

/* Set magnitude for active star and active aperture */
int cmpack_pht_star_get_id(CmpackPhtStar *s)
{
	return s->id;
}

/* Set magnitude for active star and active aperture */
void cmpack_pht_star_set_measurement(CmpackPhtStar *s, int aperture, const CmpackPhtMeasurement *data)
{
	if (aperture>0) 
		pht_star_set_data(s, aperture, data);
}

/* Set magnitude for active star and active aperture */
int cmpack_pht_star_get_measurement(CmpackPhtStar *s, int aperture, CmpackPhtMeasurement *data)
{
	if (aperture>0) {
		CmpackPhtMeasurement *item = pht_star_get_data(s, aperture);
		if (item) {
			if (data) *data = *item;
			return 0;
		}
	}
	if (data) data->mask = 0;
	return CMPACK_ERR_MAG_NOT_FOUND;
}

/* Set magnitude for active star and active aperture */
void cmpack_pht_star_set_center(CmpackPhtStar *s, double x, double y)
{
	s->mask |= CMPACK_MASK_POS;
	s->center[0] = x;
	s->center[1] = y;
}

/* Set magnitude for active star and active aperture */
int cmpack_pht_star_get_center(CmpackPhtStar *s, double *x, double *y)
{
	if (s->mask & CMPACK_MASK_POS) {
		if (x) *x = s->center[0];
		if (y) *y = s->center[1];
		return 1;
	} else {
		if (x) *x = 0.0;
		if (y) *y = 0.0;
		return 0;
	}
}

/* Set magnitude for active star and active aperture */
void cmpack_pht_star_set_refid(CmpackPhtStar *s, int refid)
{
	s->refid = refid;
}

/* Set magnitude for active star and active aperture */
int cmpack_pht_star_get_refid(CmpackPhtStar *s)
{
	return s->refid;
}

/* Clear all measurements for specified star */
void cmpack_pht_star_clear_measurements(CmpackPhtStar *s)
{
	pht_star_clear(s);
}
