/**************************************************************

head.c (C-Munipack project)
Table header
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hash.h"
#include "comfun.h"
#include "console.h"
#include "context.h"
#include "cmpack_common.h"
#include "tabhead.h"

/*************************   Constants  ***********************/

#define ALLOC_BY 	64

/***********************      Datatypes      ******************/

/* Single header item */
struct _CmpackTabHeadItem {
	char *key;						/**< Parameter name (case insensitive) */
	char *val;						/**< Parameter value */
};
	
/* File header */
struct _CmpackTabHeader {
	int	count;						/**< Number of items */
	int	capacity;					/**< Number of allocated records */
	CmpackTabHeadItem *list;		/**< List of header items */
	CmpackHashTable hash;			/**< Hash table of items */
};

/********************   Local functions    ***********************/

/* Init head item */
static void headitem_init(CmpackTabHeadItem *item, const char *key, const char *val)
{
	item->key = cmpack_strdup(key);
	item->val = cmpack_strdup(val);
}

/* Makes copy of header item */
static void headitem_copy(CmpackTabHeadItem *dst, const CmpackTabHeadItem *src)
{
	dst->key = cmpack_strdup(src->key);
	dst->val = cmpack_strdup(src->val);
}

/* Change value of header item */
static void headitem_setval(CmpackTabHeadItem *item, const char *val)
{
	if (strlen(val)>0) {
		item->val = (char*) cmpack_realloc(item->val, (strlen(val)+1)*sizeof(char));
		strcpy(item->val, val);
	} else {
		cmpack_free(item->val);
		item->val = NULL;
	}
}

/* Destroy head item */
static void headitem_clear(CmpackTabHeadItem *item)
{
	cmpack_free(item->key);
	cmpack_free(item->val);
	memset(item, 0, sizeof(CmpackTabHeadItem));
}

/* Dump head item to stdout */
static void headitem_dump(CmpackTabHeadItem *item, CmpackConsole *con)
{
	char aux[MAXLINE];

	sprintf(aux, "\t\t%s", item->key);
	if (item->val) {
		strcat(aux, " = ");
		strcat(aux, item->val);
	}
	printout(con, 1, aux);
}

/* Write item to stdout */
static void headitem_write(CmpackTabHeadItem *item, FILE *to)
{
	if (item->val)
		fprintf(to, "%s: %s", item->key, item->val);
}

/********************   Public functions    ***********************/

/* Create a new header */
CmpackTabHeader *tab_head_create(void)
{
	CmpackTabHeader *head = (CmpackTabHeader*)cmpack_malloc(sizeof(CmpackTabHeader));
	head->list = NULL;
	head->count = head->capacity = 0;
	hash_init(&head->hash);
	return head;
}	

/* Free the header */
void tab_head_free(CmpackTabHeader *head)
{
	int j;
	
	if (head->list) {
		for (j=0; j<head->count; j++)
			headitem_clear(&head->list[j]);
		cmpack_free(head->list);
		head->list = NULL;
	}
	head->count = head->capacity = 0;
	hash_clear(&head->hash);
	cmpack_free(head);
}

void tab_head_dump(CmpackTabHeader *head, CmpackConsole *con)
{
	int j;
	char aux[MAXLINE];
	
	sprintf(aux, "\tHeader: %d items\n", head->count);
	printout(con, 1, aux);
	for (j=0; j<head->count; j++)
		if (head->list[j].key)
			headitem_dump(&head->list[j], con);
}	

void tab_head_write(CmpackTabHeader *head, FILE *to, const char *prefix)
{
	int j;
	
	if (prefix)
		fprintf(to, "%s ", prefix);
	for (j=0; j<head->count; j++) {
		if (j>0)
			fprintf(to, ", ");
		headitem_write(&head->list[j], to);
	}
	fprintf(to, "\n");
}

void tab_head_copy(CmpackTabHeader *dsthead, const CmpackTabHeader *srchead)
{
	int j;
	
	tab_head_clear(dsthead);
	if (srchead->count>0) {
		dsthead->capacity = dsthead->count = srchead->count;
		dsthead->list = (CmpackTabHeadItem*) cmpack_calloc(dsthead->capacity, sizeof(CmpackTabHeadItem));
		if (dsthead->list) {
			for (j=0; j<srchead->count; j++) {
				headitem_copy(&dsthead->list[j], &srchead->list[j]);
				hash_insert(&dsthead->hash, dsthead->list[j].key, &dsthead->list[j]);
			}
		}
	}
}	


void tab_head_clear(CmpackTabHeader *head)
{
	int j;
	
	if (head->list) {
		for (j=0; j<head->count; j++)
			headitem_clear(&head->list[j]);
		cmpack_free(head->list);
		head->list = NULL;
	}
	head->count = head->capacity = 0;
	hash_clear(&head->hash);
}	

void tab_head_set(CmpackTabHeader *head, const char *key, const char *val)
{
	CmpackTabHeadItem *item = (CmpackTabHeadItem*) hash_search(&head->hash, key);
	if (item) {
		/* Modify existing item */
		if (val) headitem_setval(item, val);
	} else {
		/* Add new item */
		if (head->count>=head->capacity) {
			head->capacity += ALLOC_BY;
			head->list = (CmpackTabHeadItem*) cmpack_realloc(head->list, head->capacity*sizeof(CmpackTabHeadItem));
		}
		headitem_init(&head->list[head->count], key, val);
		hash_insert(&head->hash, key, &head->list[head->count]);
		head->count++;
	}
}

int tab_head_count(CmpackTabHeader *head)
{
	if (head)
		return head->count;
	return 0;
}

CmpackTabHeadItem *tab_head_get(CmpackTabHeader *head, int index)
{
	if (index>=0 && index<head->count)
		return &head->list[index];
	return NULL;
}

CmpackTabHeadItem *tab_head_find(CmpackTabHeader *head, const char *key)
{
	return (CmpackTabHeadItem*) hash_search(&head->hash, key);
}

/* Delete item */
void tab_head_delete(CmpackTabHeader *head, CmpackTabHeadItem *item)
{
	hash_delete(&head->hash, item->key, item);
	headitem_clear(item);
	head->count--;
	if (head->count==0) {
		/* If there is no item, clear the list */
		cmpack_free(head->list);
		head->list = NULL;
		head->count = head->capacity = 0;
	} else {
		/* Pack the list (move all following items to the left) */
		size_t len = head->count - (item - head->list);
		if (len>0) {
			CmpackTabHeadItem *aux = (CmpackTabHeadItem*) cmpack_malloc(len*sizeof(CmpackTabHeadItem));
			memcpy(aux, item+1, len*sizeof(CmpackTabHeadItem));
			memcpy(item, aux, len*sizeof(CmpackTabHeadItem));
			cmpack_free(aux);
		}
	}
}

/* Get parameter by index */
const char *tab_head_item_key(CmpackTabHeadItem *item)
{
	if (item)
		return item->key;
	return NULL;
}

/* Get parameter by index */
const char *tab_head_item_val(CmpackTabHeadItem *item)
{
	if (item)
		return item->val;
	return NULL;
}
