/**************************************************************

table.c (C-Munipack project)
Tables -- public routines
Copyright (C) 2004 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#include "comfun.h"
#include "console.h"
#include "hash.h"
#include "cmpack_common.h"
#include "cmpack_table.h"
#include "tabhead.h"
#include "tabdata.h"
#include "tabparser.h"
#include "table.h"

/******************* Constants and definitions **********************/

#define BUFFSIZE 	4096

/********************** Local functions ****************************/

static void table_clean(void *ctx);

/* Init table context */
static CmpackTable *table_create(CmpackTableType type)
{
	CmpackTable *f = (CmpackTable *)gc_create(sizeof(CmpackTable));
	f->gc.clean = &table_clean;
	f->type = type;
	f->head = tab_head_create();
	f->data = tab_data_create();
	return f;
}

/* Clear content of the file */
static void table_clear(CmpackTable *t)
{
	tab_head_clear(t->head);
	tab_data_clear(t->data);
}	

/* Clean table context */
static void table_clean(void *ctx)
{
	CmpackTable *tab = (CmpackTable *)ctx;
	tab_head_free(tab->head);
	tab_data_free(tab->data);
	gc_clean(ctx);
}

/* Copy content of the table */
static void table_copy(CmpackTable *dsttab, const CmpackTable *srctab)
{
	dsttab->type = srctab->type;
	tab_head_copy(dsttab->head, srctab->head);
	tab_data_copy(dsttab->data, srctab->data);
}

/* Write table to file */
static void table_write(CmpackTable *t, FILE *f)
{
	const char *prefix;

	if (t->type == CMPACK_TABLE_TRACKLIST)
		prefix = "#";
	else
		prefix = NULL;
	tab_data_write_cols(t->data, f, prefix);
	tab_head_write(t->head, f, NULL);
	tab_data_write_data(t->data, f);
}

/*********************   Public functions    ***********************/

/* Create a table */
CmpackTable *cmpack_tab_init(CmpackTableType type)
{
	return table_create(type);
}

/* Test the file */
int cmpack_tab_test(const char *filename)
{
	int bytes, filesize;
	char buffer[2048];

	if (filename) {
		FILE *f = fopen(filename, "rb");
		if (f) {
			fseek(f, 0, SEEK_END);
			filesize = ftell(f);
			fseek(f, 0, SEEK_SET);
			bytes = (int)fread(buffer, 1, 2048, f);
			fclose(f);
			return cmpack_tab_test_buffer(buffer, bytes, filesize);
		}
	}
	return 0;
}

/* Test if the file is a photometry file */
int cmpack_tab_test_buffer(const char *buffer, int buflen, int filesize)
{
	int res;
	char *aux;

	res = 0;
	if (buffer && buflen>16 && filesize>16) {
		/* Convert buffer to a null-terminated string */
		aux = (char*)cmpack_malloc((buflen+1)*sizeof(char));
		memcpy(aux, buffer, buflen);
		aux[buflen] = '\0';
		res = (tab_format(aux)!=CMPACK_TABLE_UNSPECIFIED);
		cmpack_free(aux);
	}
	return res;
}

/* Opens file. You can open the file for reading only or in 
 * read-write mode. Returns pointer to internal file structure 
 * handle or NULL if failed.
 */
int cmpack_tab_load(CmpackTable **ptab, const char *filename, CmpackLoadMode flags)
{
	int res;
	FILE *f;
	CmpackTable *tab;
	
	*ptab = NULL;
	f = fopen(filename, "rb");
	if (!f) 
		return CMPACK_ERR_CANT_OPEN_SRC;
	tab = table_create(CMPACK_TABLE_UNSPECIFIED);
	res = tab_load(tab, f, flags);
	fclose(f);
	if (res!=0)
		cmpack_unref(tab);
	else
		*ptab = tab;
	return res;
}

/* Destroy a table */
int cmpack_tab_save(CmpackTable *tab, const char *filename)
{
	FILE *f;

	f = fopen(filename, "wb+");
	if (!f) 
		return CMPACK_ERR_OPEN_ERROR;

	table_write(tab, f);
	fclose(f);
	return CMPACK_ERR_OK;
}

/* Clear table content */
void cmpack_tab_clear(CmpackTable *tab)
{
	if (tab)
		table_clear(tab);
}

/* Copy a table */
void cmpack_tab_copy(CmpackTable *dsttab, const CmpackTable *srctab)
{
	if (dsttab && srctab)
		table_copy(dsttab, srctab);
}

/* Set table type */
void cmpack_tab_set_type(CmpackTable *tab, CmpackTableType type)
{
	if (tab)
		tab->type = type;
}

/* Get table type */
CmpackTableType cmpack_tab_get_type(CmpackTable *tab)
{
	if (tab)
		return tab->type;
	else
		return CMPACK_TABLE_UNSPECIFIED;
}

/*********************   Table header     ***********************/

/* Set value of parameter in table header */
void cmpack_tab_pkys(CmpackTable *tab, const char *key, const char *val)
{
	if (tab && key)
		tab_head_set(tab->head, key, val);
}

/* Set value of parameter in table header */
void cmpack_tab_pkyi(CmpackTable *tab, const char *key, int val)
{
	char buf[16];

	if (tab && key) {
		sprintf(buf, "%d", val);
		tab_head_set(tab->head, key, buf);
	}
}

/* Set value of parameter in table header */
void cmpack_tab_pkyd(CmpackTable *tab, const char *key, double val, int prec)
{
	char buf[256];
	
	if (tab && key) {
		if (prec>16) 
			 prec = 16;
		if (fabs(val)<DBL_EPSILON)
			strcpy(buf, "0");
		else if (fabs(val)<1e+9)
			sprintf(buf, "%.*f", prec, val);
		else
			sprintf(buf, "%.*e", prec, val);
		tab_head_set(tab->head, key, buf);
	}
}

/* Get value of parameter in table header */
const char *cmpack_tab_gkys(CmpackTable *tab, const char *key)
{
	if (tab && key) {
		CmpackTabHeadItem *item = tab_head_find(tab->head, key);
		if (item)
			return tab_head_item_val(item);
	}
	return NULL;
}

/* Get value of parameter in table header */
int cmpack_tab_gkyi(CmpackTable *tab, const char *key, int defval)
{
	if (tab && key) {
		CmpackTabHeadItem *item = tab_head_find(tab->head, key);
		if (item) {
			const char *val = tab_head_item_val(item);
			if (val)
				return atoi(val);
		}
	}
	return defval;
}

/* Get value of parameter in table header */
double cmpack_tab_gkyd(CmpackTable *tab, const char *key, double defval)
{
	if (tab && key) {
		CmpackTabHeadItem *item = tab_head_find(tab->head, key);
		if (item) {
			const char *val = tab_head_item_val(item);
			if (val)
				return atof(val);
		}
	}
	return defval;
}

/* Return number of parameters stored in table header */
int cmpack_tab_nkey(CmpackTable *tab)
{
	if (tab) 
		return tab_head_count(tab->head); 
	else
		return 0;
}

/* Get the (key, value, comment) tripplets */
int cmpack_tab_gkyn(CmpackTable *tab, int index, const char **key, const char **val)
{
	if (tab) {
		CmpackTabHeadItem *item = tab_head_get(tab->head, index);
		if (item) {
			if (key) *key = tab_head_item_key(item);
			if (val) *val = tab_head_item_val(item);
			return 1;
		}
	}
	if (key) *key = NULL;
	if (val) *val = NULL;
	return 0;
}

/* Delete specified parameter */
void cmpack_tab_dkey(CmpackTable *tab, const char *key)
{
	if (tab && key) {
		CmpackTabHeadItem *item = tab_head_find(tab->head, key);
		if (item)
			tab_head_delete(tab->head, item);
	}
}

/*********************   Table structure     ***********************/

/* Returns number of columns */
int cmpack_tab_ncolumns(CmpackTable *tab)
{
	if (tab)
		return tab_data_ncols(tab->data);
	else
		return 0;
}

/* Append column */
int cmpack_tab_add_column(CmpackTable *tab, const char *name, CmpackColumnType type, int prec)
{
	if (tab && name) {
		if (strlen(name)>0) 
			return tab_data_add_col(tab->data, name, type, prec);
	}
	return -1;
}

/* Get column by its index */
CmpackTabColumn *cmpack_tab_get_column(CmpackTable *tab, int index)
{
	if (tab) {
		CmpackTabColumn *col = tab_data_get_col(tab->data, index);
		if (col)
			return (CmpackTabColumn*)cmpack_addref(col);
	}
	return NULL;
}

/* Get column by its name */
int cmpack_tab_find_column(CmpackTable *tab, const char *name)
{
	if (tab) 
		return tab_data_find_col(tab->data, name);
	return -1;
}

/* Get column name */
const char *cmpack_tab_column_get_name(CmpackTabColumn *col) 
{
	if (col) 
		return tab_col_name(col);
	else
		return NULL;
}

/* Get column name */
void cmpack_tab_column_set_name(CmpackTabColumn *col, const char *name) 
{
	if (col) 
		tab_col_setname(col, name);
}

/*********************    Table data    ***********************/

/* Returns number of rows */
int cmpack_tab_nrows(CmpackTable *tab)
{
	if (tab) 
		return tab_data_nrows(tab->data);
	else
		return 0;
}

/* Append empty rows to the end of a table */
int cmpack_tab_add_row(CmpackTable *tab)
{
	if (tab) 
		return tab_data_add_row(tab->data);
	else
		return -1;
}

/* Write table data */
void cmpack_tab_ptds(CmpackTable *tab, int row, int col, const char *value)
{
	if (tab) 
		tab_data_set(tab->data, col, row, value);
}

/* Write table data */
void cmpack_tab_ptdi(CmpackTable *tab, int row, int col, int value)
{
	if (tab) 
		tab_data_seti(tab->data, col, row, value);
}

/* Write table data */
void cmpack_tab_ptdd(CmpackTable *tab, int row, int col, double value)
{
	if (tab) 
		tab_data_setd(tab->data, col, row, value);
}

/* Read table data */
int cmpack_tab_gtds(CmpackTable *tab, int row, int col, char **value)
{
	if (tab) 
		return tab_data_get(tab->data, col, row, value);
	return 0;
}

/* Read table data */
int cmpack_tab_gtdi(CmpackTable *tab, int row, int col, int *value)
{
	if (tab) 
		return tab_data_geti(tab->data, col, row, value);
	return 0;
}

/* Read table data */
int cmpack_tab_gtdd(CmpackTable *tab, int row, int col, double *value)
{
	if (tab) 
		return tab_data_getd(tab->data, col, row, value);
	return 0;
}
