/**************************************************************

chart.cpp (C-Munipack project)
Widget which can draw a chart
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#define _USE_MATH_DEFINES

#include <stdlib.h>
#include <glib.h>
#include <math.h>
#include <string.h>

#include "chart.h"
#include "utils.h"
#include "configuration.h"
#include "scale.h"
#include "qphot.h"

#define CURSOR_SIZE		12

#define BAD_PIXEL_COLOR	0xFF0000

//--------------------------   HELPER FUNCTIONS   ----------------------------------

// grayscale mapping
static unsigned long grayscale(double x)
{
	x = 255.0*x;

	if (x<=0.0)
		return 0x000000;
	if (x>=255.0)
		return 0xFFFFFF;

	int ci = (int)x;
	return ci | (ci<<8) | (ci<<16);
}


//
// pseudocolors mapping (black, violet, blue, cyan, green, yellow, red, white) }
//
static unsigned long pseudocolor(double x)
{
	x = x*1785.0;

	if (x>=0x6FF)
		return 0xFFFFFF;
	if (x<=0)
		return 0x000000;

	int ci = (int)x;
	switch (ci>>8)
	{
	case 0:
		return (ci & 0xFF) | (ci & 0xFF)<<16;
	case 1:
		return (~ci & 0xFF)<<16 | 0xFF;
	case 2:
		return (ci & 0xFF)<<8 | 0xFF;
	case 3:
		return 0xFF00 | (~ci & 0xFF);
	case 4:
		return (ci & 0xFF)<<16 | 0xFF00;
	case 5:
		return 0xFF0000 | (~ci & 0xFF)<<8;
	case 6:
		return 0xFF0000 | (ci & 0xFF)<<8 | (ci & 0xFF);
	default:
		return 0;
	}
}

//--------------------------   CHART BASE CLASS   ----------------------------------

// Default constructor
CChartBase::CChartBase():m_DrawArea(NULL), m_Pixmap(NULL), m_NeedUpdate(true), 
	m_PixWidth(0), m_PixHeight(0), m_ApertureID(0), m_ApertureIndex(0), 
	m_StFocused(-1), m_StSelected(-1), m_EnableTrack(false), m_ShowImage(true), 
	m_ShowChart(true), m_BadStars(false), m_Pseudocolors(false), m_BadPixels(false), 
	m_ShowNonMatchedStars(true), m_HighlightNonMatchedStars(true), 
	m_HighlightSelectedStars(true), m_Invert(false), m_CursorX(-1), m_CursorY(-1), 
	m_pScale(NULL), m_Zoom(1.0), m_OffsetX(0.0), m_OffsetY(0.0), m_MapWidth(0), 
	m_MapHeight(0), m_Image(NULL), m_ImgWidth(0), m_ImgHeight(0), 
	m_IgnoreApertures(true)
	
{
	memset(&m_CCDData, 0, sizeof(tImageData));
	memset(&m_PhtData, 0, sizeof(tChartData));

	// Drawing area
	m_DrawArea = gtk_drawing_area_new();
	gtk_widget_set_size_request(GTK_WIDGET(m_DrawArea), 0, 0);
	gtk_widget_set_events(m_DrawArea, GDK_EXPOSURE_MASK | 
			GDK_LEAVE_NOTIFY_MASK | GDK_BUTTON_PRESS_MASK |
			GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK);

	// Signals used to handle backing pixmap 
	g_signal_connect(G_OBJECT(m_DrawArea), "expose_event", G_CALLBACK(expose_event), this);
	g_signal_connect(G_OBJECT(m_DrawArea), "configure_event", G_CALLBACK(configure_event), this);
	g_signal_connect(G_OBJECT(m_DrawArea), "motion_notify_event", G_CALLBACK(motion_notify_event), this);
	g_signal_connect(G_OBJECT(m_DrawArea), "leave-notify-event", G_CALLBACK(leave_notify_event), this);
	g_signal_connect(G_OBJECT(m_DrawArea), "button_press_event", G_CALLBACK(button_press_event), this);
}

// Destructor
CChartBase::~CChartBase()
{
	if (m_pScale)
		SetScale(NULL);
	if (m_PhtData.aper) 
		g_free(m_PhtData.aper);
	if (m_PhtData.stars)
		g_free(m_PhtData.stars);
	if (m_Image) 
		cairo_surface_destroy(m_Image);
	if (m_Pixmap)
		g_object_unref(m_Pixmap);
	if (m_CCDData.ccd)
		g_free(m_CCDData.ccd);
	gtk_widget_destroy(m_DrawArea);
}

// Set pointer to scale
void CChartBase::SetScale(CScale *pScale)
{
	CScale *old;
	if (pScale!=m_pScale) {
		old = m_pScale;
		m_pScale = NULL;
		if (old)
			old->SetChart(NULL);
		m_pScale = pScale;
		if (m_pScale)
			m_pScale->SetChart(this);
	}
}


// Clear internal data
void CChartBase::Clear(void)
{
	ClearImage();
	ClearPhotometry();
	ClearSelection();
	Resize();
	Invalidate();
}

// Resize
void CChartBase::Resize(void)
{
	m_MapWidth = (m_CCDData.width > m_PhtData.width ? m_CCDData.width : m_PhtData.width);
	m_MapHeight = (m_CCDData.height > m_PhtData.height ? m_CCDData.height : m_PhtData.height);
	OnImageSizeChanged();
}

// Invalidate the content of the backscreen buffer
void CChartBase::Invalidate()
{
	m_NeedUpdate = true;
	gtk_widget_queue_draw(m_DrawArea);
}

// Clear photometry data
void CChartBase::ClearPhotometry(void)
{
	if (m_StFocused>=0) {
		m_StFocused = -1;
	}
	if (m_StSelected>=0) {
		m_StSelected = -1;
	}
	if (m_PhtData.stars) {
		g_free(m_PhtData.stars);
		m_PhtData.stars = NULL;
	}
	m_PhtData.st_count = m_PhtData.width = m_PhtData.height = 0;
	if (m_PhtData.aper) {
		g_free(m_PhtData.aper);
		m_PhtData.aper = NULL;
	}
	m_PhtData.ap_count = 0;
	m_ApertureIndex = -1;
	m_IgnoreApertures = true;
	m_PhtInfo.Clear();
}

// Clear image data
void CChartBase::ClearImage()
{
	if (m_Image) {
		cairo_surface_destroy(m_Image);
		m_Image = NULL;
	}
	m_ImgWidth = m_ImgHeight = 0;
	if (m_CCDData.ccd) {
		g_free(m_CCDData.ccd);
		m_CCDData.ccd = NULL;
	}
	m_CCDInfo.Clear();
	m_CCDData.scale = m_CCDData.offset = 0.0;
	if (m_pScale)
		m_pScale->Update();
}

// Clear selection
void CChartBase::ClearSelection(void)
{
	m_Selection.Clear();
}

// Load image file
void CChartBase::LoadImage(const char *filename)
{
	bool	ok;
	double	a, val, datamin, datamax, minval, maxval;
	int		w, h, index, i, j, k, pixels, count, *hist;
	CmpackCcdFile *ccd_file;

	ClearImage();

	char *f = g_locale_from_utf8(filename, -1, NULL, NULL, NULL);
	if (cmpack_ccd_open(&ccd_file, f, CMPACK_OPEN_READONLY)==0) {
		// Get image info
		m_CCDInfo.Set(ccd_file);
		// Get image size
		w = h = 0;
		cmpack_ccd_get_size(ccd_file, &w, &h);
		if (w>0 && h>0) {
			m_CCDData.width = w;
			m_CCDData.height = h;
			pixels = w*h;
			m_CCDData.ccd = (double*)g_malloc(pixels*sizeof(double));
			cmpack_ccd_get_image(ccd_file, CMPACK_BITPIX_DOUBLE, 0, 0, m_CCDData.ccd);
			// Autocontrast
			minval = m_CCDData.datamin = CConfig::GetDbl("CCD", "MinValue", 0.0);
			maxval = m_CCDData.datamax = CConfig::GetDbl("CCD", "MaxValue", 65535.0);
			datamin = datamax = 0.0;
			ok = false;
			for (i=0; i<pixels; i++) {
				val = m_CCDData.ccd[i];
				if (val>minval && val<maxval) {
					if (!ok) {
						datamin = datamax = val;
						ok = true;
					} else {
						if (datamin > val)
							datamin = val;
						if (datamax < val)
							datamax = val;
					}
				}
			}
			if (ok && (datamax-datamin)>65536*DBL_EPSILON) {
				a = 65535.0/(datamax-datamin);
				hist = (int*)g_malloc0(65536*sizeof(int));
				for (i=0; i<pixels; i++) {
					val = m_CCDData.ccd[i];
					if (val>minval && val<maxval) {
						val = a*(val-datamin);
						if (val<0)
							index = 0;
						else if (val>65535)
							index = 65535;
						else
							index = (int)(val+0.5);
						hist[index]++;
					}
				}
				for (j=count=0; j<65535 && count<0.01*pixels; j++)
					count += hist[j];
				for (k=65535, count=0; k>j && count<0.01*pixels; k--)
					count += hist[k];
				if (k-j>16) {
					m_CCDData.scale  = -a/(k-j);
					m_CCDData.offset = -m_CCDData.scale*(k/a+datamin);
				} else {
					m_CCDData.scale = 1.0/(datamax-datamin);
					m_CCDData.offset = -m_CCDData.scale*datamin;
				}
				g_free(hist);
			}
		}
		cmpack_unref(ccd_file);
	}
	g_free(f);

	RebuildImage();
	Resize();
	if (m_pScale)
		m_pScale->Update();
	Invalidate();
}

// Load photometry file
void CChartBase::LoadChart(const char *filename, bool ignore_selection)
{
	int i, j, count;
	double stddev;
	CmpackPhtFile *pht_file;
	CmpackCatFile *cat_file;
	
	ClearPhotometry();
	if (!ignore_selection)
		ClearSelection();

	char *f = g_locale_from_utf8(filename, -1, NULL, NULL, NULL);
	if (cmpack_pht_load(&pht_file, f, CMPACK_LOAD_DEFAULT)==0) {
		m_PhtInfo.Set(pht_file);
		LoadPhotometryFile(pht_file);
		cmpack_unref(pht_file);
	} else 
	if (cmpack_cat_load(&cat_file, f, CMPACK_LOAD_DEFAULT)==0) {
		m_PhtInfo.Set(cat_file);
		LoadCatalogueFile(cat_file);
		if (!ignore_selection)
			m_Selection.Load(cat_file);
		cmpack_unref(cat_file);
	}
	g_free(f);

	// Compute reference magnitude
	if (m_PhtData.st_count>0) {
		double *maglist = (double*)g_malloc(m_PhtData.st_count*sizeof(double));
		for (j=0; j<m_PhtData.ap_count; j++) {
			count = 0;
			for (i=0; i<m_PhtData.st_count; i++) {
				if (m_PhtData.stars[i].data[j].mask & CMPACK_MASK_MAG)
					maglist[count++] = m_PhtData.stars[i].data[j].mag;
			}
			cmpack_robustmean(count, maglist, &m_PhtData.aper[j].RefMag, &stddev);
		}
		g_free(maglist);
	} else {
		for (j=0; j<m_PhtData.ap_count; j++)
			m_PhtData.aper[j].RefMag = 0.0;
	}
	
	UpdateTypes();
	Resize();
	Invalidate();
}

// Load photometry file
void CChartBase::LoadPhotometryFile(CmpackPhtFile *pht_file)
{
	int i, j, count, id;
	double radius, x, y;
	CmpackPhtMeasurement mag;
	GSList *stars, *ptr, *apers;

	// Chart size
	cmpack_pht_get_size(pht_file, &m_PhtData.width, &m_PhtData.height);

	// Make list of apertures
	apers = NULL;
	count = cmpack_pht_naper(pht_file);
	for (i=0; i<count; i++) {
		CmpackPhtAperture *ap = cmpack_pht_get_aperture(pht_file, i);
		id = cmpack_pht_aperture_get_id(ap);
		radius = cmpack_pht_aperture_get_radius(ap);
		if (id>0) {
			tAperRec *data = (tAperRec*)g_malloc0(sizeof(tAperRec));
			data->Id = id;
			data->Radius = radius;
			// Add to list
			apers = g_slist_append(apers, data);
		}
		cmpack_unref(ap);
	}
	// Sort by aperture id
	if (count>1)
		apers = g_slist_sort(apers, (GCompareFunc)compare_apers);
	count = g_slist_length(apers);
	if (count>0) {
		m_PhtData.aper = (tAperRec*)g_malloc(count*sizeof(tAperRec));
		for (ptr=apers, i=0; ptr!=NULL; ptr=ptr->next, i++) {
			m_PhtData.aper[i] = *((tAperRec*)ptr->data);
			g_free(ptr->data);
		}
	}
	g_slist_free(apers);
	m_PhtData.ap_count = count;
	if (m_ApertureID>0) {
		m_ApertureIndex = -1;
		for (i=0; i<m_PhtData.ap_count; i++) {
			if (m_PhtData.aper[i].Id == m_ApertureID) {
				m_ApertureIndex = i;
				break;
			}
		}
	} else {
		m_ApertureIndex = (count>0 ? 0 : -1);
	}
	m_IgnoreApertures = false;

	// Make list of stars
	stars = NULL;
	count = cmpack_pht_nstar(pht_file);
	for (i=0; i<count; i++) {
		CmpackPhtStar *st = cmpack_pht_get_star(pht_file, i);
		id = cmpack_pht_star_get_id(st);
		if (id>=0 && cmpack_pht_star_get_center(st, &x, &y)) {
			tStarRec *data = (tStarRec*)g_malloc0(sizeof(tStarRec));
			data->id = id;
			data->x = x;
			data->y = y;
			data->data = (tMagRec*)g_malloc((m_PhtData.ap_count)*sizeof(tMagRec));
			for (j=0; j<m_PhtData.ap_count; j++) {
				// Get magnitude & error
				data->data[j].mask = 0;
				if (cmpack_pht_star_get_measurement(st, m_PhtData.aper[j].Id, &mag)==0) {
					data->data[j].mask = mag.mask & (CMPACK_MASK_MAG | CMPACK_MASK_ERR);
					data->data[j].mag = mag.magnitude;
					data->data[j].err = mag.error;
				}
			}
			// Get reference id
			data->xref = cmpack_pht_star_get_refid(st);
			data->matched = data->xref>=0;
			// Add to list
			stars = g_slist_prepend(stars, data);
		}
		cmpack_unref(st);
	}
	// Sort by magnitude in descending order
	if (count>1)
		stars = g_slist_sort(stars, (GCompareFunc)compare_stars);
	count = g_slist_length(stars);
	if (count>0) {
		m_PhtData.stars = (tStarRec*)g_malloc(count*sizeof(tStarRec));
		for (ptr=stars, i=0; ptr!=NULL; ptr=ptr->next, i++) {
			m_PhtData.stars[i] = *((tStarRec*)ptr->data);
			g_free(ptr->data);
		}
	}
	g_slist_free(stars);
	m_PhtData.st_count = count;
}

// Load photometry file
void CChartBase::LoadCatalogueFile(CmpackCatFile *cat_file)
{
	int i, count, id;
	double x, y;
	GSList *stars, *ptr;

	// Chart size
	cmpack_cat_get_size(cat_file, &m_PhtData.width, &m_PhtData.height);

	// Make list of apertures
	m_PhtData.aper = (tAperRec*)g_malloc(sizeof(tAperRec));
	m_PhtData.aper[0].Id = 1;
	m_PhtData.ap_count = 1;
	m_ApertureID = m_ApertureIndex = 0;
	m_IgnoreApertures = true;

	// Make list of stars
	stars = NULL;
	count = cmpack_cat_nstar(cat_file);
	for (i=0; i<count; i++) {
		CmpackCatStar *st = cmpack_cat_get_star(cat_file, i);
		id = cmpack_cat_star_get_id(st);
		if (id>=0 && cmpack_cat_star_get_center(st, &x, &y)) {
			tStarRec *data = (tStarRec*)g_malloc0(sizeof(tStarRec));
			data->id = data->xref = id;
			data->x = x;
			data->y = y;
			data->matched = true;
			data->data = (tMagRec*)g_malloc(sizeof(tMagRec));
			if (cmpack_cat_star_get_magnitude(st, &data->data[0].mag)) 
				data->data[0].mask = CMPACK_MASK_MAG;
			else
				data->data[0].mask = 0;
			// Add to list
			stars = g_slist_prepend(stars, data);
		}
		cmpack_unref(st);
	}
	// Sort by magnitude in descending order
	if (count>1)
		stars = g_slist_sort(stars, (GCompareFunc)compare_stars);
	count = g_slist_length(stars);
	if (count>0) {
		m_PhtData.stars = (tStarRec*)g_malloc(count*sizeof(tStarRec));
		for (ptr=stars, i=0; ptr!=NULL; ptr=ptr->next, i++) {
			m_PhtData.stars[i] = *((tStarRec*)ptr->data);
			g_free(ptr->data);
		}
	}
	g_slist_free(stars);
	m_PhtData.st_count = count;
}

/* Create a new backing pixmap of the appropriate size */
gint CChartBase::configure_event(GtkWidget *widget, GdkEventConfigure *event, CChartBase *pMe)
{
	pMe->OnConfigure(widget, event);
	return TRUE;
}

void CChartBase::OnConfigure(GtkWidget *widget, GdkEventConfigure *event)
{
	int width, height;

	width = widget->allocation.width;
	height = widget->allocation.height;
	if (width!=m_PixWidth || height!=m_PixHeight) {
		if (m_Pixmap) {
			g_object_unref(m_Pixmap);
			m_Pixmap = NULL;
		}
		if (width>0 && height>0) {
			m_Pixmap = gdk_pixmap_new(widget->window, width, height, -1);
			m_PixWidth = width;
			m_PixHeight = height;
		} else {
			m_PixHeight = m_PixWidth = 0;
		}
		Invalidate();
	}
}

/* Show/hide image */
void CChartBase::ShowImage(bool show)
{
	m_ShowImage = show;
	Invalidate();
}

/* Set pseudocolors mode */
void CChartBase::SetPseudocolors(bool enable)
{
	m_Pseudocolors = enable;
	if (m_pScale)
		m_pScale->Update();
	RebuildImage();
	Invalidate();
}

/* Set bad pixels mode */
void CChartBase::SetBadPixels(bool enable)
{
	m_BadPixels = enable;
	if (m_pScale)
		m_pScale->Update();
	RebuildImage();
	Invalidate();
}

/* Show chart */
void CChartBase::ShowChart(bool show)
{
	m_ShowChart = show;
	Invalidate();
}

/* Show matched stars only */
void CChartBase::ShowBadStars(bool show)
{
	m_BadStars = show;
	Invalidate();
}

/* Show matched stars only */
void CChartBase::ShowNonMatchedStars(bool show)
{
	m_ShowNonMatchedStars = show;
	Invalidate();
}

/* Show matched stars only */
void CChartBase::HighlightNonMatchedStars(bool show)
{
	m_HighlightNonMatchedStars = show;
	Invalidate();
}

/* Show matched stars only */
void CChartBase::HighlightSelectedStars(bool show)
{
	m_HighlightSelectedStars = show;
	Invalidate();
}

/* Set active aperture */
void CChartBase::SetAperture(int aperture)
{
	int i;

	if (aperture!=m_ApertureID && !m_IgnoreApertures) {
		m_ApertureID = aperture;
		if (m_ApertureID>0) {
			m_ApertureIndex = -1;
			for (i=0; i<m_PhtData.ap_count; i++) {
				if (m_PhtData.aper[i].Id == m_ApertureID) {
					m_ApertureIndex = i;
					break;
				}
			}
		} else {
			m_ApertureIndex = (m_PhtData.ap_count>0 ? 0 : -1);
		}
		Invalidate();
	}
}

/* Redraw the screen from the backing pixmap */
gint CChartBase::expose_event(GtkWidget *widget, GdkEventExpose *event, CChartBase *pMe)
{
	pMe->OnExposeEvent(widget, event);
	return FALSE;
}

void CChartBase::OnExposeEvent(GtkWidget *widget, GdkEventExpose *event)
{
	int r, x, y;
	GdkGCValues values;

	if (m_NeedUpdate) {
		m_NeedUpdate = false;
		Redraw();
	}

	gdk_draw_drawable(widget->window, widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		m_Pixmap, event->area.x, event->area.y, event->area.x, event->area.y,
		event->area.width, event->area.height);

	if (m_StFocused>=0 && m_ApertureIndex>=0) {
		tStarRec *pStar = &m_PhtData.stars[m_StFocused];
		x = (int)(XToView(pStar->x)+0.5);
		y = (int)(YToView(pStar->y)+0.5);
		r = RoundToInt(ComputeRadius(pStar, m_Zoom));
		gdk_gc_get_values(widget->style->black_gc, &values);
		gdk_gc_set_function(widget->style->black_gc, GDK_INVERT);
		gdk_draw_line(GDK_DRAWABLE(widget->window), widget->style->black_gc,
			x - r - CURSOR_SIZE, y, x - r - 1, y);
		gdk_draw_line(GDK_DRAWABLE(widget->window), widget->style->black_gc,
			x + r + CURSOR_SIZE, y, x + r + 1, y);
		gdk_draw_line(GDK_DRAWABLE(widget->window), widget->style->black_gc,
			x, y - r - CURSOR_SIZE, x, y - r - 1);
		gdk_draw_line(GDK_DRAWABLE(widget->window), widget->style->black_gc,
			x, y + r + CURSOR_SIZE, x, y + r + 1);
		gdk_gc_set_function(widget->style->black_gc, values.function);
	}
}


//
// Draw a rectangle on the screen 
//
void CChartBase::Redraw(void)
{
	if (m_Pixmap && m_PixWidth>0 && m_PixHeight>0) {
		cairo_t *cr = gdk_cairo_create(m_Pixmap);
		Draw(cr, m_Zoom, m_OffsetX, m_OffsetY);
		cairo_destroy(cr);
	}
}


//
// Draw to buffer
//
void CChartBase::Draw(cairo_t *cr, double zoom, double offsetx, double offsety) const
{
	bool invert, have_image, have_chart;
	int i, width, height;

	// Draw background
	cairo_set_source_rgb(cr, 0.5, 0.5, 0.5);
	cairo_paint(cr);
	if (m_MapWidth>0 && m_MapHeight>0) {
		have_image = m_Image && m_ImgWidth>0 && m_ImgHeight>0;
		have_chart = m_PhtData.width>0 && m_PhtData.height>0 && m_ApertureIndex>=0;
		if (have_image && (!have_chart || m_ShowImage || !m_ShowChart)) {
			// Draw image
			cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
			cairo_identity_matrix(cr);
			cairo_pattern_t *pat = cairo_pattern_create_for_surface(m_Image);
			cairo_pattern_set_filter(pat, (zoom<4.0 ? CAIRO_FILTER_GOOD : CAIRO_FILTER_NEAREST));
			cairo_translate(cr, offsetx, offsety);
			cairo_scale(cr, zoom, zoom);
			cairo_set_source(cr, pat);
			cairo_paint(cr);
			cairo_pattern_destroy(pat);
		} else {
			// Draw clear frame 
			cairo_identity_matrix(cr);
			width = (int)(zoom*m_MapWidth+0.5);
			height = (int)(zoom*m_MapHeight+0.5);
			cairo_rectangle(cr, offsetx, offsety, width, height);
			cairo_set_source_rgb(cr, 0.2, 0.0, 0.4);
			cairo_fill(cr);
		}
		if (have_chart) {
			width = (int)(zoom*m_PhtData.width+0.5);
			height = (int)(zoom*m_PhtData.height+0.5);
			invert = m_Invert && m_ShowImage && m_Image;
			cairo_identity_matrix(cr);
			cairo_rectangle(cr, offsetx, offsety, width, height);			
			cairo_clip(cr);
			cairo_set_line_width(cr, 1);
			cairo_set_antialias(cr, CAIRO_ANTIALIAS_DEFAULT);
			if (!have_image || !m_ShowImage || m_ShowChart) {
				// Draw all stars
				for (i=0; i<m_PhtData.st_count; i++) {
					DrawStar(cr, zoom, offsetx, offsety, &m_PhtData.stars[i], i==m_StSelected, invert);
				}
			} else if (m_HighlightSelectedStars) {
				// Draw selected stars only
				for (i=0; i<m_PhtData.st_count; i++) {
					if (m_PhtData.stars[i].stype!=0 && i!=m_StSelected) 
						DrawStar(cr, zoom, offsetx, offsety, &m_PhtData.stars[i], i==m_StSelected, invert);
				}
				if (m_StSelected>=0) {
					DrawStar(cr, zoom, offsetx, offsety, &m_PhtData.stars[m_StSelected], true, invert);
				}
			}
			// Draw labels
			cairo_set_font_size(cr, 12.0);
			cairo_select_font_face(cr, "", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD);
			cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
			for (i=0; i<m_PhtData.st_count; i++) {
				if (m_PhtData.stars[i].stype!=0 && i!=m_StSelected) 
					DrawLabel(cr, zoom, offsetx, offsety, &m_PhtData.stars[i], i==m_StSelected, invert);
			}
			if (m_StSelected>=0) {
				DrawLabel(cr, zoom, offsetx, offsety, &m_PhtData.stars[m_StSelected], true, invert);
			}
		}
	}
}

/* Draw a rectangle to a context */
void CChartBase::Export(cairo_surface_t *surface, int width, int height) const
{
	int		w, h;
	double	kx, ky, k, dx, dy;
	
	cairo_t *cr = cairo_create(surface);
	if (!cr)
		return;

	// Clear background
	cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
	cairo_set_source_rgb(cr, 0.5, 0.5, 0.5);
	cairo_rectangle(cr, 0, 0, width, height);
	cairo_fill(cr);

	if (m_MapWidth>0 && m_MapHeight>0) {
		// Compute mapping coefficients
		kx = (double)width/m_MapWidth;
		ky = (double)height/m_MapHeight;
		k = (kx < ky ? kx : ky);
		w = (int)ceil(k*m_MapWidth);
		h = (int)ceil(k*m_MapHeight);
		dx = (width - w)/2;
		dy = (height - h)/2;
		Draw(cr, k, dx, dy);
	}

	cairo_destroy(cr);
}

// Find aperture by identifier
int CChartBase::FindApIndex(int id) const
{
	int i;

	for (i=0; i<m_PhtData.ap_count; i++) {
		if (m_PhtData.aper[i].Id == id)
			return i;
	}
	return -1;
}

// Find star by identifier
int CChartBase::FindStarById(int id)
{
	int i;

	for (i=0; i<m_PhtData.st_count; i++) {
		if (m_PhtData.stars[i].id == id)
			return i;
	}
	return -1;
}

// Find star by identifier
int CChartBase::FindStarByRefId(int xref)
{
	int i;

	for (i=0; i<m_PhtData.st_count; i++) {
		if (m_PhtData.stars[i].matched && m_PhtData.stars[i].xref == xref)
			return i;
	}
	return -1;
}

// Find star by position
int CChartBase::FindStarByPos(int x, int y, bool visible_only)
{
	bool	bad;
	int		i, j;
	double	dx, dy, r2, min_r2;

	if (m_ApertureIndex>=0) {
		j = -1;
		min_r2 = 100/(m_Zoom*m_Zoom);
		for (i=0; i<m_PhtData.st_count; i++) {
			tStarRec *star = &m_PhtData.stars[i];
			bad = (star->data[m_ApertureIndex].mask & CMPACK_MASK_MAG)==0;
			if (!visible_only || ((m_BadStars || !bad) && (m_ShowNonMatchedStars || star->matched))) {
				dx = (star->x - x);
				dy = (star->y - y);
				r2 = (dx*dx) + (dy*dy);
				if (r2<min_r2) {
					j = i;
					min_r2 = r2;
				}
			}
		}
		return j;
	}
	return -1;
}

void CChartBase::MouseMove(GdkModifierType state, int x, int y)
{
	if (m_EnableTrack && x!=m_CursorX || y!=m_CursorY) {
		m_CursorX = x;
		m_CursorY = y;
		Callback(CB_MOUSE_MOVE);
	}
}

void CChartBase::SetFocusedStar(int index)
{
	int		r, x, y;

	if (index!=m_StFocused) {
		if (m_StFocused>=0) {
			tStarRec *stold = &m_PhtData.stars[m_StFocused];
			x = RoundToInt(XToView(stold->x));
			y = RoundToInt(YToView(stold->y));
			r = (int)ceil(ComputeRadius(stold, m_Zoom));
			gtk_widget_queue_draw_area(m_DrawArea, x - CURSOR_SIZE - r, y - CURSOR_SIZE - r, 
				2*(CURSOR_SIZE + r) + 2, 2*(CURSOR_SIZE + r) + 2);
		}
		m_StFocused = index;
		if (m_StFocused>=0) {
			tStarRec *stnew = &m_PhtData.stars[m_StFocused];
			x = RoundToInt(XToView(stnew->x));
			y = RoundToInt(YToView(stnew->y));
			r = (int)ceil(ComputeRadius(stnew, m_Zoom));
			gtk_widget_queue_draw_area(m_DrawArea, x - CURSOR_SIZE - r, y - CURSOR_SIZE - r, 
				2*(CURSOR_SIZE + r) + 2, 2*(CURSOR_SIZE + r) + 2);
		}
	}
}

void CChartBase::SelectStar(int index)
{
	if (index!=m_StSelected) {
		m_StSelected = index;
		Invalidate();
	}
}

gint CChartBase::button_press_event(GtkWidget *widget, GdkEventButton *event, CChartBase *pMe)
{
	pMe->OnButtonPress(widget, event);
	return true;
}

void CChartBase::OnButtonPress(GtkWidget *widget, GdkEventButton *event)
{
	switch (event->type)
	{
	case GDK_BUTTON_PRESS:
		switch (event->button)
		{
		case 1:
			MouseMove((GdkModifierType)event->state, (int)event->x, (int)event->y);
			Callback(CB_LEFT_CLICK, m_StFocused, event);
			break;
		case 3:
			MouseMove((GdkModifierType)event->state, (int)event->x, (int)event->y);
			Callback(CB_RIGHT_CLICK, m_StFocused, event);
			break;
		}
		break;
	
	case GDK_2BUTTON_PRESS:
		switch (event->button)
		{
		case 1:
			MouseMove((GdkModifierType)event->state, (int)event->x, (int)event->y);
			Callback(CB_LEFT_DBLCLICK, m_StFocused, event);
			break;
		}
		break;
		
	default:
		break;
	}
}

gint CChartBase::motion_notify_event(GtkWidget *widget, GdkEventMotion *event, CChartBase *pMe)
{
	pMe->OnMotionNotify(widget, event);
	return true;
}

void CChartBase::OnMotionNotify(GtkWidget *widget, GdkEventMotion *event)
{
	gint x, y;
	GdkModifierType state;

	if (m_EnableTrack) {
		if (event->is_hint) {
			gdk_window_get_pointer(event->window, &x, &y, &state);
		} else {
			x = (int)event->x;
			y = (int)event->y;
			state = (GdkModifierType)event->state;
		}
		MouseMove(state, x, y);
	}
}

gint CChartBase::leave_notify_event(GtkWidget *widget, GdkEventCrossing *event, CChartBase *pMe)
{
	pMe->OnLeaveNotify(widget, event);
	return true;
}

void CChartBase::OnLeaveNotify(GtkWidget *widget, GdkEventCrossing *event)
{
	if (m_EnableTrack && (m_CursorX>=0 || m_CursorY>=0)) {
		m_CursorX = m_CursorY = -1;
		Callback(CB_MOUSE_MOVE);
	}
}

void CChartBase::DrawStar(cairo_t *cr, double zoom, double offsetx, double offsety, 
		const tStarRec *star, bool selected, bool invert) const
{
	double r, x, y;
	bool bad;
	tStarType type;

	bad = (star->data[m_ApertureIndex].mask & CMPACK_MASK_MAG)==0;
	if ((!m_BadStars && bad) || (!m_ShowNonMatchedStars && !star->matched))
		return;

	x = star->x*zoom + offsetx;
	y = star->y*zoom + offsety;
	r = ComputeRadius(star, zoom);
	type = (m_HighlightSelectedStars ? star->stype : ST_NOT_SELECTED);

	cairo_new_sub_path(cr);
	cairo_arc(cr, x, y, r, 0.0, 2.0*M_PI);
	if (star->data[m_ApertureIndex].mask & CMPACK_MASK_MAG) {
		if (selected)
			cairo_set_source_rgb(cr, 0.7, 0.7, 0);
		else if (type == ST_VARIABLE)
			cairo_set_source_rgb(cr, 0.7, 0, 0);
		else if (type == ST_COMPARISON)
			cairo_set_source_rgb(cr, 0, 0.5, 0);
		else if (type == ST_CHECK)
			cairo_set_source_rgb(cr, 0, 0, 0.9);
		else if (!invert) {
			if (m_HighlightNonMatchedStars && !star->matched)
				cairo_set_source_rgb(cr, 0.6, 0.6, 0.6);
			else
				cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
		} else {
			if (m_HighlightNonMatchedStars && !star->matched)
				cairo_set_source_rgb(cr, 0.4, 0.4, 0.4);
			else
				cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
		}
		cairo_fill_preserve(cr);
		if (invert) 
			cairo_set_source_rgb (cr, 1, 1, 1);
		else
			cairo_set_source_rgb (cr, 0, 0, 0);
		cairo_stroke(cr);
	} else {
		if (selected)
			cairo_set_source_rgb(cr, 0.7, 0.7, 0);
		else if (type == ST_VARIABLE)
			cairo_set_source_rgb(cr, 0.7, 0, 0);
		else if (type == ST_COMPARISON)
			cairo_set_source_rgb(cr, 0, 0.5, 0);
		else if (type == ST_CHECK)
			cairo_set_source_rgb(cr, 0, 0, 0.9);
		else if (m_HighlightNonMatchedStars && !star->matched) {
			if (!invert)
				cairo_set_source_rgb(cr, 0.6, 0.6, 0.6);
			else
				cairo_set_source_rgb(cr, 0.4, 0.4, 0.4);
		} else
			cairo_set_source_rgb(cr, 1.0, 0, 0);
		cairo_stroke(cr);
	}
}

void CChartBase::DrawLabel(cairo_t *cr, double zoom, double offsetx, double offsety, 
		const tStarRec *star, bool selected, bool invert) const
{
	double r;
	int i, x, y, index;
	char buf[64];
	cairo_text_extents_t extents;
	tStarType type;
	
	if (star->matched && star->stype!=ST_NOT_SELECTED && m_HighlightSelectedStars) {
		i = m_Selection.FindStar(star->xref);
		if (i>=0) {
			type = m_Selection.GetType(i);
			index = m_Selection.GetIndex(i);
			// Position
			cairo_text_extents(cr, "X", &extents);
			r = ComputeRadius(star, zoom);
			x = RoundToInt(star->x*zoom + offsetx + 0.71*r + 2);	
			y = RoundToInt(star->y*zoom + offsety + 0.71*r + 2 + extents.height);
			// Caption
			if (type==ST_VARIABLE) {
				if (index==1)
					strcpy(buf, "var");
				else
					sprintf(buf, "var #%d", index);
			} else 
			if (type==ST_COMPARISON) {
				if (index==1)
					strcpy(buf, "comp");
				else
					sprintf(buf, "comp #%d", index);
			} else {
				sprintf(buf, "check #%d", index);
			}
			// Color
			if (type==ST_VARIABLE)
				cairo_set_source_rgb(cr, 1, 0, 0);
			else if (type==ST_COMPARISON)
				cairo_set_source_rgb(cr, 0, 1, 0);
			else
				cairo_set_source_rgb(cr, 0, 0, 1);
			/*// Frame
			cairo_rectangle(cr, x - 3, y - extents.height - 3, extents.width + 8, extents.height + 6);
			cairo_fill_preserve(cr);
			cairo_set_source_rgb(cr, 0, 0, 0);
			cairo_stroke(cr);
			// Text
			cairo_set_source_rgb(cr, 0, 0, 0);*/
			cairo_move_to(cr, x, y);
			cairo_show_text(cr, buf);
		}
	}
	if (selected) {
		// Caption
		sprintf(buf, "%d", star->id);
		// Position
		r = ComputeRadius(star, zoom);
		x = RoundToInt(star->x*zoom + offsetx + 0.71*r + 2);
		y = RoundToInt(star->y*zoom + offsety - 0.71*r + 2);
		/*// Frame
		cairo_set_source_rgb(cr, 1, 1, 0.7);
		cairo_rectangle(cr, x - 3, y - extents.height - 3, extents.width + 8, extents.height + 6);
		cairo_fill_preserve(cr);
		cairo_set_source_rgb(cr, 0, 0, 0);
		cairo_stroke(cr);
		// Text
		cairo_set_source_rgb(cr, 0, 0, 0);*/
		cairo_move_to(cr, x, y);
		cairo_show_text(cr, buf);
	}
}

// Compare stars by magnitude
gint CChartBase::compare_stars(const tStarRec *a, const tStarRec *b)
{
	if ((a->data[0].mask & CMPACK_MASK_MAG)==0) 
		return ((b->data[0].mask & CMPACK_MASK_MAG)==0 ? 0 : -1);
	else if ((b->data[0].mask & CMPACK_MASK_MAG)==0)
		return 1;
	else if (a->data[0].mag < b->data[0].mag)
		return -1;
	else if (a->data[0].mag > b->data[0].mag)
		return +1;
	else
		return 0;
}

// Compare apertures by id
gint CChartBase::compare_apers(const tAperRec *a, const tAperRec *b)
{
	if (a->Id < b->Id)
		return -1;
	else if (a->Id > b->Id)
		return +1;
	else
		return 0;
}

void CChartBase::EnableTracking(bool enable)
{
	if (enable!=m_EnableTrack) {
		m_EnableTrack = enable;
		m_StFocused = -1;
	}
}

// Compute circle radius for given star
double CChartBase::ComputeRadius(const struct tStarRec *star, double zoom) const
{
	double radius;

	if (m_ApertureIndex>=0 && (star->data[m_ApertureIndex].mask & CMPACK_MASK_MAG))
		radius = 3.0 + 0.75*(m_PhtData.aper[m_ApertureIndex].RefMag - star->data[m_ApertureIndex].mag);
	else
		radius = 3.0;
	return LimitValue(radius, 2.0, 12.0)*sqrt(zoom);
}

// Position transformation
double CChartBase::XToView(double x) const
{
	return x*m_Zoom + m_OffsetX;
}
double CChartBase::YToView(double y) const
{
	return y*m_Zoom + m_OffsetY;
}

// Reverse transformation
double CChartBase::ViewToX(double u) const
{
	return (u-m_OffsetX)/m_Zoom;
}
double CChartBase::ViewToY(double v) const
{
	return (v-m_OffsetY)/m_Zoom;
}


/* Rebuild image */
void CChartBase::RebuildImage()
{
	bool ok;
	int i, width, height, pixels;
	double scale, offset;

	m_Invert = CConfig::GetBool("Environment", "ImageInvert");

	width = m_CCDInfo.Width();
	height = m_CCDInfo.Height();
	if (m_ImgHeight!=height || m_ImgWidth!=width) {
		// Rebuild image surface
		if (m_Image) {
			cairo_surface_destroy(m_Image);
			m_Image = NULL;
		}
		m_ImgWidth = width;
		m_ImgHeight = height;
		if (m_ImgWidth>0 && m_ImgHeight>0) {
			m_Image = cairo_image_surface_create(CAIRO_FORMAT_RGB24, m_ImgWidth, m_ImgHeight);
		}
	}

	if (m_Image && m_ImgWidth>0 && m_ImgHeight>0) {
		// Refresh image
		unsigned long *dptr = (unsigned long*)cairo_image_surface_get_data(m_Image);
		if (m_Invert) {
			scale = m_CCDData.scale;
			offset = m_CCDData.offset;
		} else {
			scale = -m_CCDData.scale;
			offset = 1.0-m_CCDData.offset;
		}
		pixels = m_ImgWidth*m_ImgHeight;
		const double *sptr = m_CCDData.ccd;
		if (!m_Pseudocolors) {
			for (i=0; i<pixels; i++) {
				double val = *sptr++;
				if (m_BadPixels) 
					ok = (val>m_CCDData.datamin && val<m_CCDData.datamax);
				else
					ok = true;
				if (ok)
					*dptr++ = grayscale(scale*val + offset);
				else
					*dptr++ = BAD_PIXEL_COLOR;
			}
		} else {
			for (i=0; i<pixels; i++) {
				double val = *sptr++;
				if (m_BadPixels) 
					ok = (val>m_CCDData.datamin && val<m_CCDData.datamax);
				else
					ok = true;
				if (ok) 
					*dptr++ = pseudocolor(scale*val + offset);
				else
					*dptr++ = BAD_PIXEL_COLOR;
			}
		}
	}
}

// Get scale min and max 
void CChartBase::GetScaleMinMax(double *min, double *max)
{
	if (m_CCDData.scale!=0.0) {
		if (min)
			*min = (1.0 - m_CCDData.offset)/m_CCDData.scale;
		if (max)
			*max = (0.0 - m_CCDData.offset)/m_CCDData.scale;
	} else {
		if (min)
			*min = 0.0;
		if (max)
			*max = 0.0;
	}
}

bool CChartBase::GetCursorPos(int *px, int *py) const
{
	int x, y;

	if (m_EnableTrack && m_CursorX>=0 && m_CursorY>=0) {
		x = (int)ViewToX(m_CursorX);
		y = (int)ViewToY(m_CursorY);
		if (x>=0 && x<m_MapWidth && y>=0 && y<m_MapHeight) {
			if (px) *px = x;
			if (py) *py = y;
			return true;
		}
	}
	return false;
}

// Get pixel value
double CChartBase::GetValue(int x, int y) const
{
	if (x>=0 && x<m_CCDData.width && y>=0 && y<m_CCDData.height) 
		return m_CCDData.ccd[x+y*m_CCDData.width];
	else
		return 0;
}

// Get star data
bool CChartBase::GetStar(int index, tStarInfo *info) const
{
	memset(info, 0, sizeof(tStarInfo));
	if (index>=0 && index<m_PhtData.st_count) {
		tStarRec *rec = &m_PhtData.stars[index];
		info->id		= rec->id;
		info->x			= rec->x;
		info->y			= rec->y;
		info->matched	= rec->matched;
		info->xref		= rec->xref;
		if (rec->matched && rec->stype!=ST_NOT_SELECTED) {
			int i = m_Selection.FindStar(rec->xref);
			if (i>=0) {
				info->sel_type = m_Selection.GetType(i);
				info->sel_index = m_Selection.GetIndex(i);
			}
		}
		if (m_ApertureIndex>=0) {
			tMagRec *mag = &rec->data[m_ApertureIndex];
			info->mag_valid	= (mag->mask & CMPACK_MASK_MAG)!=0;
			info->mag		= mag->mag;
			info->err_valid	= (mag->mask & CMPACK_MASK_ERR)!=0;
			info->err		= mag->err;
		}
		return true;
	}
	return false;
}


//
// Quick photometry
//
bool CChartBase::Photometry(int x, int y, double aperture, tPhotResult *result)
{
	double skyin, skyout;

	if (x>=0 && y>=0 && x<m_CCDData.width && y<m_CCDData.height && aperture>0) {
		skyin = CConfig::GetDbl("Photometry", "SkyInRadius", 20.0);
		skyout = CConfig::GetDbl("Photometry", "SkyOutRadius", 30.0);
		if (skyin>0 && skyout>skyin) {
			return QuickPhotometry(m_CCDData.ccd, m_CCDData.width, m_CCDData.height, 
				m_CCDData.datamin, m_CCDData.datamax, aperture, skyin, skyout, x, y, result);
		}
	}
	return false;
}


//
// Set selection
//
void CChartBase::SetSelection(const CSelection *sel)
{
	m_Selection = *sel;
	UpdateTypes();
	Invalidate();
	Invalidate();
}


//
// Selection callback
//
void CChartBase::SelectionCallback(int message, int wparam, void* lparam, void* cb_data)
{
	CChartBase *pMe = (CChartBase*)cb_data;

	switch (message)
	{
	case CSelection::CB_SELECT_CHANGE:
		pMe->UpdateTypes();
		pMe->Invalidate();
		break;
	}
}


//
// Update star types
//
void CChartBase::UpdateTypes(void)
{
	int i, j;

	for (i=0; i<m_PhtData.st_count; i++) {
		tStarRec *star = &m_PhtData.stars[i];
		star->stype = ST_NOT_SELECTED;
		if (star->matched) {
			j = m_Selection.FindStar(star->xref);
			if (j>=0)
				star->stype = m_Selection.GetType(j);
		}
	}
}

//--------------------------   CHART CLASS   ----------------------------------

// Default constructor
CChart::CChart()
{
	m_Selection.RegisterCallback(SelectionCallback, this);
}

// Auto zoom
void CChart::Redraw(void)
{
	double kx, ky;
	int width, height;

	if (m_PixWidth>0 && m_PixHeight>0 && m_MapWidth>0 && m_MapHeight>0) {
		kx = (double)m_PixWidth/m_MapWidth;
		ky = (double)m_PixHeight/m_MapHeight;
		m_Zoom = (kx < ky ? kx : ky);
		width = (int)ceil(m_Zoom*m_MapWidth);
		height = (int)ceil(m_Zoom*m_MapHeight);
		m_OffsetX = (m_PixWidth - width)/2;
		m_OffsetY = (m_PixHeight - height)/2;
		CChartBase::Redraw();
	}
}
