/**************************************************************

config_dlg.cpp (C-Munipack project)
Configuration dialog
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <cmunipack.h>

#include "utils.h"
#include "configuration.h"
#include "config_dlg.h"
#include "main.h"
#include "project.h"

// Model columns
#define NUM_MCOLS 3

enum tTreeColumn {
	MCOL_ID,				// Section id
	MCOL_CAPTION,			// Caption
	MCOL_PARENT				// Parent id
};

// Section identifiers
enum tSectionId {
	SEC_ROOT = 0,
	SEC_CCD,
	SEC_CCD_ADVANCED,
	SEC_ENVIRONMENT,
	SEC_ENV_ADVANCED,
	SEC_INPUT_FILES,
	SEC_REDUCTION,
	SEC_CONVERSION,
	SEC_CALIBRATION,
	SEC_PHOTOMETRY,
	SEC_STAR_DETECTION,
	SEC_DET_ADVANCED,
	SEC_SKY_ANNULUS,
	SEC_APERTURES,
	SEC_PLOTTING,
	SEC_LIGHT_CURVE,
	SEC_OBJECT,
	SEC_OBSERVER,
	SEC_TOOLS,
	SEC_MASTER_BIAS,
	SEC_MASTER_DARK,
	SEC_MASTER_FLAT,
	SEC_MUNIFIND,
	SEC_MATCHING
};
	
// Section definitions
const static struct tSection {
	int			id;			// Section identifier (0 = list terminator)
	const char	*caption;	// Section display name
	int			parent;		// Parent section (0 = first level sections)
} SecList[] = {
	{ SEC_ENVIRONMENT,		"Environment" },
	{ SEC_INPUT_FILES,		"Input files",			SEC_ENVIRONMENT },
	{ SEC_ENV_ADVANCED,		"Advanced settings",	SEC_ENVIRONMENT },
	{ SEC_CCD,				"Source frames" },
	{ SEC_CCD_ADVANCED,		"Advanced settings",	SEC_CCD },
	{ SEC_REDUCTION,		"Reduction" },
	{ SEC_CONVERSION,		"Fetch/convert",		SEC_REDUCTION },
	{ SEC_CALIBRATION,		"Calibration",			SEC_REDUCTION },
	{ SEC_STAR_DETECTION,	"Star detection",		SEC_REDUCTION },
	{ SEC_DET_ADVANCED,		"Advanced settings",	SEC_STAR_DETECTION },
	{ SEC_PHOTOMETRY,		"Photometry",			SEC_REDUCTION },
	{ SEC_APERTURES,		"Apertures",			SEC_PHOTOMETRY },
	{ SEC_SKY_ANNULUS,		"Sky annulus",			SEC_PHOTOMETRY },
	{ SEC_MATCHING,			"Matching",				SEC_REDUCTION },
	{ SEC_PLOTTING,			"Plotting" },
	{ SEC_LIGHT_CURVE,		"Light curve",			SEC_PLOTTING },
	{ SEC_OBJECT,			"Catalogues",			SEC_PLOTTING },
	{ SEC_OBSERVER,			"Observer",				SEC_PLOTTING },
	{ SEC_TOOLS,			"Tools" },
	{ SEC_MASTER_BIAS,		"Master-bias",			SEC_TOOLS },
	{ SEC_MASTER_DARK,		"Master-dark",			SEC_TOOLS },
	{ SEC_MASTER_FLAT,		"Master-flat",			SEC_TOOLS },
	{ SEC_MUNIFIND,			"Find variables",		SEC_TOOLS },
	{ 0, NULL }
};

// Parameter control types
enum tCtrlType {
	CTRL_CHECK,			// Check box
	CTRL_RANGE,			// Edit box with spins
	CTRL_EDIT,			// Edit box
	CTRL_CHECK_GROUP,	// Group of toggle buttons
	CTRL_RADIO_GROUP,	// Group of radio buttons
	CTRL_SELECT,		// Combo box
	CTRL_FILE_CHOOSER 	// File chooser
};

// Parameter data types
enum tDataType {
	DATA_INT,
	DATA_BOOL,
	DATA_DBL,
	DATA_STR
};

union tValue
{
	int			Int;
	bool		Bool;
	char		*Str;
	double		Dbl;
};

struct tStringList
{
	const char *str;			// Caption (NULL = terminator)
	int id;						// Value, bitmask
};

// Options for image rotation
static const tStringList BitpixStrings[] = {
	{ "Autodetection",					CMPACK_BITPIX_AUTO },
	{ "Signed short (2 bytes)",			CMPACK_BITPIX_SSHORT },
	{ "Unsigned short (2 bytes)",		CMPACK_BITPIX_USHORT },
	{ "Signed long (4 bytes)",			CMPACK_BITPIX_SLONG },
	{ "Unsigned long (4 bytes)",		CMPACK_BITPIX_ULONG },
	{ "Single precision FP (4 bytes)",	CMPACK_BITPIX_FLOAT },
	{ "Double precision FP (8 bytes)",	CMPACK_BITPIX_DOUBLE },
	{ NULL }
};

// Available columns
static const tStringList ColStrings[] = {
	{ "Date and time",		1 << 0 },
	{ "Julian date",		1 << 1 },
	{ "Exposure",			1 << 2 },
	{ "Temperature",		1 << 3 },
	{ "Filter",				1 << 4 },
	{ "Stars",				1 << 5 },
	{ "Original file",		1 << 6 },
	{ "Temporary file",		1 << 7 },
	{ NULL }
};

// Icon sizes
static const tStringList IconSizeStrings[] = {
	{ "80 x 80 pixels",		 80 },
	{ "100 x 100 pixels",	100 },
	{ "120 x 120 pixels",	120 },
	{ "140 x 140 pixels",	140 },
	{ "160 x 160 pixels",	160 },
	{ NULL }
};

// Matching algorithm
static const tStringList MatchStrings[] = {
	{ "Never",				0 },
	{ "Auto",				1 },
	{ "Always",				2 },
	{ NULL }
};

// Time of observation
static const tStringList JDFormat[] = {
	{ "Geocentric julian date",		0 },
	{ "Heliocentric julian date",	1 },
	{ NULL }
};

// Optional columns
static const tStringList OptColumns[] = {
	{ "Heliocentric correction",	(1<<0) },
	{ "Air-mass coefficient",		(1<<1) },
	{ "Altitude",					(1<<2) },
	{ NULL }
};

// Parameter definitions
static const struct tParameter 
{
	const char	*group;			// Parameter group (NULL = list terminator)	
	const char	*key;			// Parameter key
	tDataType	data;			// Parameter value type
	const char  *defval;		// Default value
	tSectionId	section;		// Section id
	tCtrlType	ctrl;			// Control type
	const char	*caption;		// Displayed parameter name
	const char  *unit;			// Displayed unit
	double		minval;			// Min value (range controls)
	double		maxval;			// Max value (range controls)
	double		step;			// Step increment (range controls)
	int			digits;			// Decimal places (range controls), max. text length (edit controls)
	const tStringList *strings;	// List of strings (check group, radio group, select)
} ParList[] = {
	{ "Environment", "NoPreviews", DATA_BOOL, "0", SEC_ENVIRONMENT, CTRL_CHECK, "Disable previews" },
	{ "Environment", "ImageInvert", DATA_BOOL, "0", SEC_ENVIRONMENT, CTRL_CHECK, "Invert images" },
	{ "Environment", "Debug", DATA_INT, "0", SEC_ENV_ADVANCED, CTRL_CHECK, "Debug outputs" },
	{ "Environment", "Columns", DATA_INT, "53", SEC_INPUT_FILES, CTRL_CHECK_GROUP, "Displayed columns", NULL, 0, 0, 0, 0, ColStrings },

	{ "CCD", "MinValue", DATA_DBL, "0.0", SEC_CCD, CTRL_RANGE, "Min. pixel value", "ADU", -0x7FFFFFFF, 0x7FFFFFFF, 1.0, 1 },
	{ "CCD", "MaxValue", DATA_DBL, "65535.0", SEC_CCD, CTRL_RANGE, "Max. pixel value", "ADU", -0x7FFFFFFF, 0x7FFFFFFF, 1.0, 1 },
	{ "CCD", "BorderLeft", DATA_INT, "0", SEC_CCD_ADVANCED, CTRL_RANGE, "Border - left", "pxl", 0.0, 65535.0, 1.0, 0 },
	{ "CCD", "BorderRight", DATA_INT, "0", SEC_CCD_ADVANCED, CTRL_RANGE, "Border - right", "pxl", 0.0, 65535.0, 1.0, 0 },
	{ "CCD", "BorderTop", DATA_INT, "0", SEC_CCD_ADVANCED, CTRL_RANGE, "Border - top", "pxl", 0.0, 65535.0, 1.0, 0 },
	{ "CCD", "BorderBottom", DATA_INT, "0", SEC_CCD_ADVANCED, CTRL_RANGE, "Border - bottom", "pxl", 0.0, 65535.0, 1.0, 0 },

	{ "Conversion", "Format", DATA_INT, "0", SEC_CONVERSION, CTRL_SELECT, "Image data format", NULL, 0, 0, 0, 0, BitpixStrings },

	{ "Calibration", "Advanced", DATA_INT, "0", SEC_CALIBRATION, CTRL_CHECK, "Advanced calibration scheme" },

	{ "MasterBias", "Format", DATA_INT, "0", SEC_MASTER_BIAS, CTRL_SELECT, "Image data format", NULL, 0, 0, 0, 0, BitpixStrings },

	{ "MasterDark", "Format", DATA_INT, "0", SEC_MASTER_DARK, CTRL_SELECT, "Image data format", NULL, 0, 0, 0, 0, BitpixStrings },
	{ "MasterDark", "Scalable", DATA_INT, "0", SEC_MASTER_DARK, CTRL_CHECK, "Make scalable dark frame" },

	{ "MasterFlat", "Format", DATA_INT, "0", SEC_MASTER_FLAT, CTRL_SELECT, "Image data format", NULL, 0, 0, 0, 0, BitpixStrings },
	{ "MasterFlat", "Level", DATA_DBL, "10000.0", SEC_MASTER_FLAT, CTRL_RANGE, "Output mean level", "ADU", 0, 0xFFFFFFFF, 1, 1 },

	{ "MuniFind", "Threshold", DATA_DBL, "60.0", SEC_MUNIFIND, CTRL_RANGE, "Valid measurements required", "%", 0.0, 100.0, 1.0, 0 },

	{ "Photometry", "ReadNoise", DATA_DBL, "15.0", SEC_PHOTOMETRY, CTRL_RANGE, "Readout noise", "e-", 0.01, 99.99, 0.1, 2 },
	{ "Photometry", "ADCGain", DATA_DBL, "2.3", SEC_CCD, CTRL_RANGE, "ADC gain", "e-/ADU", 0.01, 99.99, 0.1, 2 },
	{ "Photometry", "Fwhm", DATA_DBL, "3.0", SEC_STAR_DETECTION, CTRL_RANGE, "Average FWHM of objects", "pxl", 3.0, 99.99, 0.1, 2 },
	{ "Photometry", "Threshold", DATA_DBL, "4.0", SEC_STAR_DETECTION, CTRL_RANGE, "Detection threshold", NULL, 1.0, 99.99, 0.1, 2 },
	{ "Photometry", "MinSharpness", DATA_DBL, "0.2", SEC_DET_ADVANCED, CTRL_RANGE, "Min. sharpness", NULL, 0.01, 9.99, 0.1, 2 },
	{ "Photometry", "MaxSharpness", DATA_DBL, "1.0", SEC_DET_ADVANCED, CTRL_RANGE, "Max. sharpness", NULL, 0.01, 9.99, 0.1, 2 },
	{ "Photometry", "MinRoundness", DATA_DBL, "-1.0", SEC_DET_ADVANCED, CTRL_RANGE, "Min. roundness", NULL, -9.99, 9.99, 0.1, 2 },
	{ "Photometry", "MaxRoundness", DATA_DBL, "1.0", SEC_DET_ADVANCED, CTRL_RANGE, "Max. roundness", NULL, -9.99, 9.99, 0.1, 2 },

	{ "Photometry", "Aperture1", DATA_DBL, "2.0", SEC_APERTURES, CTRL_RANGE, "Aperture #1", NULL, 1.0, 99.9, 0.1, 1 },
	{ "Photometry", "Aperture2", DATA_DBL, "2.7273", SEC_APERTURES, CTRL_RANGE, "Aperture #2", NULL, 1.0, 99.9, 0.1, 1 },
	{ "Photometry", "Aperture3", DATA_DBL, "3.8182", SEC_APERTURES, CTRL_RANGE, "Aperture #3", NULL, 1.0, 99.9, 0.1, 1 },
	{ "Photometry", "Aperture4", DATA_DBL, "5.2727", SEC_APERTURES, CTRL_RANGE, "Aperture #4", NULL, 1.0, 99.9, 0.1, 1 },
	{ "Photometry", "Aperture5", DATA_DBL, "7.0909", SEC_APERTURES, CTRL_RANGE, "Aperture #5", NULL, 1.0, 99.9, 0.1, 1 },
	{ "Photometry", "Aperture6", DATA_DBL, "9.2727", SEC_APERTURES, CTRL_RANGE, "Aperture #6", NULL, 1.0, 99.9, 0.1, 1 },
	{ "Photometry", "Aperture7", DATA_DBL, "11.8182", SEC_APERTURES, CTRL_RANGE, "Aperture #7", NULL, 1.0, 99.9, 0.1, 1 },
	{ "Photometry", "Aperture8", DATA_DBL, "14.7273", SEC_APERTURES, CTRL_RANGE, "Aperture #8", NULL, 1.0, 99.9, 0.1, 1 },
	{ "Photometry", "Aperture9", DATA_DBL, "18.0", SEC_APERTURES, CTRL_RANGE, "Aperture #9", NULL, 1.0, 99.9, 0.1, 1 },
	{ "Photometry", "Aperture10", DATA_DBL, "21.6364", SEC_APERTURES, CTRL_RANGE, "Aperture #10", NULL, 1.0, 99.9, 0.1, 1 },
	{ "Photometry", "Aperture11", DATA_DBL, "25.6364", SEC_APERTURES, CTRL_RANGE, "Aperture #11", NULL, 1.0, 99.9, 0.1, 1 },
	{ "Photometry", "Aperture12", DATA_DBL, "30.0", SEC_APERTURES, CTRL_RANGE, "Aperture #12", NULL, 1.0, 99.9, 0.1, 1 },

	{ "Photometry", "SkyInRadius", DATA_DBL, "20.0", SEC_SKY_ANNULUS, CTRL_RANGE, "Sky annulus inner radius", NULL, 1.0, 99.9, 0.1, 1 },
	{ "Photometry", "SkyOutRadius", DATA_DBL, "30.0", SEC_SKY_ANNULUS, CTRL_RANGE, "Sky annulus outer radius", NULL, 1.0, 99.9, 0.1, 1 },

	{ "Matching", "RStars", DATA_INT, "10", SEC_MATCHING, CTRL_RANGE, "Number of stars", NULL, 3, 99, 1 },
	{ "Matching", "IStars", DATA_INT, "5", SEC_MATCHING, CTRL_RANGE, "Number of vertices", NULL, 3, 10, 1 },
	{ "Matching", "Clip", DATA_DBL, "2.5", SEC_MATCHING, CTRL_RANGE, "Clipping threshold", NULL, 0.1, 100.0, 0.1, 1 },
	{ "Matching", "SpFields", DATA_INT, "0", SEC_MATCHING, CTRL_SELECT, "Use algorithm for sparse fields", NULL, 0, 0, 0, 0, MatchStrings },
	{ "Matching", "MaxOffset", DATA_INT, "3", SEC_MATCHING, CTRL_RANGE, "Max. position offset (sparse fields only)", "pxl", 1, 999, 1, 0 },

	{ "LightCurve", "JDFormat", DATA_INT, "0", SEC_LIGHT_CURVE, CTRL_RADIO_GROUP, "Date & time format", NULL, 0, 0, 0, 0, JDFormat },
	{ "LightCurve", "OptColumns", DATA_INT, "0", SEC_LIGHT_CURVE, CTRL_CHECK_GROUP, "Optional columns", NULL, 0, 0, 0, 0, OptColumns },
	{ "LightCurve", "OmitEmptyLines", DATA_BOOL, "0", SEC_LIGHT_CURVE, CTRL_CHECK, "Omit empty lines" },

	{ "BRKA", "Enable", DATA_BOOL, "0", SEC_OBJECT, CTRL_CHECK, "B.R.K.A. catalogue" },
	{ "BRKA", "Path", DATA_STR, "", SEC_OBJECT, CTRL_FILE_CHOOSER, "Path" },

	{ "GCVS", "Enable", DATA_BOOL, "0", SEC_OBJECT, CTRL_CHECK, "GCVS catalogue" },
	{ "GCVS", "Path", DATA_STR, "", SEC_OBJECT, CTRL_FILE_CHOOSER, "Path" },

	{ "NSV", "Enable", DATA_BOOL, "0", SEC_OBJECT, CTRL_CHECK, "NSV catalogue" },
	{ "NSV", "Path", DATA_STR, "", SEC_OBJECT, CTRL_FILE_CHOOSER, "Path" },

	{ "NSVS", "Enable", DATA_BOOL, "0", SEC_OBJECT, CTRL_CHECK, "NSVS catalogue" },
	{ "NSVS", "Path", DATA_STR, "", SEC_OBJECT, CTRL_FILE_CHOOSER, "Path" },

	{ "Observer", "Location", DATA_STR, "", SEC_OBSERVER, CTRL_EDIT, "Default location", NULL, 0, 0, 0, 80 },
	{ "Observer", "Longitude", DATA_STR, "", SEC_OBSERVER, CTRL_EDIT, "Longitude", "E ddd mm ss", 0, 0, 0, 11 },
	{ "Observer", "Latitude", DATA_STR, "", SEC_OBSERVER, CTRL_EDIT, "Latitude", "N dd mm ss", 0, 0, 0, 10 },

	{ NULL }
};

// Parameter runtime data
struct tParData
{
	const tParameter *info;		// Static data
	tValue		value;			// Parameter value
	int			nwidgets;		// Number of widgets
	GtkWidget	**widgets;		// List of widgets
};

// Tab descriptor
struct tTabData
{
	const tSection *info;		// Static data
	int			index;			// Tab index 
	int         count;			// Number of parameters
	tParData	*list;			// Parameter data
};

//-------------------------   HELPER FUNCTIONS   --------------------------------

struct tFindData
{
	int section;
	GtkTreeRowReference *rowref;
};

gboolean find_section(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter,
			tFindData *data)
{
	int section;

	gtk_tree_model_get(model, iter, MCOL_ID, &section, -1);
	if (section==data->section) {
		data->rowref = gtk_tree_row_reference_new(model, path);
		return TRUE;
	}
	return FALSE;
}

//-------------------------   CONFIGURATION DIALOG   --------------------------------

CConfigDlg::CConfigDlg(GtkWindow *pParent)
{
	GtkWidget *hbox;

	m_Section = SEC_ROOT;
	m_Tabs = NULL;
	m_TabCount = 0;
	m_Updating = false;

	MakeData();

	// Make list store
	m_TreeStore = gtk_tree_store_new(NUM_MCOLS, G_TYPE_INT, G_TYPE_STRING, G_TYPE_INT);
	MakeSection(m_TreeStore, SecList, SEC_ROOT, NULL);

	// Dialog with buttons
	m_pDlg = gtk_dialog_new_with_buttons("Preferences", pParent, 
		(GtkDialogFlags)(GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR),
		GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, NULL);

	// Dialog icon
	gchar *icon = get_icon_file("preferences");
	gtk_window_set_icon(GTK_WINDOW(m_pDlg), gdk_pixbuf_new_from_file(icon, NULL));
	g_free(icon);

	// Dialog layout
	hbox = gtk_hbox_new(FALSE, 4);
	gtk_container_add (GTK_CONTAINER(GTK_DIALOG(m_pDlg)->vbox), hbox);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 4);

	// Left panel with categories
	MakeTreeView();
	gtk_widget_set_size_request(m_ScrWnd, 200, -1);
	gtk_box_pack_start(GTK_BOX(hbox), m_ScrWnd, FALSE, FALSE, 0);

	// Right panel with parameters
	MakeNotebook();
	gtk_widget_set_size_request(GTK_WIDGET(m_Notebook), 360, 360);
	gtk_box_pack_start(GTK_BOX(hbox), m_Notebook, TRUE, TRUE, 0);

	gtk_widget_show_all(hbox);

	// Signals:
	GtkTreeSelection *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(m_TreeView));
	g_signal_connect(G_OBJECT(selection), "changed", G_CALLBACK(selection_changed), this);
}

CConfigDlg::~CConfigDlg()
{
	int i, j;

	gtk_widget_destroy(m_pDlg);
	g_object_unref(m_TreeStore);

	if (m_TabCount>0) {
		for (i=0; i<m_TabCount; i++) {
			struct tTabData *tab = &m_Tabs[i];
			if (tab->count>0) {
				for (j=0; j<tab->count; j++) {
					struct tParData *par = &tab->list[j];
					if (par->info->data==DATA_STR && par->value.Str)
						g_free(par->value.Str);
				}
				g_free(tab->list);
			}
		}
		g_free(m_Tabs);
	}
}

gint CConfigDlg::selection_changed(GtkTreeSelection *widget, CConfigDlg *data)
{
	data->OnSelectionChanged();
	return FALSE;
}

void CConfigDlg::Execute()
{
	bool res;

	LoadConfig();
	res = gtk_dialog_run(GTK_DIALOG(m_pDlg))==GTK_RESPONSE_ACCEPT;
	if (res) {
		SaveConfig();
		Global.pProject->OnConfigChange();
		CConfig::Flush();
	}
}

void CConfigDlg::LoadConfig()
{
	int i, j;

	for (i=0; i<m_TabCount; i++) {
		struct tTabData *tab = &m_Tabs[i];
		for (j=0; j<tab->count; j++) {
			LoadData(&tab->list[j]);
			SetData(&tab->list[j]);
		}
	}
}

void CConfigDlg::LoadData(tParData *pData)
{
	switch (pData->info->data)
	{
	case DATA_BOOL:
		{
			bool defval = (pData->info->defval ? atoi(pData->info->defval)!=0 : false);
			pData->value.Bool = CConfig::GetBool(pData->info->group, pData->info->key, defval);
		}
		break;

	case DATA_INT:
		{
			int defval = (pData->info->defval ? atoi(pData->info->defval) : 0);
			pData->value.Int = CConfig::GetInt(pData->info->group, pData->info->key, defval);
		}
		break;

	case DATA_DBL:
		{
			double defval = (pData->info->defval ? atof(pData->info->defval) : 0.0);
			pData->value.Dbl = CConfig::GetDbl(pData->info->group, pData->info->key, defval);
		}
		break;

	case DATA_STR:
		{
			const char *defval = (pData->info->defval ? pData->info->defval : "");
			pData->value.Str = CConfig::GetStr(pData->info->group, pData->info->key, defval);
		}
		break;
	}
}

void CConfigDlg::SaveData(const tParData *pData)
{
	switch (pData->info->data)
	{
	case DATA_BOOL:
		CConfig::SetBool(pData->info->group, 
			pData->info->key, pData->value.Bool);
		break;

	case DATA_INT:
		CConfig::SetInt(pData->info->group, 
			pData->info->key, pData->value.Int);
		break;

	case DATA_DBL:
		CConfig::SetDbl(pData->info->group, 
			pData->info->key, pData->value.Dbl);
		break;

	case DATA_STR:
		CConfig::SetStr(pData->info->group, 
			pData->info->key, pData->value.Str);
		break;
	}
}


void CConfigDlg::MakeControl(const tParameter *info, GtkWidget ***widgets, int *nwidgets)
{
	int i, count;
	GtkWidget *widget;
	GtkObject *adj;

	switch (info->ctrl)
	{
	case CTRL_CHECK:
		*nwidgets = 1;
		*widgets = (GtkWidget**)g_malloc0(sizeof(GtkWidget*));
		(*widgets)[0] = widget = gtk_check_button_new_with_label("");
		break;

	case CTRL_RANGE:
		*nwidgets = 1;
		*widgets = (GtkWidget**)g_malloc0(sizeof(GtkWidget*));
		adj = gtk_adjustment_new(0, info->minval, info->maxval, 
			info->step, 10*info->step, 0);
		(*widgets)[0] = widget = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, info->digits);
		gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(widget), true);
		break;

	case CTRL_EDIT:
		*nwidgets = 1;
		*widgets = (GtkWidget**)g_malloc0(sizeof(GtkWidget*));
		(*widgets)[0] = widget = gtk_entry_new_with_max_length(info->digits);
		break;

	case CTRL_CHECK_GROUP:
		count = 0;
		for (i=0; info->strings[i].str!=NULL; i++)
			count++;
		*nwidgets = count;
		*widgets = (GtkWidget**)g_malloc0(count*sizeof(GtkWidget*));
		for (i=0; i<count; i++)
			(*widgets)[i] = gtk_check_button_new_with_label(info->strings[i].str);
		break;

	case CTRL_RADIO_GROUP:
		count = 0;
		for (i=0; info->strings[i].str!=NULL; i++)
			count++;
		*nwidgets = count;
		*widgets = (GtkWidget**)g_malloc0(count*sizeof(GtkWidget*));
		for (i=0, widget=NULL; i<count; i++) {
			widget = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(widget), 
				info->strings[i].str);
			(*widgets)[i] = widget;
		}
		break;

	case CTRL_SELECT:
		*nwidgets = 1;
		*widgets = (GtkWidget**)g_malloc0(sizeof(GtkWidget*));
		(*widgets)[0] = widget = gtk_combo_box_new_text();
		for (i=0; info->strings[i].str!=NULL; i++)
			gtk_combo_box_append_text(GTK_COMBO_BOX(widget), info->strings[i].str);
		break;

	case CTRL_FILE_CHOOSER:
		*nwidgets = 2;
		*widgets = (GtkWidget**)g_malloc0(2*sizeof(GtkWidget*));
		(*widgets)[0] = widget = gtk_entry_new();
		(*widgets)[1] = widget = gtk_button_new_with_label("...");
		g_signal_connect(G_OBJECT((*widgets)[1]), "clicked", G_CALLBACK(filebutton_clicked), this);
		break;

	default:
		*nwidgets = 0;
		*widgets = NULL;
	}
}


void CConfigDlg::SetData(const tParData *pData)
{
	bool active;
	double value;
	int i, index, option, bitmask;

	switch (pData->info->ctrl)
	{
	case CTRL_CHECK:
		switch (pData->info->data)
		{
		case DATA_BOOL:
			active = pData->value.Bool;
			break;
		case DATA_INT:
			active = pData->value.Int!=0;
			break;
		case DATA_STR:
			active = (pData->value.Str ? atoi(pData->value.Str)!=0 : false);
			break;
		default:
			active = false;
		}
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pData->widgets[0]), active);
		break;

	case CTRL_RANGE:
		switch (pData->info->data)
		{
		case DATA_INT:
			value = (double)(pData->value.Int);
			break;
		case DATA_DBL:
			value = pData->value.Dbl;
			break;
		case DATA_STR:
			value = (pData->value.Str ? atof(pData->value.Str) : 0.0);
			break;
		default:
			value = 0.0;
		}
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(pData->widgets[0]), value);
		break;

	case CTRL_SELECT:
		switch (pData->info->data)
		{
		case DATA_INT:
			option = pData->value.Int;
			break;
		case DATA_STR:
			option = (pData->value.Str ? atoi(pData->value.Str) : -1);
			break;
		default:
			option = -1;
		}
		index = -1;
		if (option!=-1 && pData->info->strings) {
			for (i=0; pData->info->strings[i].str; i++) {
				if (pData->info->strings[i].id == option) {
					index = i;
					break;
				}
			}
		}
		gtk_combo_box_set_active(GTK_COMBO_BOX(pData->widgets[0]), index);
		break;

	case CTRL_RADIO_GROUP:
		switch (pData->info->data)
		{
		case DATA_INT:
			option = pData->value.Int;
			break;
		case DATA_STR:
			option = (pData->value.Str ? atoi(pData->value.Str) : -1);
			break;
		default:
			option = -1;
		}
		for (i=0; i<pData->nwidgets; i++) {
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pData->widgets[i]), 
				pData->info->strings[i].id == option);
		}
		break;

	case CTRL_CHECK_GROUP:
		switch (pData->info->data)
		{
		case DATA_BOOL:
			bitmask = (pData->value.Bool ? 1 : 0);
		case DATA_INT:
			bitmask = pData->value.Int;
			break;
		case DATA_STR:
			bitmask = (pData->value.Str ? atoi(pData->value.Str) : 0);
			break;
		default:
			bitmask = 0;
		}
		for (i=0; i<pData->nwidgets; i++) {
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pData->widgets[i]), 
				(bitmask & pData->info->strings[i].id)!=0);
		}
		break;

	case CTRL_FILE_CHOOSER:
	case CTRL_EDIT:
		if (pData->info->data == DATA_STR) 
			gtk_entry_set_text(GTK_ENTRY(pData->widgets[0]), pData->value.Str);
		break;

	default:
		break;
	}
}

void CConfigDlg::SetDefault(const tParData *pData)
{
	bool active;
	double value;
	int i, index, option, bitmask;

	switch (pData->info->ctrl)
	{
	case CTRL_CHECK:
		active = (pData->info->defval ? atoi(pData->info->defval)!=0 : false);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pData->widgets[0]), active);
		break;

	case CTRL_RANGE:
		value = (pData->info->defval ? atof(pData->info->defval) : 0.0);
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(pData->widgets[0]), value);
		break;

	case CTRL_SELECT:
		option = (pData->info->defval ? atoi(pData->info->defval) : -1);
		index = -1;
		if (option!=-1 && pData->info->strings) {
			for (i=0; pData->info->strings[i].str; i++) {
				if (pData->info->strings[i].id == option) {
					index = i;
					break;
				}
			}
		}
		gtk_combo_box_set_active(GTK_COMBO_BOX(pData->widgets[0]), index);
		break;

	case CTRL_RADIO_GROUP:
		option = (pData->info->defval ? atoi(pData->info->defval) : -1);
		for (i=0; i<pData->nwidgets; i++) {
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pData->widgets[i]), 
				pData->info->strings[i].id == option);
		}
		break;

	case CTRL_CHECK_GROUP:
		bitmask = (pData->info->defval ? atoi(pData->info->defval) : 0);
		for (i=0; i<pData->nwidgets; i++) {
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pData->widgets[i]), 
				(bitmask & pData->info->strings[i].id)!=0);
		}
		break;

	case CTRL_FILE_CHOOSER:
	case CTRL_EDIT:
		gtk_entry_set_text(GTK_ENTRY(pData->widgets[0]), pData->info->defval);
		break;

	default:
		break;
	}
}

void CConfigDlg::GetData(tParData *pData)
{
	int i, index, option, bitmask;
	bool active;
	double value;

	switch (pData->info->ctrl)
	{
	case CTRL_CHECK:
		active = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pData->widgets[0]))!=0;
		switch (pData->info->data)
		{
		case DATA_BOOL:
			pData->value.Bool = active;
			break;
		case DATA_INT:
			pData->value.Int = (active ? 1 : 0);
			break;
		case DATA_STR:
			pData->value.Str = (char*)g_realloc(pData->value.Str, 2*sizeof(char));
			strcpy(pData->value.Str, (active ? "0" : "1"));
			break;
		default:
			break;
		}
		break;

	case CTRL_RANGE:
		value = gtk_spin_button_get_value(GTK_SPIN_BUTTON(pData->widgets[0]));
		switch (pData->info->data)
		{
		case DATA_INT:
			pData->value.Int = (int)(value>=0.0 ? value+0.5 : value-0.5);
			break;
		case DATA_DBL:
			pData->value.Dbl = value;
			break;
		case DATA_STR:
			pData->value.Str = (char*)g_realloc(pData->value.Str, 64*sizeof(char));
			sprintf(pData->value.Str, "%.*f", pData->info->digits, value);
			break;
		default:
			break;
		}
		break;

	case CTRL_SELECT:
		index = gtk_combo_box_get_active(GTK_COMBO_BOX(pData->widgets[0]));
		if (index>=0 && pData->info->strings)
			option = pData->info->strings[index].id;
		else
			option = -1;
		switch (pData->info->data)
		{
		case DATA_INT:
			pData->value.Int = option;
			break;
		case DATA_STR:
			pData->value.Str = (char*)g_realloc(pData->value.Str, 64*sizeof(char));
			sprintf(pData->value.Str, "%d", option);
			break;
		default:
			break;
		}
		break;

	case CTRL_RADIO_GROUP:
		option = -1;
		for (i=0; i<pData->nwidgets; i++) {
			if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pData->widgets[i])))
				option = pData->info->strings[i].id;
		}
		switch (pData->info->data)
		{
		case DATA_INT:
			pData->value.Int = option;
			break;
		case DATA_STR:
			pData->value.Str = (char*)g_realloc(pData->value.Str, 64*sizeof(char));
			sprintf(pData->value.Str, "%d", option);
			break;
		default:
			break;
		}
		break;

	case CTRL_CHECK_GROUP:
		bitmask = 0;
		for (i=0; i<pData->nwidgets; i++) {
			if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pData->widgets[i])))
				bitmask |= pData->info->strings[i].id;
		}
		switch (pData->info->data)
		{
		case DATA_INT:
			pData->value.Int = bitmask;
			break;
		case DATA_STR:
			pData->value.Str = (char*)g_realloc(pData->value.Str, 64*sizeof(char));
			sprintf(pData->value.Str, "%d", bitmask);
			break;
		default:
			break;
		}
		break;

	case CTRL_FILE_CHOOSER:
	case CTRL_EDIT:
		if (pData->info->data == DATA_STR) {
			const char *text = gtk_entry_get_text(GTK_ENTRY(pData->widgets[0]));
			pData->value.Str = (char*)g_realloc(pData->value.Str, (strlen(text)+1));
			strcpy(pData->value.Str, text);
		}
		break;

	default:
		break;
	}
}

void CConfigDlg::SaveConfig()
{
	int i, j;

	for (i=0; i<m_TabCount; i++) {
		struct tTabData *tab = &m_Tabs[i];
		for (j=0; j<tab->count; j++) {
			GetData(&tab->list[j]);
			SaveData(&tab->list[j]);
		}
	}
}

void CConfigDlg::MakeData(void)
{
	int i, j, k, count;

	// Number of tabs 
	count = 0;
	for (i=0; SecList[i].id>0; i++) 
		count++;
	if (count<=0)
		return;

	m_TabCount = count;
	m_Tabs = (tTabData*)g_malloc0(count*sizeof(tTabData));
	for (i=0; i<m_TabCount; i++) {
		int section = SecList[i].id;
		m_Tabs[i].info = &SecList[i];
		// Number of parameters in each tab
		count = 0;
		for (j=0; ParList[j].group!=NULL; j++) {
			if (ParList[j].section==section)
				count++;
		}
		if (count>0) {
			m_Tabs[i].count = count;
			m_Tabs[i].list = (tParData*)g_malloc0(count*sizeof(tParData));
			for (j=0, k=0; ParList[j].group!=NULL; j++) {
				if (ParList[j].section==section) {
					m_Tabs[i].list[k].info = &ParList[j];
					k++;
				}
			}
		}
	}
}

void CConfigDlg::MakeTreeView(void)
{
	// Make tree view
	m_TreeView = gtk_tree_view_new();
	GtkTreeViewColumn *col = gtk_tree_view_column_new();

	// Set column name and alignment
	gtk_tree_view_column_set_title(col, "Category");
	gtk_tree_view_append_column(GTK_TREE_VIEW(m_TreeView), col);

	// Text renderer
	GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(col, renderer, TRUE);
	gtk_tree_view_column_add_attribute(col, renderer, "text", MCOL_CAPTION);
	gtk_tree_view_set_model(GTK_TREE_VIEW(m_TreeView), GTK_TREE_MODEL(m_TreeStore));
	gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(m_TreeView)), 
		GTK_SELECTION_BROWSE);

	// Scrolled window
	m_ScrWnd = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(m_ScrWnd),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(m_ScrWnd), 
		GTK_SHADOW_ETCHED_IN);
	gtk_container_add(GTK_CONTAINER(m_ScrWnd), m_TreeView);

	gtk_tree_view_expand_all(GTK_TREE_VIEW(m_TreeView));
	//gtk_tree_view_set_show_expanders(GTK_TREE_VIEW(m_TreeView), false);
	//gtk_tree_view_set_level_indentation(GTK_TREE_VIEW(m_TreeView), 20);
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(m_TreeView), false);

	GtkTreeSelection *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(m_TreeView));
	GtkTreePath *first = gtk_tree_path_new_first();
	gtk_tree_selection_select_path(selection, first);
	gtk_tree_path_free(first);
}

void CConfigDlg::MakeNotebook(void)
{
	int i, j, k, row, nwidgets;
	GtkWidget **widgets, *table, *label, *btnbox, *defbtn, *separ;

	m_Notebook = gtk_notebook_new();
	// First page
	gtk_notebook_append_page(GTK_NOTEBOOK(m_Notebook),
		gtk_label_new("Select a category"), NULL);
	// Other pages
	for (i=0; i<m_TabCount; i++) {
		GtkWidget *child = NULL;
		if (m_Tabs[i].count>0) {
			child = gtk_vbox_new(false, 0);
			gtk_container_set_border_width(GTK_CONTAINER(child), 4);
			table = gtk_table_new(m_Tabs[i].count, 3, false);
			for (j=0, row=0; j<m_Tabs[i].count; j++) {
				MakeControl(m_Tabs[i].list[j].info, &widgets, &nwidgets);
				if (widgets && nwidgets>0) {
					if (m_Tabs[i].list[j].info->caption) {
						label = gtk_label_new(m_Tabs[i].list[j].info->caption);
						gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
						gtk_table_attach(GTK_TABLE(table), label, 0, 1, row, row+1,
							(GtkAttachOptions)(GTK_FILL | GTK_EXPAND), GTK_FILL, 0, 0);
					}
					if (m_Tabs[i].list[j].info->unit) {
						label = gtk_label_new(m_Tabs[i].list[j].info->unit);
						gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
						gtk_table_attach(GTK_TABLE(table), label, 2, 3, row, row+1, 
							GTK_FILL, GTK_FILL, 0, 0);
					}
					switch (m_Tabs[i].list[j].info->ctrl)
					{
					case CTRL_EDIT:
					case CTRL_RANGE:
					case CTRL_SELECT:
						for (k=0; k<nwidgets; k++) {
							gtk_table_attach_defaults(GTK_TABLE(table), widgets[k], 1, 2, row, row+1);
							row++;
						}
						row--;
						break;

					case CTRL_FILE_CHOOSER:
						for (k=0; k<nwidgets; k++)
							gtk_table_attach_defaults(GTK_TABLE(table), widgets[k], k+1, k+2, row, row+1);
						break;

					default:
						for (k=0; k<nwidgets; k++) {
							gtk_table_attach_defaults(GTK_TABLE(table), widgets[k], 1, 3, row, row+1);
							row++;
						}
						row--;
					}
					row++;
					m_Tabs[i].list[j].widgets = widgets;
					m_Tabs[i].list[j].nwidgets = nwidgets;
				}
			}
			gtk_table_set_row_spacings(GTK_TABLE(table), 4);
			gtk_table_set_col_spacings(GTK_TABLE(table), 8);
			gtk_box_pack_start(GTK_BOX(child), table, FALSE, FALSE, 0);
			// Fill
			label = gtk_label_new("");
			gtk_box_pack_start(GTK_BOX(child), label, TRUE, FALSE, 0);
			// Separator
			separ = gtk_hseparator_new();
			gtk_box_pack_start(GTK_BOX(child), separ, FALSE, FALSE, 2);
			// Buttons
			btnbox = gtk_hbutton_box_new();
			gtk_button_box_set_layout(GTK_BUTTON_BOX(btnbox), GTK_BUTTONBOX_END);
			defbtn = gtk_button_new_with_label("Set defaults");
			g_signal_connect(G_OBJECT(defbtn), "clicked", G_CALLBACK(setdefaults_clicked), this);
			gtk_box_pack_start(GTK_BOX(btnbox), defbtn, false, false, 0);
			gtk_box_pack_start(GTK_BOX(child), btnbox, FALSE, FALSE, 2);
		} else {
			child = gtk_label_new("There are no options for this category"); 
		}
		if (child) 
			m_Tabs[i].index = gtk_notebook_append_page(GTK_NOTEBOOK(m_Notebook),
				child, NULL);
		else
			m_Tabs[i].index = -1;
	}

	gtk_notebook_set_current_page(GTK_NOTEBOOK(m_Notebook), 0);
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(m_Notebook), false);
}

void CConfigDlg::MakeSection(GtkTreeStore *tree, const tSection *list,
		int parent_id, GtkTreeIter *parent)
{
	int i;
	GtkTreeIter iter;
	
	for (i=0; list[i].id>0; i++) {
		if (list[i].parent==parent_id && list[i].id!=list[i].parent) {
			gtk_tree_store_append(tree, &iter, parent);
			gtk_tree_store_set(tree, &iter, MCOL_ID, list[i].id, MCOL_CAPTION,
				list[i].caption, MCOL_PARENT, list[i].parent, -1);
			MakeSection(tree, list, list[i].id, &iter);
		}
	}
}

void CConfigDlg::SetCategory(int section)
{
	int i, index;
	tFindData info;

	if (section!=m_Section) {
		m_Section = section;
		// Show tab
		index = 0;
		for (i=0; i<m_TabCount; i++) {
			if (m_Tabs[i].info->id == section) {
				index = m_Tabs[i].index;
				break;
			}
		}
		gtk_notebook_set_current_page(GTK_NOTEBOOK(m_Notebook), index);
		// Select category
		info.section = section;
		info.rowref = NULL;
		gtk_tree_model_foreach(GTK_TREE_MODEL(m_TreeStore), (GtkTreeModelForeachFunc)find_section, &info);
		if (info.rowref) {
			GtkTreeSelection *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(m_TreeView));
			gtk_tree_selection_select_path(selection, gtk_tree_row_reference_get_path(info.rowref));
			gtk_tree_row_reference_free(info.rowref);
		}
	}
}

void CConfigDlg::OnSelectionChanged(void)
{
	int section;
	GtkTreeIter iter;

	if (!m_Updating) {
		m_Updating = true;
		GtkTreeSelection *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(m_TreeView));
		GList* rows = gtk_tree_selection_get_selected_rows(selection, NULL);
		if (g_list_length(rows)>0) {
			gtk_tree_model_get_iter(GTK_TREE_MODEL(m_TreeStore), &iter, (GtkTreePath*)rows->data);
			gtk_tree_model_get(GTK_TREE_MODEL(m_TreeStore), &iter, MCOL_ID, &section, -1);
			SetCategory(section);
		} else {
			SetCategory(-1);
		}
		g_list_foreach(rows, (GFunc)gtk_tree_path_free, NULL);
		g_list_free(rows);
		m_Updating = false;
	}
}

void CConfigDlg::setdefaults_clicked(GtkButton *pWidget, CConfigDlg *pDlg)
{
	pDlg->OnSetDefaultsClicked(pWidget);
}

void CConfigDlg::OnSetDefaultsClicked(GtkButton *pButton)
{
	int i, j;

	for (i=0; i<m_TabCount; i++) {
		struct tTabData *tab = &m_Tabs[i];
		if (tab->info->id == m_Section) {
			for (j=0; j<tab->count; j++)
				SetDefault(&tab->list[j]);
		}
	}
}

void CConfigDlg::filebutton_clicked(GtkButton *pWidget, CConfigDlg *pDlg)
{
	pDlg->OnFileButtonClicked(pWidget);
}

void CConfigDlg::OnFileButtonClicked(GtkButton *pButton)
{
	int i, j;

	for (i=0; i<m_TabCount; i++) {
		struct tTabData *tab = &m_Tabs[i];
		if (tab->info->id == m_Section) {
			for (j=0; j<tab->count; j++) {
				if (tab->list[j].info->ctrl == CTRL_FILE_CHOOSER) {
					if (GTK_BUTTON(tab->list[j].widgets[1]) == pButton) {
						ChooseFile(&tab->list[j]);
						break;
					}
				}
			}
		}
	}
}

void CConfigDlg::ChooseFile(tParData *data)
{
	GtkWidget *pDlg = gtk_file_chooser_dialog_new(data->info->caption,
		GTK_WINDOW(m_pDlg), GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL, 
		GTK_RESPONSE_REJECT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, NULL);
	gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(pDlg), 
		gtk_entry_get_text(GTK_ENTRY(data->widgets[0])));
	if (gtk_dialog_run(GTK_DIALOG(pDlg)) == GTK_RESPONSE_ACCEPT) {
		char *filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(pDlg));
		if (filename) {
			data->value.Str = (char*)g_realloc(data->value.Str, (strlen(filename)+1));
			strcpy(data->value.Str, filename);
			gtk_entry_set_text(GTK_ENTRY(data->widgets[0]), filename);
			g_free(filename);
		} else {
			data->value.Str = (char*)g_realloc(data->value.Str, 1);
			strcpy(data->value.Str, "");
			gtk_entry_set_text(GTK_ENTRY(data->widgets[0]), "");
		}
	  }
	gtk_widget_destroy(pDlg);
}
