/**************************************************************

graph.cpp (C-Munipack project)
Widget which can draw a graph
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#define _USE_MATH_DEFINES

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "graph.h"
#include "graph_data.h"
#include "utils.h"

//-----------------------   CONSTANTS AND DEFINITIONS   ------------------------

#define ALLOC_BY		8

#define CURSOR_SIZE		12

#define POINT_SIZE		4

#define FONT_SIZE		10

#define SCALE_YPADDING	4
#define SCALE_XPADDING	4

#define NAME_XPADDING	4
#define NAME_YPADDING	4

// Number of auto-scale steps
#define NumSteps 75

// This array enumerates allowed steps of the scales in FIXED mode
static const double Steps[NumSteps] = {
		1.0e-12, 2.0e-12, 5.0e-12, 1.0e-11, 2.0e-11, 5.0e-11, 1.0e-10, 2.0e-10, 5.0e-10, 
		1.0e-9,  2.0e-9,  5.0e-9,  1.0e-8,  2.0e-8,  5.0e-8,  1.0e-7,  2.0e-7,  5.0e-7, 
		1.0e-6,  2.0e-6,  5.0e-6,  1.0e-5,  2.0e-5,  5.0e-5,  1.0e-4,  2.0e-4,  5.0e-4,
		1.0e-3,  2.0e-3,  5.0e-3,  1.0e-2,  2.0e-2,  5.0e-2,  1.0e-1,  2.0e-1,  5.0e-1,  
		1.0e+0,  2.0e+0,  5.0e+0,  1.0e+1,  2.0e+1,  5.0e+1,  1.0e+2,  2.0e+2,  5.0e+2, 
		1.0e+3,  2.0e+3,  5.0e+3,  1.0e+4,  2.0e+4,  5.0e+4,  1.0e+5,  2.0e+5,  5.0e+5, 
		1.0e+6,  2.0e+6,  5.0e+6,  1.0e+7,  2.0e+7,  5.0e+7,  1.0e+8,  2.0e+8,  5.0e+8, 
		1.0e+9,  2.0e+9,  5.0e+9,  1.0e+10, 2.0e+10, 5.0e+10, 1.0e+11, 2.0e+11, 5.0e+11,
		1.0e+12, 2.0e+12, 5.0e+12 
};
        
// This array contains number of decimal places, which are needed 
// for scale labels for each step size. 
static const char MinPrecs[NumSteps] = {
        12, 12, 12, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 
		4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

#define NumTSteps 20

// This array enumerates allowed step of the scale in TIME mode (in seconds)
const double TSteps[NumTSteps] = {1.0, 2.0, 5.0, 10.0, 15.0, 30.0, 60.0, 120.0, 300.0, 600.0, 
        900.0, 1800.0, 3600.0, 7200.0, 10800, 21600.0, 43200.0, 86400.0, 172800.0, 
        259200.0};

//-----------------------   GRAPH CLASS ------------------------------------------

// Default constructor
CGraph::CGraph():m_pData(NULL), m_ColCount(0), m_ColCapacity(0), m_Cols(NULL), 
	m_DrawArea(NULL), m_Pixmap(NULL), m_Tags(NULL), m_PixWidth(0), m_PixHeight(0), 
	m_NeedUpdate(true), m_GraphWidth(0), m_GraphHeight(0), m_EnableTrack(false), 
	m_FocusedItem(-1), m_FocusedCol(-1), m_MouseX(-1), m_MouseY(-1)
{
	m_ZoomBase = pow(100.0, 1.0/100.0);
	memset(&m_GraphRect, 0, sizeof(tRect));
	SetRect(&m_GraphPadding, 2, 2, 2, 2);
	SetRect(&m_PanelPadding, 2, 2, 2, 2);
	memset(&m_X, 0, sizeof(tAxis));
	memset(&m_Y, 0, sizeof(tAxis));
	m_X.ShowLabels = m_Y.ShowLabels = true;
	SetXParams("X", 0.0, 1.0, 0.1, 0, 1);
	SetYParams("Y", -1.0, 1.0, 0.1, 0, 1);

	// Create drawing area
	m_DrawArea = gtk_drawing_area_new();
	gtk_widget_set_size_request(GTK_WIDGET(m_DrawArea), 0, 0);
	gtk_widget_set_events(m_DrawArea, GDK_EXPOSURE_MASK | 
			GDK_LEAVE_NOTIFY_MASK | GDK_BUTTON_PRESS_MASK |
			GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK);

	// Signals used to handle backing pixmap 
	g_signal_connect(G_OBJECT(m_DrawArea), "expose_event", G_CALLBACK(expose_event), this);
	g_signal_connect(G_OBJECT(m_DrawArea), "configure_event", G_CALLBACK(configure_event), this);
	g_signal_connect(G_OBJECT(m_DrawArea), "motion_notify_event", G_CALLBACK(motion_notify_event), this);
	g_signal_connect(G_OBJECT(m_DrawArea), "button_press_event", G_CALLBACK(button_press_event), this);
}


//
// Destructor
//
CGraph::~CGraph()
{
	if (m_pData)
		m_pData->Disconnect(this);
	if (m_Cols)
		g_free(m_Cols);
	for (GSList *ptr=m_Tags; ptr!=NULL; ptr=ptr->next) 
		g_free(ptr->data);
	g_slist_free(m_Tags);
	gtk_widget_destroy(m_DrawArea);
}

//----------------------   X AXIS   ---------------------------------------

//
// Set X axis parameters
//
void CGraph::SetXParams(const char *name, double min, double max, double eps, 
						int flags, int Prec)
{
	double a, b;

	if ((max-min) < eps) {
		a = 0.5*(max+min-eps);
		b = 0.5*(max+min+eps);
		min = a;
		max = b;
	}
	if (name)
		strcpy(m_X.Name, name);
	else
		strcpy(m_X.Name, "");
	m_X.Log = (flags & LOG_SCALE)!=0;
	m_X.ShowLabels = (flags & NO_LABELS)==0;
	m_X.ShowGrid = (flags & NO_GRID)==0;
	m_X.Reverse = (flags & REVERSE_SCALE)!=0;
	m_X.LabelsOpposite = (flags & LABELS_TOP)!=0;
	m_X.LabelsInside = (flags & LABELS_INSIDE)!=0;
	m_X.Min = Min(min, max);
	m_X.Max = Max(min, max);
	m_X.Eps = eps;
	m_X.PMin = XToProj(!m_X.Reverse ? m_X.Min : m_X.Max);
	m_X.PMax = XToProj(!m_X.Reverse ? m_X.Max : m_X.Min);
	m_X.Pos = (m_X.PMin + m_X.PMax)/2.0;
	m_X.ZMin = 0.0;
	m_X.ZMax = log((m_X.PMax-m_X.PMin)/m_X.Eps)/log(m_ZoomBase);
	m_X.ZoomPos = m_X.ZMin;
	m_X.Scale = 1.0;
	m_X.Offset = 0.0;
	if (flags & EXP_LABELS)
		m_X.Format = F_EXP;
	else if (flags & TIME_LABELS)
		m_X.Format = F_TIME;
	else
		m_X.Format = F_DEC;
	m_X.Prec = Prec;
	UpdateRectangles();
	UpdatePxlSizeX();
	UpdatePxlSizeY();
	LimitPosX();
	LimitPosY();
	Invalidate();
}


//
// Set X axis style
//
void CGraph::SetXFlags(int mask, int flags)
{
	flags = (GetXFlags() & ~mask) | flags;
	m_X.Log = (flags & LOG_SCALE)!=0;
	m_X.ShowLabels = (flags & NO_LABELS)==0;
	m_X.ShowGrid = (flags & NO_GRID)==0;
	m_X.Reverse = (flags & REVERSE_SCALE)!=0;
	m_X.LabelsOpposite = (flags & LABELS_TOP)!=0;
	m_X.LabelsInside = (flags & LABELS_INSIDE)!=0;
	m_X.PMin = XToProj(!m_X.Reverse ? m_X.Min : m_X.Max);
	m_X.PMax = XToProj(!m_X.Reverse ? m_X.Max : m_X.Min);
	m_X.Pos = LimitValue(m_X.Pos, m_X.PMin, m_X.PMax);
	m_X.ZMin = 0.0;
	m_X.ZMax = log((m_X.PMax-m_X.PMin)/m_X.Eps)/log(m_ZoomBase);
	m_X.ZoomPos = LimitValue(m_X.ZoomPos, m_X.ZMin, m_X.ZMax);
	if (flags & EXP_LABELS)
		m_X.Format = F_EXP;
	else if (flags & TIME_LABELS)
		m_X.Format = F_TIME;
	else
		m_X.Format = F_DEC;
	UpdateRectangles();
	UpdatePxlSizeX();
	UpdatePxlSizeY();
	LimitPosX();
	LimitPosY();
	Invalidate();
}


//
// Set X axis name
//
void CGraph::SetXName(const char *name)
{
	if (name)
		strcpy(m_X.Name, name);
	else
		strcpy(m_X.Name, "");
	UpdateRectangles();
	UpdatePxlSizeX();
	UpdatePxlSizeY();
	LimitPosX();
	LimitPosY();
	Invalidate();
}


//
// Get X axis style
//
int CGraph::GetXFlags(void) const
{
	return (m_X.Log ? LOG_SCALE : 0) |
		(m_X.Reverse ? REVERSE_SCALE : 0) |
		(m_X.ShowLabels ? 0 : NO_LABELS) |
		(m_X.ShowGrid ? 0 : NO_GRID) |
		(m_X.LabelsInside ? LABELS_INSIDE : 0) |
		(m_X.LabelsOpposite ? LABELS_RIGHT : 0) |
		(m_X.Format==F_EXP ? EXP_LABELS : 0) |
		(m_X.Format==F_TIME ? TIME_LABELS : 0);
}


//
// Physical units -> projection units
//
double CGraph::XToProj(double x) const
{
	if (m_X.Log) {
		if (!m_X.Reverse) 
			return log10(x);
		else
			return -log10(x);
	} else {
		if (!m_X.Reverse)
			return x;
		else
			return -x;
	}
}

//
// Projection units -> physical units
//
double CGraph::ProjToX(double u) const
{
	if (m_X.Log) {
		if (!m_X.Reverse) 
			return pow(10.0, u);
		else
			return pow(10.0, -u);
	} else {
		if (!m_X.Reverse)
			return u;
		else
			return -u;
	}
}


//
// Projection units -> display units
//
double CGraph::XProjToView(double x) const
{
	return m_OffsetX + (x - m_X.Pos)/m_X.PxlSize;
}


//
// Display units -> projection units
//
double CGraph::ViewToProjX(double u) const
{
	return (u - m_OffsetX)*m_X.PxlSize + m_X.Pos;
}


//
// Physical units -> display units
//
double CGraph::XToView(double x) const
{
	return XProjToView(XToProj(x));
}


//
// Physical units -> projection units
//
double CGraph::ViewToX(double u) const
{
	return ProjToX(ViewToProjX(u));
}


//
// Update pixelsize
//
void CGraph::UpdatePxlSizeX()
{
	if (m_GraphWidth>0) {
		double zoom = pow(m_ZoomBase, m_X.ZoomPos);
		m_X.PxlSize = (m_X.PMax - m_X.PMin)/(zoom*m_GraphWidth);
	} else {
		m_X.PxlSize = 0.0;
	}
}


//
// Limit position
//
void CGraph::LimitPosX()
{
	if (fabs(m_X.PxlSize)>0) {
		double left = m_OffsetX + (m_X.PMin - m_X.Pos)/m_X.PxlSize;
		if (left > m_GraphRect.left)
			 m_X.Pos = m_X.PMin + (m_OffsetX - m_GraphRect.left)*m_X.PxlSize;
		else {
			double right = m_OffsetX + (m_X.PMax - m_X.Pos)/m_X.PxlSize;
			if (right < m_GraphRect.right)
				m_X.Pos = m_X.PMax + (m_OffsetX - m_GraphRect.right)*m_X.PxlSize;
		}
	}
}


//
// Set X zoom
//
void CGraph::SetXZoom(double zoom)
{
	if (zoom!=m_X.ZoomPos) {
		m_X.ZoomPos = LimitValue(zoom, m_X.ZMin, m_X.ZMax);
		UpdatePxlSizeX();
		LimitPosX();
		Invalidate();
	}
}


//
// Set X resolution
//
void CGraph::SetXResolution(double res)
{
	SetXZoom(log((m_X.PMax-m_X.PMin)/res)/log(m_ZoomBase));
}


//
// Set X position
//
void CGraph::SetXPosition(double pos)
{
	pos = XToProj(pos);
	if (m_X.Pos!=pos) {
		m_X.Pos = pos;
		LimitPosX();
		Invalidate();
	}
}


//
// Auto zoom for X axis
//
void CGraph::AutoXZoom(void)
{
	bool first;
	int i, count;
	double a, b, f, x, xmin, xmax;

	if (!m_pData)
		return;

	first = true;
	xmin = xmax = 0;
	count = m_pData->Count();
	for (i=0; i<count; i++) {
		x = m_pData->GetX(i);
		if (x>=m_X.Min && x<=m_X.Max) {
			x = XToProj(x);
			if (first) {
				xmin = xmax = x;
				first = false;
			} else {
				if (xmin > x)
					xmin = x;
				if (xmax < x)
					xmax = x;
			}
		}
	}
	if (!first) {
		if (xmax-xmin < m_X.Eps) {
            f = ((xmin+xmax)/2.0-m_X.PMin)/(m_X.PMax-m_X.PMin);
            a = xmin - f * (m_X.Eps - (xmax - xmin));
            b = xmax + (1.0-f) * (m_X.Eps - (xmax - xmin));
		} else {
			a = xmin;
			b = xmax;
		}
		m_X.ZoomPos = log((m_X.PMax - m_X.PMin)/(b - a))/log(m_ZoomBase);
		m_X.ZoomPos = LimitValue(m_X.ZoomPos, m_X.ZMin, m_X.ZMax);
		m_X.Pos = 0.5*(a+b);
		UpdatePxlSizeX();
		LimitPosX();
		Invalidate();
	}
}

//---------------------------   Y AXIS   -----------------------------

//
// Set Y axis parameters
//
void CGraph::SetYParams(const char *name, double min, double max, double eps, 
						int flags, int Prec)
{
	double a, b;

	if ((max-min) < eps) {
		a = 0.5*(max+min-eps);
		b = 0.5*(max+min+eps);
		min = a;
		max = b;
	}
	if (name)
		strcpy(m_Y.Name, name);
	else
		strcpy(m_Y.Name, "");
	m_Y.Log = (flags & LOG_SCALE)!=0;
	m_Y.ShowLabels = (flags & NO_LABELS)==0;
	m_Y.ShowGrid = (flags & NO_GRID)==0;
	m_Y.Reverse = (flags & REVERSE_SCALE)!=0;
	m_Y.LabelsOpposite = (flags & LABELS_RIGHT)!=0;
	m_Y.LabelsInside = (flags & LABELS_INSIDE)!=0;
	m_Y.Min = Min(min, max);
	m_Y.Max = Max(min, max);
	m_Y.Eps = eps;
	m_Y.PMin = YToProj(!m_Y.Reverse ? m_Y.Min : m_Y.Max);
	m_Y.PMax = YToProj(!m_Y.Reverse ? m_Y.Max : m_Y.Min);
	m_Y.Pos = (m_Y.PMin + m_Y.PMax)/2.0;
	m_Y.ZMin = 0.0;
	m_Y.ZMax = log((m_Y.PMax-m_Y.PMin)/m_Y.Eps)/log(m_ZoomBase);
	m_Y.ZoomPos = m_Y.ZMin;
	m_Y.Scale = 1.0;
	m_Y.Offset = 0.0;
	if (flags & EXP_LABELS)
		m_Y.Format = F_EXP;
	else if (flags & TIME_LABELS)
		m_Y.Format = F_TIME;
	else
		m_Y.Format = F_DEC;
	m_Y.Prec = Prec;
	UpdateRectangles();
	UpdatePxlSizeX();
	UpdatePxlSizeY();
	LimitPosX();
	LimitPosY();
	Invalidate();
}


//
// Set Y axis style
//
void CGraph::SetYFlags(int mask, int flags)
{
	flags = (GetYFlags() & ~mask) | flags;
	m_Y.Log = (flags & LOG_SCALE)!=0;
	m_Y.ShowLabels = (flags & NO_LABELS)==0;
	m_Y.ShowGrid = (flags & NO_GRID)==0;
	m_Y.Reverse = (flags & REVERSE_SCALE)!=0;
	m_Y.LabelsOpposite = (flags & LABELS_RIGHT)!=0;
	m_Y.LabelsInside = (flags & LABELS_INSIDE)!=0;
	m_Y.PMin = YToProj(!m_Y.Reverse ? m_Y.Min : m_Y.Max);
	m_Y.PMax = YToProj(!m_Y.Reverse ? m_Y.Max : m_Y.Min);
	m_Y.Pos = LimitValue(m_Y.Pos, m_Y.PMin, m_Y.PMax);
	m_Y.ZMin = 0.0;
	m_Y.ZMax = log((m_Y.PMax-m_Y.PMin)/m_Y.Eps)/log(m_ZoomBase);
	m_Y.ZoomPos = LimitValue(m_Y.ZoomPos, m_Y.ZMin, m_Y.ZMax);
	if (flags & EXP_LABELS)
		m_Y.Format = F_EXP;
	else if (flags & TIME_LABELS)
		m_Y.Format = F_TIME;
	else
		m_Y.Format = F_DEC;
	UpdateRectangles();
	UpdatePxlSizeX();
	UpdatePxlSizeY();
	LimitPosX();
	LimitPosY();
	Invalidate();
}


//
// Set Y axis name
//
void CGraph::SetYName(const char *name)
{
	if (name)
		strcpy(m_Y.Name, name);
	else
		strcpy(m_Y.Name, "");
	UpdateRectangles();
	UpdatePxlSizeX();
	UpdatePxlSizeY();
	LimitPosX();
	LimitPosY();
	Invalidate();
}


//
// Get Y axis style
//
int CGraph::GetYFlags(void) const
{
	return (m_Y.Log ? LOG_SCALE : 0) |
		(m_Y.Reverse ? REVERSE_SCALE : 0) |
		(m_Y.ShowLabels ? 0 : NO_LABELS) |
		(m_Y.ShowGrid ? 0 : NO_GRID) |
		(m_Y.LabelsInside ? LABELS_INSIDE : 0) |
		(m_Y.LabelsOpposite? LABELS_RIGHT : 0) |
		(m_Y.Format==F_EXP ? EXP_LABELS : 0) |
		(m_Y.Format==F_TIME ? TIME_LABELS : 0);
}


//
// Physical units -> projection units
//
double CGraph::YToProj(double y) const
{
	if (m_Y.Log) {
		if (!m_Y.Reverse) 
			return log10(y);
		else
			return -log10(y);
	} else {
		if (!m_Y.Reverse)
			return y;
		else
			return -y;
	}
}

//
// Projection units -> physical units
//
double CGraph::ProjToY(double v) const
{
	if (m_Y.Log) {
		if (!m_Y.Reverse) 
			return pow(10.0, v);
		else
			return pow(10.0, -v);
	} else {
		if (!m_Y.Reverse)
			return v;
		else
			return -v;
	}
}


//
// Projection units -> display units
//
double CGraph::YProjToView(double y) const
{
	return m_OffsetY - (y - m_Y.Pos)/m_Y.PxlSize;
}


//
// Display units -> projection units
//
double CGraph::ViewToProjY(double v) const
{
	return m_Y.Pos - (v - m_OffsetY)*m_Y.PxlSize;
}


//
// Physical units -> display units
//
double CGraph::YToView(double y) const
{
	return YProjToView(YToProj(y));
}


//
// Physical units -> projection units
//
double CGraph::ViewToY(double v) const
{
	return ProjToY(ViewToProjY(v));
}


//
// Update pixelsize
//
void CGraph::UpdatePxlSizeY()
{
	if (m_GraphHeight>0) {
		double zoom = pow(m_ZoomBase, m_Y.ZoomPos);
		m_Y.PxlSize = (m_Y.PMax - m_Y.PMin)/(zoom*m_GraphHeight);
	} else {
		m_Y.PxlSize = 0.0;
	}
}


//
// Limit position
//
void CGraph::LimitPosY()
{
	if (fabs(m_Y.PxlSize)>0) {
		double top = m_OffsetY - (m_Y.PMax - m_Y.Pos)/m_Y.PxlSize;
		if (top > m_GraphRect.top)
			 m_Y.Pos = m_Y.PMax - (m_OffsetY - m_GraphRect.top)*m_Y.PxlSize;
		else {
			double bottom = m_OffsetY - (m_Y.PMin - m_Y.Pos)/m_Y.PxlSize;
			if (bottom < m_GraphRect.bottom)
				m_Y.Pos = m_Y.PMin - (m_OffsetY - m_GraphRect.bottom)*m_Y.PxlSize;
		}
	}
}


//
// Set Y zoom
//
void CGraph::SetYZoom(double zoom)
{
	if (zoom!=m_Y.ZoomPos) {
		m_Y.ZoomPos = LimitValue(zoom, m_Y.ZMin, m_Y.ZMax);
		UpdatePxlSizeY();
		LimitPosY();
		Invalidate();
	}
}


//
// Set Y resolution
//
void CGraph::SetYResolution(double res)
{
	SetYZoom(log((m_Y.PMax-m_Y.PMin)/res)/log(m_ZoomBase));
}


//
// Set Y position
//
void CGraph::SetYPosition(double pos)
{
	pos = YToProj(pos);
	if (m_Y.Pos!=pos) {
		m_Y.Pos = pos;
		LimitPosY();
		Invalidate();
	}
}


//
// Auto zoom for Y axis
//
void CGraph::AutoYZoom(void)
{
	bool first;
	int i, j, count;
	double a, b, f, x, y, ymin, ymax;
	const double *yval;

	if (!m_pData || m_ColCount<=0)
		return;

	first = true;
	ymin = ymax = 0;
	count = m_pData->Count();
	for (i=0; i<count; i++) {
		x = m_pData->GetX(i);
		if (x>=m_X.Min && x<=m_X.Max) {
			yval = m_pData->GetValues(i);
			for (j=0; j<m_ColCount; j++) {
				if (m_Cols[j].Visible && m_Cols[j].ValChannel>=0) {
					y = yval[m_Cols[j].ValChannel];
					if (y>=m_Y.Min && y<=m_Y.Max) {
						y = YToProj(y);
						if (first) {
							ymin = ymax = y;
							first = false;
						} else {
							if (ymin > y)
								ymin = y;
							if (ymax < y)
								ymax = y;
						}
					}
				}
			}
		}
	}
	if (!first) {
		if (ymax-ymin < m_Y.Eps) {
            f = ((ymin+ymax)/2.0-m_Y.PMin)/(m_Y.PMax-m_Y.PMin);
            a = ymin - f * (m_Y.Eps - (ymax - ymin));
            b = ymax + (1.0-f) * (m_Y.Eps - (ymax - ymin));
		} else {
			a = ymin;
			b = ymax;
		}
		m_Y.ZoomPos = log((m_Y.PMax - m_Y.PMin)/(b - a))/log(m_ZoomBase);
		m_Y.ZoomPos = LimitValue(m_Y.ZoomPos, m_Y.ZMin, m_Y.ZMax);
		m_Y.Pos = 0.5*(a+b);
		UpdatePxlSizeY();
		LimitPosY();
		Invalidate();
	}
}

//---------------------------   COLUMNS    ----------------------------

//
// Clear all columns
//
void CGraph::ClearColumns(void)
{
	if (m_Cols) {
		g_free(m_Cols);
		m_Cols = NULL;
	}
	m_ColCount = m_ColCapacity = 0;
}

//
// Change number of columns
//
int CGraph::AddColumn(int val_channel, int err_channel, int flags, tColor color)
{
	int index;

	if (m_ColCount>=m_ColCapacity) {
		m_ColCapacity += ALLOC_BY;
		m_Cols = (tColumn*)g_realloc(m_Cols, m_ColCapacity*sizeof(tColumn));
	}
	index = m_ColCount;
	m_Cols[index].ValChannel = val_channel;
	m_Cols[index].ErrChannel = err_channel;
	m_Cols[index].Visible = (flags & HIDDEN)==0;
	m_Cols[index].ShowErrors = (flags & NO_ERRORS)==0;
	m_Cols[index].Color = color;
	m_ColCount++;
	Invalidate();
	return index;
}

//
// Change number of columns
//
void CGraph::SetColumn(int col, int val_channel, int err_channel, int flags, tColor color)
{
	if (col>=0 && col<m_ColCount) {
		m_Cols[col].ValChannel = val_channel;
		m_Cols[col].ErrChannel = err_channel;
		m_Cols[col].Visible = (flags & HIDDEN)==0;
		m_Cols[col].ShowErrors = (flags & NO_ERRORS)==0;
		m_Cols[col].Color = color;
		Invalidate();
	}
}


//
// Change number of columns
//
void CGraph::SetColumnFlags(int col, int mask, int flags)
{
	if (col>=0 && col<m_ColCount) {
		flags = (GetColumnFlags(col) & ~mask) | flags;
		m_Cols[col].Visible = (flags & HIDDEN)==0;
		m_Cols[col].ShowErrors = (flags & NO_ERRORS)==0;
		Invalidate();
	}
}


//
// Get column mask
//
int CGraph::GetColumnFlags(int col) const
{
	if (col>=0 && col<m_ColCount) {
		return (m_Cols[col].Visible ? 0 : HIDDEN) |
			(m_Cols[col].ShowErrors ? 0 : NO_ERRORS);
	}
	return 0;
}

//
// Show/hide column
//
void CGraph::ShowColumn(int col, bool show)
{
	int i;

	if (col>=0 && col<m_ColCount) {
		if (m_Cols[col].Visible!=show) {
			m_Cols[col].Visible = show;
			Invalidate();
		}
	} else if (col<0) {
		for (i=0; i<m_ColCount; i++) 
			m_Cols[i].Visible = show;
		Invalidate();
	}
}

//---------------------------   PAINTING    ----------------------------

// Invalidate the content of the backscreen buffer
void CGraph::Invalidate()
{
	m_NeedUpdate = true;
	if (m_DrawArea)
		gtk_widget_queue_draw(m_DrawArea);
}

// Create a new backing pixmap of the appropriate size 
gint CGraph::configure_event(GtkWidget *widget, GdkEventConfigure *event, CGraph *pMe)
{
	pMe->OnConfigure(widget, event);
	return TRUE;
}

void CGraph::OnConfigure(GtkWidget *widget, GdkEventConfigure *event)
{
	int width, height;

	width = widget->allocation.width;
	height = widget->allocation.height;
	if (width!=m_PixWidth || height!=m_PixHeight) {
		if (m_Pixmap) {
			g_object_unref(m_Pixmap);
			m_Pixmap = NULL;
		}
		if (width>0 && height>0) {
			m_Pixmap = gdk_pixmap_new(widget->window, width, height, -1);
			m_PixWidth = width;
			m_PixHeight = height;
		} else {
			m_PixHeight = m_PixWidth = 0;
		}
		UpdateRectangles();
		Invalidate();
	}
}

// Redraw the screen from the backing pixmap 
gint CGraph::expose_event(GtkWidget *widget, GdkEventExpose *event, CGraph *pMe)
{
	pMe->OnExpose(widget, event);
	return FALSE;
}

void CGraph::OnExpose(GtkWidget *widget, GdkEventExpose *event)
{
	int x, y, r;
	GdkGCValues values;

	if (m_NeedUpdate) {
		m_NeedUpdate = false;
		Redraw();
	}

	gdk_draw_drawable(widget->window, widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		m_Pixmap, event->area.x, event->area.y, event->area.x, event->area.y,
		event->area.width, event->area.height);

	if (m_pData && m_FocusedItem>=0 && m_FocusedCol>=0 && m_FocusedCol<m_ColCount) {
		if (m_FocusedItem<m_pData->Count()) {
			x = RoundToInt(XToView(m_pData->GetX(m_FocusedItem)));
			y = RoundToInt(YToView(m_pData->GetY(m_FocusedItem, m_Cols[m_FocusedCol].ValChannel)));
			r = POINT_SIZE + 2;
			gdk_gc_get_values(widget->style->black_gc, &values);
			gdk_gc_set_function(widget->style->black_gc, GDK_INVERT);
			gdk_draw_line(GDK_DRAWABLE(widget->window), widget->style->black_gc,
				x - r - CURSOR_SIZE, y, x - r - 1, y);
			gdk_draw_line(GDK_DRAWABLE(widget->window), widget->style->black_gc,
				x + r + CURSOR_SIZE, y, x + r + 1, y);
			gdk_draw_line(GDK_DRAWABLE(widget->window), widget->style->black_gc,
				x, y - r - CURSOR_SIZE, x, y - r - 1);
			gdk_draw_line(GDK_DRAWABLE(widget->window), widget->style->black_gc,
				x, y + r + CURSOR_SIZE, x, y + r + 1);
			gdk_gc_set_function(widget->style->black_gc, values.function);
		}
	}
}

// Paint graph
void CGraph::Redraw(void)
{
	if (m_Pixmap && m_PixWidth>0 && m_PixHeight>0) {
		cairo_t *cr = gdk_cairo_create(m_Pixmap);

		// Draw canvas background
		cairo_set_source_rgb(cr, 0.5, 0.5, 0.5);
		cairo_paint(cr);

		// Draw panel
		if (m_PanelRect.right>=m_PanelRect.left && m_PanelRect.bottom>=m_PanelRect.top) {
			cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
			cairo_set_line_width(cr, 1.0);
			cairo_rectangle(cr, m_PanelRect.left, m_PanelRect.top, 
				m_PanelRect.right-m_PanelRect.left, m_PanelRect.bottom-m_PanelRect.top);
			cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
			cairo_fill_preserve(cr);
			cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
			cairo_stroke(cr);
		}

		// Draw grid
		if (m_X.PxlSize>0.0 && m_Y.PxlSize>0.0) {
			if (m_X.ShowGrid)
				PaintXGrid(cr);
			if (m_Y.ShowGrid)
				PaintYGrid(cr);

			// Draw graph data
			if (m_ColCount>0)
				PaintData(cr);

			// Draw scale labels
			if (m_X.ShowLabels)
				PaintXLabels(cr);
			if (m_Y.ShowLabels)
				PaintYLabels(cr);
		}

		cairo_destroy(cr);
	}
}


//
// Compute X scale height
//
int CGraph::GetXScaleHeight(cairo_t *cr)
{
	cairo_text_extents_t extents;

	cairo_text_extents(cr, "X", &extents);
	return (int)ceil(extents.height);
}


//
// Compute Y label height
//
int CGraph::GetYLabelHeight(cairo_t *cr)
{
	cairo_text_extents_t extents;

	cairo_text_extents(cr, "X", &extents);
	return (int)ceil(extents.height);
}


// 
// Compute Y scale width
//
int CGraph::GetYScaleWidth(cairo_t *cr)
{
	cairo_text_extents_t extents;
	char	buf[256];
    double	val;
	
	switch (m_Y.Format) 
	{
    case F_EXP:
        // Exponential mode
		val = Max(fabs(m_Y.Scale*(m_Y.Min+m_Y.Offset)), 
			fabs(m_Y.Scale*(m_Y.Max+m_Y.Offset)));
	    sprintf(buf, "%.*e", m_Y.Prec, -val);
		cairo_text_extents(cr, buf, &extents);
		return (int)ceil(extents.width);

	case F_DEC:
		// Number with fixed number of decimal places
		val = Max(fabs(m_Y.Scale*(m_Y.Min+m_Y.Offset)), 
			fabs(m_Y.Scale*(m_Y.Max+m_Y.Offset)));
	    sprintf(buf, "%.*f", m_Y.Prec, -val);
		cairo_text_extents(cr, buf, &extents);
		return (int)ceil(extents.width);

    case F_TIME:
        // Time mode
		if (m_Y.Prec>=3)
			cairo_text_extents(cr, "00:00", &extents);
		else
			cairo_text_extents(cr, "00:00:00", &extents);
		return (int)ceil(extents.width);
    }
	return 0;
}


// 
// Compute Y scale width
//
int CGraph::GetXLabelWidth(cairo_t *cr, int prec)
{
	cairo_text_extents_t extents;
	char	buf[256];
    double	val;
	
	switch (m_X.Format) 
	{
    case F_EXP:
        // Exponential mode
		val = Max(fabs(m_X.Scale*(m_X.Min+m_X.Offset)), 
			fabs(m_X.Scale*(m_X.Max+m_X.Offset)));
	    sprintf(buf, "%.*e", prec, -val);
		cairo_text_extents(cr, buf, &extents);
		return (int)ceil(extents.width);

	case F_DEC:
		// Number with fixed number of decimal places
		val = Max(fabs(m_X.Scale*(m_X.Min+m_X.Offset)), 
			fabs(m_X.Scale*(m_X.Max+m_X.Offset)));
	    sprintf(buf, "%.*f", prec, -val);
		cairo_text_extents(cr, buf, &extents);
		return (int)ceil(extents.width);

    case F_TIME:
        // Time mode
		if (prec>=3)
			cairo_text_extents(cr, "00:00", &extents);
		else
			cairo_text_extents(cr, "00:00:00", &extents);
		return (int)ceil(extents.width);
    }
	return 0;
}


//
// Print value
//
void CGraph::PrintLabel(char *buf, double value, tFormat fmt, int prec)
{
	int seconds, minutes;

	switch (fmt) 
	{
	case F_EXP:
        // Always exponential form
		sprintf(buf, "%.*e", prec, value);
		if (buf[0]=='-' && buf[strspn(buf, "-0.,e")]=='\0')
			sprintf(buf, "%.*e", prec, 0.0);
		break;

	case F_DEC:
		// Number with fixed number of decimal places
		sprintf(buf, "%.*f", prec, value);
		if (buf[0]=='-' && buf[strspn(buf, "-0.,e")]=='\0')
			sprintf(buf, "%.*f", prec, 0.0);
		break;

	case F_TIME:
        // Time in the form hh:mm
		if (prec>=3) {
			seconds = RoundToInt((value-floor(value)*86400.0));
			sprintf(buf, "%d:%02d:%02d", seconds/3600, (seconds/60)%60, seconds%60);
		} else {
			minutes = RoundToInt((value-floor(value)*14400.0));
			sprintf(buf, "%d:%02d", minutes/60, minutes%60);
		}
		break;
	}
}


//
// Compute text width
//
int CGraph::GetTextWidth(cairo_t *cr, const char *text)
{
	cairo_text_extents_t extents;

	if (text) {
		cairo_text_extents(cr, text, &extents);
		return (int)ceil(extents.width);
	}
	return 0;
}


//
// Compute text height
//
int CGraph::GetTextHeight(cairo_t *cr, const char *text)
{
	cairo_text_extents_t extents;

	if (text) {
		cairo_text_extents(cr, text, &extents);
		return (int)ceil(extents.height);
	}
	return 0;
}

void CGraph::GetTextExtents(cairo_t *cr, const char *text, int *width, int *height)
{
	cairo_text_extents_t extents;

	if (text) {
		cairo_text_extents(cr, text, &extents);
		if (width) 
			*width = (int)ceil(extents.width);
		if (height) 
			*height = (int)ceil(extents.height);
	} else {
		if (width) 
			*width = 0;
		if (height) 
			*height = 0;
	}
}

int CGraph::button_press_event(GtkWidget *widget, GdkEventButton *event, CGraph *pMe)
{
	pMe->OnButtonPress(widget, event);
	return true;
}

void CGraph::OnButtonPress(GtkWidget *widget, GdkEventButton *event)
{
	switch (event->type)
	{
	case GDK_BUTTON_PRESS:
		switch (event->button)
		{
		case 1:
			MouseMove((GdkModifierType)event->state, (int)event->x, (int)event->y);
			Callback(CB_LEFT_CLICK, 0, event);
			break;
		case 3:
			MouseMove((GdkModifierType)event->state, (int)event->x, (int)event->y);
			Callback(CB_RIGHT_CLICK, 0, event);
			break;
		}
		break;

	case GDK_2BUTTON_PRESS:
		switch (event->button)
		{
		case 1:
			MouseMove((GdkModifierType)event->state, (int)event->x, (int)event->y);
			Callback(CB_LEFT_DBLCLICK, 0, event);
			break;
		}
		break;
		
	default:
		break;
	}
}

gint CGraph::motion_notify_event(GtkWidget *widget, GdkEventMotion *event, CGraph *pMe)
{
	pMe->OnMotionNotify(widget, event);
	return true;
}

void CGraph::OnMotionNotify(GtkWidget *widget, GdkEventMotion *event)
{
	gint x, y;
	GdkModifierType state;

	if (m_EnableTrack) {
		if (event->is_hint) {
			gdk_window_get_pointer(event->window, &x, &y, &state);
		} else {
			x = (int)event->x;
			y = (int)event->y;
			state = (GdkModifierType)event->state;
		}
		MouseMove(state, x, y);
	}
}

gint CGraph::leave_notify_event(GtkWidget *widget, GdkEventCrossing *event, CGraph *pMe)
{
	pMe->OnLeaveNotify(widget, event);
	return true;
}

void CGraph::OnLeaveNotify(GtkWidget *widget, GdkEventCrossing *event)
{
	if (m_EnableTrack && (m_MouseX>=0 || m_MouseY>=0)) {
		m_MouseX = m_MouseY = -1;
		Callback(CB_MOUSE_MOVE);
	}
}


//
// Enable/disable mouse tracking
//
void CGraph::EnableTracking(bool enable)
{
	if (enable!=m_EnableTrack) {
		m_EnableTrack = enable;
		Callback(CB_MOUSE_MOVE);
	}
}


//
// Set mouse position
//
void CGraph::MouseMove(GdkModifierType state, int x, int y)
{
	if (m_EnableTrack) {
		if (x>=m_PanelRect.left && x<m_PanelRect.right && 
			y>=m_PanelRect.top && y<m_PanelRect.bottom) {
				if (x!=m_MouseX || y!=m_MouseY) {
					m_MouseX = x;
					m_MouseY = y;
					Callback(CB_MOUSE_MOVE);
				}
		} else {
			if (m_MouseX>=0 || m_MouseY>=0) {
				m_MouseX = m_MouseY = -1;
					Callback(CB_MOUSE_MOVE);
				}
		}
	}
}


//
// Mouse position in physical units
//
bool CGraph::GetMousePos(double *px, double *py) const
{
	double x, y;

	if (m_EnableTrack && m_MouseX>=0 && m_MouseY>=0) {
		x = ViewToProjX(m_MouseX);
		y = ViewToProjY(m_MouseY);
		if (x>=m_X.PMin && x<=m_X.PMax && y>=m_Y.PMin && y<=m_Y.PMax) {
			if (px)
				*px = ProjToX(x);
			if (py)
				*py = ProjToY(y);
			return true;
		}
	}
	return false;
}


//
// Set rectangle coordinates
//
void CGraph::SetRect(tRect *rect, int left, int top, int right, int bottom)
{
	rect->left = left;
	rect->top = top;
	rect->right = right;
	rect->bottom = bottom;
}


//
// Set graph data
//
void CGraph::SetData(CGraphData *pData)
{
	if (pData!=m_pData) {
		CGraphData *pOld = m_pData;
		m_pData = NULL;
		if (pOld)
			pOld->Disconnect(this);
		m_pData = pData;
		m_FocusedItem = m_FocusedCol = -1;
		if (m_pData)
			m_pData->Connect(this);
	}
}

//
// Set graph padding
//
void CGraph::SetGraphPadding(int left, int top, int right, int bottom)
{
	SetRect(&m_GraphPadding, left, top, right, bottom);
	UpdateRectangles();
	Invalidate();
}


//
// Update rectangles
//
void CGraph::UpdateRectangles()
{
	int yscale_width, xscale_height, xname_width, yname_height;
	tRect panel_margins, graph_margins;

	if (!m_Pixmap || m_PixWidth<=0 || m_PixHeight<=0)
		return;

	SetRect(&panel_margins, 0, 0, 0, 0);
	SetRect(&graph_margins, 0, 0, 0, 0);

	// Compute margins
	cairo_t *cr = gdk_cairo_create(m_Pixmap);
	cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
	cairo_set_font_size(cr, FONT_SIZE);
	cairo_select_font_face(cr, "monospace", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
	xscale_height = Max(GetXScaleHeight(cr), GetTextHeight(cr, m_X.Name)) + 2*SCALE_YPADDING;
	xname_width = GetTextWidth(cr, m_X.Name) + 2*NAME_XPADDING;
	if (m_X.ShowLabels) {
		if (!m_X.LabelsInside) {
			if (!m_X.LabelsOpposite) {
				// Outside bottom
				panel_margins.bottom += xscale_height;
			} else {
				// Outside top
				panel_margins.top += xscale_height;
			}
		} else {
			if (!m_X.LabelsOpposite) {
				// Inside bottom
				graph_margins.bottom += xscale_height;
			} else {
				// Inside top
				graph_margins.top += xscale_height;
			}
		}
	}
	yscale_width = Max(GetYScaleWidth(cr), GetTextWidth(cr, m_Y.Name)) + 2*SCALE_XPADDING;
	yname_height = GetTextHeight(cr, m_Y.Name) + 2*NAME_YPADDING;
	if (m_Y.ShowLabels) {
		if (!m_Y.LabelsInside) {
			if (!m_Y.LabelsOpposite) {
				// Outside left
				panel_margins.left += yscale_width;
			} else {
				// Outside right
				panel_margins.right += yscale_width;
			}
		} else {
			if (!m_Y.LabelsOpposite) {
				// Inside left
				graph_margins.left += yscale_width;
			} else {
				// Inside right
				graph_margins.right += yscale_width;
			}
		}
	}

	// Panel rectangle
	m_PanelRect.left = m_PanelPadding.left + panel_margins.left;
	m_PanelRect.top = m_PanelPadding.top + panel_margins.top;
	m_PanelRect.right = m_PixWidth - m_PanelPadding.right - panel_margins.right;
	m_PanelRect.bottom = m_PixHeight - m_PanelPadding.bottom - panel_margins.bottom;

	// Graph rectangle and size
	m_GraphRect.left = m_PanelRect.left + m_GraphPadding.left + graph_margins.left + POINT_SIZE;
	m_GraphRect.top = m_PanelRect.top + m_GraphPadding.top + graph_margins.top + POINT_SIZE;
	m_GraphRect.right = m_PanelRect.right - m_GraphPadding.right - graph_margins.right - POINT_SIZE;
	m_GraphRect.bottom = m_PanelRect.bottom - m_GraphPadding.bottom - graph_margins.bottom - POINT_SIZE;
	m_GraphWidth = m_GraphRect.right - m_GraphRect.left;
	m_GraphHeight = m_GraphRect.bottom - m_GraphRect.top;
	m_OffsetX = (m_GraphRect.left + m_GraphRect.right)/2.0;
	m_OffsetY = (m_GraphRect.top + m_GraphRect.bottom)/2.0;

	// Compute rectangles
	memset(&m_XScaleRect, 0, sizeof(tRect));
	memset(&m_YScaleRect, 0, sizeof(tRect));
	memset(&m_XNameRect, 0, sizeof(tRect));
	memset(&m_YNameRect, 0, sizeof(tRect));
	if (m_X.ShowLabels) {
		m_XScaleRect.left = m_PanelRect.left;
		m_XScaleRect.right = m_PanelRect.right;
		if (!m_X.LabelsInside) {
			if (!m_X.LabelsOpposite) {
				// Outside bottom
				m_XScaleRect.top = m_PanelRect.bottom;
				m_XScaleRect.bottom = m_XScaleRect.top + xscale_height;
			} else {
				// Outside top
				m_XScaleRect.bottom = m_PanelRect.top;
				m_XScaleRect.top = m_XScaleRect.bottom - xscale_height;
			}
		} else {
			if (!m_X.LabelsOpposite) {
				// Inside bottom
				m_XScaleRect.bottom = m_PanelRect.bottom;
				m_XScaleRect.top = m_XScaleRect.bottom - xscale_height;
			} else {
				// Inside top
				m_XScaleRect.top = m_PanelRect.top;
				m_XScaleRect.bottom = m_XScaleRect.top + xscale_height;
			}
		}
		if (m_Y.ShowLabels && m_Y.LabelsInside) {
			if (!m_Y.LabelsOpposite) {
				// Y labels inside left
				m_XScaleRect.left += yscale_width;
			} else {
				// Y labels inside right
				m_XScaleRect.right -= yscale_width;
			}
		}
		if (m_X.Name!=NULL) {
			m_XNameRect = m_XScaleRect;
			if (!m_Y.LabelsOpposite) {
				// Y labels left
				m_XNameRect.left = m_XScaleRect.right - xname_width;
				m_XScaleRect.right = m_XNameRect.left;
			} else {
				// Y labels right
				m_XNameRect.right = m_XScaleRect.left + xname_width;
				m_XScaleRect.left = m_XNameRect.right;
			}
		}
	}
	if (m_Y.ShowLabels) {
		m_YScaleRect.top = m_PanelRect.top;
		m_YScaleRect.bottom = m_PanelRect.bottom;
		if (!m_Y.LabelsInside) {
			if (!m_Y.LabelsOpposite) {
				// Outside left
				m_YScaleRect.right = m_PanelRect.left;
				m_YScaleRect.left = m_YScaleRect.right - yscale_width;
			} else {
				// Outside right
				m_YScaleRect.left = m_PanelRect.right;
				m_YScaleRect.right = m_YScaleRect.left + yscale_width;
			}
		} else {
			if (!m_Y.LabelsOpposite) {
				// Inside left
				m_YScaleRect.left = m_PanelRect.left;
				m_YScaleRect.right = m_YScaleRect.right + yscale_width;
			} else {
				// Inside right
				m_YScaleRect.right = m_PanelRect.right;
				m_YScaleRect.left = m_YScaleRect.right - yscale_width;
			}
		}
		if (m_X.ShowLabels && m_X.LabelsInside) {
			if (!m_X.LabelsOpposite) {
				// X labels inside bottom
				m_YScaleRect.bottom -= xscale_height;
			} else {
				// Y labels inside top
				m_YScaleRect.top += xscale_height;
			}
		}
		if (m_Y.Name!=NULL) {
			m_YNameRect = m_YScaleRect;
			if (!m_X.LabelsOpposite) {
				// X labels bottom
				m_YNameRect.bottom = m_YScaleRect.top + yname_height;
				m_YScaleRect.top = m_YNameRect.bottom;
			} else {
				// X labels top
				m_YNameRect.top = m_YScaleRect.bottom - yname_height;
				m_YScaleRect.bottom = m_YNameRect.top;
			}
		}
	}
	cairo_destroy(cr);

	UpdatePxlSizeX();
	LimitPosX();
	UpdatePxlSizeY();
	LimitPosY();
}

void CGraph::PrintXLabel(char *buf, double value, int minprec)
{
	if (m_X.Format!=F_DEC || (minprec<m_X.Prec))
		PrintLabel(buf, value, m_X.Format, m_X.Prec);
	else
		PrintLabel(buf, value, m_X.Format, minprec);
}

void CGraph::PrintYLabel(char *buf, double value, int minprec)
{
	if (m_Y.Format!=F_DEC || (minprec<m_Y.Prec))
		PrintLabel(buf, value, m_Y.Format, m_Y.Prec);
	else
		PrintLabel(buf, value, m_Y.Format, minprec);
}


//
// Compute horizontal grid step and distance between two labels
// on horizontal scale
//
void CGraph::GetGridXStep(cairo_t *cr, double minval, double maxval, double *step, int *minprec)
{
    int		i, cx, xprec, oldprec;
	double	xstep;
	
	switch (m_X.Format)
	{
	case F_TIME:
        // Time mode
		cx = GetXLabelWidth(cr, 0);
		for (i=NumTSteps-1; i>=0; i--) {
			if (TSteps[i]/m_X.PxlSize < (cx+24))
				break;
		}
		*step = TSteps[i+1];
		*minprec = 0;
		break;

	case F_DEC:
	case F_EXP:
        // Decimal and exponential mode
		xprec = m_X.Prec;
		do {
			oldprec = xprec;
			cx = GetXLabelWidth(cr, xprec);
			for (i=NumSteps-1; i>=0; i--) {
				if (Steps[i]/m_X.PxlSize < (cx+24))
					break;
			}
			xstep = Steps[i+1];
			xprec = MinPrecs[i+1];
		} while (xprec > oldprec);
		*step = xstep;
		*minprec = xprec;
		break;
	}
}


//
// Compute vertical grid step and distance between two labels
// on vertical scale
//
void CGraph::GetGridYStep(cairo_t *cr, double minval, double maxval, double *step, int *minprec)
{
    int		i, cy;

	cy = GetYLabelHeight(cr);
	switch (m_Y.Format)
	{
	case F_TIME:
        // Time mode
		for (i=NumTSteps-1; i>=0; i--) {
			if (TSteps[i]/m_Y.PxlSize < (cy+48))
				break;
		}
		*step = TSteps[i+1];
		*minprec = 0;
		break;

	case F_DEC:
	case F_EXP:
        // Decimal mode
		for (i=NumSteps-1; i>=0; i--) {
			if (Steps[i]/m_Y.PxlSize < (cy+48))
				break;
		}
		*step = Steps[i+1];
		*minprec = MinPrecs[i+1];
		break;
	}
}

//
// Paint scale label
//
void CGraph::PaintXScaleLabel(cairo_t *cr, double val, int prec, int y)
{
	int x, w;
	char buf[256];

	x = RoundToInt(XToView(val/m_X.Scale - m_X.Offset));
	if (x>=m_XScaleRect.left && x<m_XScaleRect.right) {
		PrintXLabel(buf, val, prec);
		w = GetTextWidth(cr, buf);
		x = x - w/2;
		if (x>=m_XScaleRect.left && x+w<m_XScaleRect.right) {
			cairo_move_to(cr, x, y);
			cairo_show_text(cr, buf);
		}
	}
}


//
// Paint labels on x-axis scale
//
void CGraph::PaintXLabels(cairo_t *cr)
{
	double val, step, xmin, xmax;
	int y, minprec;

    if (!m_X.ShowLabels)
		return;

    // Visible region in physical units
	if (!m_X.Reverse) {
		xmin = LimitValue(ViewToX(m_PanelRect.left), m_X.Min, m_X.Max); 
		xmax = LimitValue(ViewToX(m_PanelRect.right), m_X.Min, m_X.Max);
	} else {
		xmin = LimitValue(ViewToX(m_PanelRect.right), m_X.Min, m_X.Max); 
		xmax = LimitValue(ViewToX(m_PanelRect.left), m_X.Min, m_X.Max);
	}
	xmin = m_X.Scale*(xmin+m_X.Offset);
	xmax = m_X.Scale*(xmax+m_X.Offset);

	y =  m_XScaleRect.bottom - SCALE_YPADDING;

	cairo_save(cr);
	cairo_reset_clip(cr);
	cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
	cairo_set_font_size(cr, FONT_SIZE);
	cairo_select_font_face(cr, "monospace", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);

	// Print unit name
	if (m_X.Name) {
		//cairo_reset_clip(cr);
		//cairo_rectangle(cr, m_XNameRect.left, m_XNameRect.top, m_XNameRect.right, m_XNameRect.bottom);
		//cairo_clip(cr);
		cairo_set_source_rgb(cr, 0, 0, 0);
		cairo_move_to(cr, m_XNameRect.left + NAME_XPADDING, y);
		cairo_show_text(cr, m_X.Name);
	}

	//cairo_reset_clip(cr);
	//cairo_rectangle(cr, m_XScaleRect.left, m_XScaleRect.top, m_XScaleRect.right, m_XScaleRect.bottom);
	//cairo_clip(cr);
	cairo_set_source_rgb(cr, 0, 0, 0);
	if (m_X.Log) {
		if (m_X.Log && log10(xmax/xmin)>4) {
			// Log scale if range is greater than three periods
			xmin = log10(Max(DBL_MIN, xmin));
			xmax = log10(Max(DBL_MIN, xmax));
			GetGridXStep(cr, xmin, xmax, &step, &minprec);
			if (step<m_X.Scale*m_X.PxlSize) 
				step = m_X.Scale*m_X.PxlSize;
			xmin = ceil(xmin);
			while (xmin <= xmax) {
				PaintXScaleLabel(cr, pow(10, xmin), minprec, y);
				xmin += step;
			}
		} else
		if (m_X.Log && log10(xmax/xmin)>1) {
			// Log scale if range is greater than one period
			xmin = log10(Max(DBL_MIN, xmin));
			xmax = log10(Max(DBL_MIN, xmax));
			step = 1.0; minprec = 1;
			xmin = floor(xmin);
			while (xmin <= xmax) {
				val = pow(10, xmin);
				PaintXScaleLabel(cr, 1.0*val, minprec, y);
				PaintXScaleLabel(cr, 2.0*val, minprec, y);
				PaintXScaleLabel(cr, 5.0*val, minprec, y);
				xmin += step;
			}
		} else {
			// Lin scale or log scale if range is smaller than one period
			GetGridXStep(cr, xmin, xmax, &step, &minprec);
			if (step<m_X.Scale*m_X.PxlSize) 
				step = m_X.Scale*m_X.PxlSize;
			xmin = ceil(xmin/step)*step;
			while (xmin <= xmax) {
				PaintXScaleLabel(cr, xmin, minprec, y);
				xmin += step;
			}
		}
	} else {
		// Linear scale
		GetGridXStep(cr, xmin, xmax, &step, &minprec);
		if (step<m_X.Scale*m_X.PxlSize) 
			step = m_X.Scale*m_X.PxlSize;
		xmin = ceil(xmin/step)*step;
		while (xmin <= xmax) {
			PaintXScaleLabel(cr, xmin, minprec, y);
			xmin += step;
		}
	}
	cairo_restore(cr);
}


//
// Paint Y axis label
//
void CGraph::PaintYScaleLabel(cairo_t *cr, double val, int prec, int x)
{
	int y, w, h;
	char buf[256];

	y = RoundToInt(YToView(val/m_Y.Scale - m_Y.Offset));
	if (y>=m_YScaleRect.top && y<m_YScaleRect.bottom) {
		PrintYLabel(buf, val, prec);
		GetTextExtents(cr, buf, &w, &h);
		x = x - w;
		y = y + h/2;
		if (y-h>=m_YScaleRect.top && y<m_YScaleRect.bottom) {
			cairo_move_to(cr, x, y);
			cairo_show_text(cr, buf);
		}
	}
}


//
// Paint labels on y-axis scale
//
void CGraph::PaintYLabels(cairo_t *cr)
{
	double val, step, ymin, ymax;
	int x, minprec;

    if (!m_Y.ShowLabels)
        return;

    // Visible region in physical units
	if (!m_Y.Reverse) {
		ymin = LimitValue(ViewToY(m_PanelRect.bottom), m_Y.Min, m_Y.Max);
		ymax = LimitValue(ViewToY(m_PanelRect.top), m_Y.Min, m_Y.Max);
	} else {
		ymin = LimitValue(ViewToY(m_PanelRect.top), m_Y.Min, m_Y.Max);
		ymax = LimitValue(ViewToY(m_PanelRect.bottom), m_Y.Min, m_Y.Max);
	}
	ymin = m_Y.Scale*(ymin+m_Y.Offset);
	ymax = m_Y.Scale*(ymax+m_Y.Offset);

    // Compute the height of the characters
	cairo_save(cr);
	cairo_reset_clip(cr);
	cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
	cairo_set_font_size(cr, FONT_SIZE);
	cairo_select_font_face(cr, "monospace", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);

	// Print unit name
	if (m_Y.Name) {
		//cairo_reset_clip(cr);
		//cairo_rectangle(cr, m_YNameRect.left, m_YNameRect.top, m_YNameRect.right, m_YNameRect.bottom);
		//cairo_clip(cr);
		cairo_set_source_rgb(cr, 0, 0, 0);
		x = (m_YNameRect.left + m_YNameRect.right - GetTextWidth(cr, m_Y.Name))/2;
		cairo_move_to(cr, x, m_YNameRect.bottom - NAME_YPADDING);
		cairo_show_text(cr, m_Y.Name);
	}

	//cairo_reset_clip(cr);
	//cairo_rectangle(cr, m_YScaleRect.left, m_YScaleRect.top, m_YScaleRect.right, m_YScaleRect.bottom);
	//cairo_clip(cr);
	cairo_set_source_rgb(cr, 0, 0, 0);
	x = m_YScaleRect.right - SCALE_XPADDING;
	if (m_Y.Log && log10(ymax/ymin)>4) {
		// Log scale if range is greater than three periods
		ymin = log10(Max(DBL_MIN, ymin));
		ymax = log10(Max(DBL_MIN, ymax));
		GetGridYStep(cr, ymin, ymax, &step, &minprec);
		//if (step<YPxlsize) step = YPxlsize;
		ymin = ceil(ymin);
		while (ymin <= ymax) {
			PaintYScaleLabel(cr, pow(10,ymin), minprec, x);
			ymin += step;
		}
	} else 
	if (m_Y.Log && log10(ymax/ymin)>1) {
		// Log scale if range is greater than one period
		ymin = log10(Max(DBL_MIN, ymin));
		ymax = log10(Max(DBL_MIN, ymax));
		step = 1.0; minprec = 1;
		ymin = floor(ymin);
		while (ymin <= ymax) {
			val = pow(10,ymin);
			PaintYScaleLabel(cr, 1.0*val, minprec, x);
			PaintYScaleLabel(cr, 2.0*val, minprec, x);
			PaintYScaleLabel(cr, 5.0*val, minprec, x);
			ymin += step;
		}
	} else {
		// Lin scale or log scale if range is smaller than one period
		GetGridYStep(cr, ymin, ymax, &step, &minprec);
		//if (step<YPxlsize) step = YPxlsize;
		ymin = ceil(ymin/step)*step;
		while (ymin <= ymax) {
			PaintYScaleLabel(cr, ymin, minprec, x);
			ymin += step;
		}
	}
	cairo_restore(cr);
}


//
// Paint X axis grid line
//
void CGraph::PaintXGridLine(cairo_t *cr, double val)
{
	int x = RoundToInt(XToView(val/m_X.Scale - m_X.Offset));
	if (x>=m_PanelRect.left && x<m_PanelRect.right) {
		cairo_move_to(cr, x, m_PanelRect.top+1);
		cairo_line_to(cr, x, m_PanelRect.bottom-1);
	}
}


//
// Paints x-axis grid
//
void CGraph::PaintXGrid(cairo_t *cr)
{
	const double dashsize = 3;
	double val, step, xmin, xmax;
	int minprec;

    // Visible region in scale units
	if (!m_X.Reverse) {
		xmin = LimitValue(ViewToX(m_PanelRect.left), m_X.Min, m_X.Max); 
		xmax = LimitValue(ViewToX(m_PanelRect.right), m_X.Min, m_X.Max);
	} else {
		xmin = LimitValue(ViewToX(m_PanelRect.right), m_X.Min, m_X.Max); 
		xmax = LimitValue(ViewToX(m_PanelRect.left), m_X.Min, m_X.Max);
	}
	xmin = m_X.Scale*(xmin+m_X.Offset);
	xmax = m_X.Scale*(xmax+m_X.Offset);

	cairo_save(cr);
	cairo_reset_clip(cr);
	cairo_set_source_rgb(cr, 0.7, 0.7, 0.7);
	cairo_set_dash(cr, &dashsize, 1, 0);
	if (m_X.Log && log10(xmax/xmin)>4) {
		// Log scale if range is greater than three periods
		xmin = log10(Max(DBL_MIN, xmin));
		xmax = log10(Max(DBL_MIN, xmax));
		GetGridXStep(cr, xmin, xmax, &step, &minprec);
		if (step<m_X.Scale*m_X.PxlSize) 
			step = m_X.Scale*m_X.PxlSize;
		xmin = ceil(xmin);
		while (xmin <= xmax) {
			PaintXGridLine(cr, pow(10, xmin));
			xmin += step;
		}
	} else
	if (m_X.Log && log10(xmax/xmin)>1) {
		// Log scale if range is greater than one period
		xmin = log10(Max(DBL_MIN, xmin));
		xmax = log10(Max(DBL_MIN, xmax));
		step = 1.0; minprec = 1;
		xmin = floor(xmin);
		while (xmin <= xmax) {
			val = pow(10, xmin);
			PaintXGridLine(cr, 1.0*val);
			PaintXGridLine(cr, 2.0*val);
			PaintXGridLine(cr, 5.0*val);
			xmin += step;
		}
	} else {
		// Lin scale or log scale if range is smaller than one period
		GetGridXStep(cr, xmin, xmax, &step, &minprec);
		if (step<m_X.Scale*m_X.PxlSize) 
			step = m_X.Scale*m_X.PxlSize;
		xmin = ceil(xmin/step)*step;
		while (xmin <= xmax) {
			PaintXGridLine(cr, xmin);
			xmin += step;
		}
	}
	cairo_stroke(cr);
	cairo_restore(cr);
}


//
// Paint Y axis grid line
//
void CGraph::PaintYGridLine(cairo_t *cr, double val)
{
	int y = RoundToInt(YToView(val/m_Y.Scale - m_Y.Offset));
	if (y>=m_PanelRect.top && y<m_PanelRect.bottom) {
		cairo_move_to(cr, m_PanelRect.left+1, y);
		cairo_line_to(cr, m_PanelRect.right-1, y);
	}
}


//
// Paint y-axis grid
//
void CGraph::PaintYGrid(cairo_t *cr)
{
	const double dashsize = 3;
	double val, step, ymin, ymax;
	int minprec;

    // Visible region in physical units
	if (!m_Y.Reverse) {
		ymin = LimitValue(ViewToY(m_PanelRect.bottom), m_Y.Min, m_Y.Max);
		ymax = LimitValue(ViewToY(m_PanelRect.top), m_Y.Min, m_Y.Max);
	} else {
		ymin = LimitValue(ViewToY(m_PanelRect.top), m_Y.Min, m_Y.Max);
		ymax = LimitValue(ViewToY(m_PanelRect.bottom), m_Y.Min, m_Y.Max);
	}
	ymin = m_Y.Scale*(ymin+m_Y.Offset);
	ymax = m_Y.Scale*(ymax+m_Y.Offset);

	cairo_save(cr);
	cairo_reset_clip(cr);
	cairo_set_source_rgb(cr, 0.7, 0.7, 0.7);
	cairo_set_dash(cr, &dashsize, 1, 0);
	if (m_Y.Log && log10(ymax/ymin)>4) {
		// Log scale if range is greater than three periods
		ymin = log10(Max(DBL_MIN, ymin));
		ymax = log10(Max(DBL_MIN, ymax));
		GetGridYStep(cr, ymin, ymax, &step, &minprec);
		//if (step<YPxlsize) step = YPxlsize;
		ymin = ceil(ymin);
		while (ymin <= ymax) {
			PaintYGridLine(cr, pow(10, ymin));
			ymin += step;
		}
	} else 
	if (m_Y.Log && log10(ymax/ymin)>1) {
		// Log scale if range is greater than one period
		ymin = log10(Max(DBL_MIN, ymin));
		ymax = log10(Max(DBL_MIN, ymax));
		step = 1.0; minprec = 1;
		ymin = floor(ymin);
		while (ymin <= ymax) {
			val = pow(10, ymin);
			PaintYGridLine(cr, 1.0*val);
			PaintYGridLine(cr, 2.0*val);
			PaintYGridLine(cr, 5.0*val);
			ymin += step;
		}
	} else {
		// Lin scale or log scale if range is smaller than one period
		GetGridYStep(cr, ymin, ymax, &step, &minprec);
		//if (step<YPxlsize) step = YPxlsize;
		ymin = ceil(ymin/step)*step;
		while (ymin <= ymax) {
			PaintYGridLine(cr, ymin);
			ymin += step;
		}
	}
	cairo_stroke(cr);
	cairo_restore(cr);
}


//
// Paint curves
//
void CGraph::PaintData(cairo_t *cr)
{
	int i, j, x, y, y1, y2, samples;
    double xval, yval, yerr;
	const double *yvals;
	tColor color;
    
	cairo_save(cr);

	cairo_reset_clip(cr);
	cairo_rectangle(cr, m_PanelRect.left+1, m_PanelRect.top+1, m_PanelRect.right-1, m_PanelRect.bottom-1);
	cairo_clip(cr);

	// Paint data
	cairo_set_source_rgb(cr, 0, 0, 0);
	cairo_set_line_width(cr, 1.0);
	cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
	samples = m_pData->Count();
	for (i=0; i<samples; i++) {
		xval = m_pData->GetX(i);
		if (xval>=m_X.Min && xval<=m_X.Max) {
			x = RoundToInt(XToView(xval));
			if (x>=m_PanelRect.left && x<m_PanelRect.right) {
				yvals = m_pData->GetValues(i);
				for (j=0; j<m_ColCount; j++) {
					if (m_Cols[j].Visible && m_Cols[j].ValChannel>=0) {
						yval = yvals[m_Cols[j].ValChannel];
						if (yval>=m_Y.Min && yval<=m_Y.Max) {
							y = RoundToInt(YToView(yval));
							if (y>=m_PanelRect.top && y<m_PanelRect.bottom) {
								color = m_Cols[j].Color;
								cairo_set_source_rgb(cr, GetR(color)/255.0, GetG(color)/255.0, GetB(color)/255.0);
								cairo_arc(cr, x, y, 3.0, 0.0, 2.0*M_PI);
								cairo_fill(cr);
								if (m_Cols[j].ErrChannel>=0 && m_Cols[j].ShowErrors) {
									yerr = yvals[m_Cols[j].ErrChannel];
									if (yerr>0) {
										y1 = RoundToInt(YToView(yval - yerr));
										y2 = RoundToInt(YToView(yval + yerr));
										cairo_move_to(cr, x-POINT_SIZE, y1);
										cairo_line_to(cr, x+POINT_SIZE, y1);
										cairo_move_to(cr, x-POINT_SIZE, y2);
										cairo_line_to(cr, x+POINT_SIZE, y2);
										cairo_move_to(cr, x, y1);
										cairo_line_to(cr, x, y2);
										cairo_stroke(cr);
									}
								}
							}
						}
					}
				}
			}
		}
	}

	// Paint tags
	cairo_set_source_rgb(cr, 0, 0, 0);
	cairo_set_font_size(cr, 12.0);
	cairo_select_font_face(cr, "", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD);
	cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
	for (GSList *ptr=m_Tags; ptr!=NULL; ptr=ptr->next) {
		tTag *tag = (tTag*)ptr->data;
		if (tag->Row>=0 && tag->Row<m_pData->Count() && tag->Col>=0 && tag->Col<m_ColCount) {
			if (m_Cols[tag->Col].Visible && m_Cols[tag->Col].ValChannel>=0) {
				xval = m_pData->GetX(tag->Row);
				if (xval>=m_X.Min && xval<=m_X.Max) {
					x = RoundToInt(XToView(xval));
					if (x>=m_PanelRect.left && x<m_PanelRect.right) {
						yval = m_pData->GetY(tag->Row, m_Cols[tag->Col].ValChannel);
						if (yval>=m_Y.Min && yval<=m_Y.Max) {
							y = RoundToInt(YToView(yval));
							if (y>=m_PanelRect.top && y<m_PanelRect.bottom) {
								color = tag->Color;
								cairo_set_source_rgb(cr, GetR(color)/255.0, GetG(color)/255.0, GetB(color)/255.0);
								cairo_arc(cr, x, y, 3.0, 0.0, 2.0*M_PI);
								cairo_fill(cr);
								if (tag->Text) {
									cairo_move_to(cr, x+4, y-4);
									cairo_show_text(cr, tag->Text);
								}
							}
						}
					}
				}
			}
		}
	}
	cairo_restore(cr);
}


//
// Find a sample near to the specified position
//
int CGraph::FindByPos(double x, double y, int column) const
{
	int i, count, index;
	double r2min, u, v, u0, v0, r2;

	if (!m_pData || column<0 || column>=m_ColCount)
		return -1;

	u0 = XToView(x);
	v0 = YToView(y);

	count = m_pData->Count();
	r2min = 400;
	index = -1;
	for (i=0; i<count; i++) {
		u = XToView(m_pData->GetX(i));
		v = YToView(m_pData->GetY(i, m_Cols[column].ValChannel));
		r2 = (u-u0)*(u-u0) + (v-v0)*(v-v0);
		if (r2<=r2min) {
			index = i;
			r2min = r2;
		}
	}
	return index;
}


//
// Set focused item
//
void CGraph::SetFocusedItem(int index, int column)
{
	if (m_FocusedItem!=index || m_FocusedCol!=column) {
		m_FocusedItem = index;
		m_FocusedCol = column;
		Invalidate();
	}
}


//
// Autozoom
//
void CGraph::AutoZoom(bool x_axis, bool y_axis)
{
	if (x_axis)
		AutoXZoom();
	if (y_axis)
		AutoYZoom();
}


//
// Set tags
//
void CGraph::SetTag(int column, int row, const char *text, tColor color)
{
	bool ok = false;
	GSList *ptr;

	if (row>=0 && row<m_pData->Count() && column>=0 && column<m_ColCount) {
		for (ptr=m_Tags; ptr!=NULL; ptr=ptr->next) {
			if (((tTag*)ptr->data)->Row==row && ((tTag*)ptr->data)->Col==column) {
				strncpy(((tTag*)ptr->data)->Text, text, 63);
				((tTag*)ptr->data)->Color = color;
				ok = true;
				break;
			}
		}
		if (!ok) {
			tTag *tag = (tTag*)g_malloc(sizeof(tTag));
			tag->Row = row;
			tag->Col = column;
			strncpy(tag->Text, text, 63);
			tag->Color = color;
			m_Tags = g_slist_append(m_Tags, tag);
		}
		Invalidate();
	}
}


//
// Delete tag
//
void CGraph::DeleteTag(int column, int row)
{
	GSList *ptr;

	for (ptr=m_Tags; ptr!=NULL; ptr=ptr->next) {
		if (((tTag*)ptr->data)->Row==row && ((tTag*)ptr->data)->Col==column) {
			g_free(ptr->data);
			m_Tags = g_slist_remove(m_Tags, ptr);
			Invalidate();
			break;
		}
	}
}


//
// Delete all tags
//
void CGraph::DeleteAllTags(void)
{
	GSList *ptr;

	if (m_Tags) {
		for (ptr=m_Tags; ptr!=NULL; ptr=ptr->next) 
			g_free(ptr->data);
		g_slist_free(m_Tags);
		m_Tags = NULL;
		Invalidate();
	}
}
