/**************************************************************

selection.h (C-Munipack project)
Table of selected stars
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#ifndef MUNIWIN_HELPER_CLASSES_H
#define MUNIWIN_HELPER_CLASSES_H

#include <gtk/gtk.h>
#include <cmunipack.h>

#include "callback.h"

// Types of stars:
enum tStarType
{
	ST_NOT_SELECTED,
	ST_VARIABLE,
	ST_COMPARISON,
	ST_CHECK
};

#define SELECTION_TYPES 4

//
// Selection of stars
//
class CSelection:public CCBObject
{
public:
	enum tMessageId
	{
		CB_SELECT_CHANGE		// Selection changed
	};

public:
	// Constructor(s)
	CSelection();
	CSelection(const CSelection &orig);
	
	// Destructor
	virtual ~CSelection();

	// assignment operator
	const CSelection &operator=(const CSelection &orig);

	// Clear data
	void Clear();

	// Load selection from a file
	void Load(CmpackCatFile *file);

	// Save selection to a file
	void Save(CmpackCatFile *file) const;

	// Select a star (automatically
	void Select(int id, tStarType type);

	// Unselect a star
	void Unselect(int id);

	// Find record that matches specified type and index
	// Returns the index of the record
	int FindStar(tStarType type, int index) const;

	// Find record that matches specified star identifier
	// Returns the index of the record
	int FindStar(int id) const;

	// Get number of selected stars
	int Count(void) const
	{ return m_Count; }

	// Get number of stars of specified type
	int CountStars(tStarType type) const;

	// Enumerate selected stars
	bool Get(int i, int *id, tStarType *type, int *index) const;
	int GetId(int i) const;
	tStarType GetType(int i) const;
	int GetIndex(int i) const;

	// Get identifiers of selected star
	int GetStarList(tStarType type, int *list, int maxcount) const;
	char *GetStarList(tStarType type) const;

	// Set list of selected stars
	void SetStarList(tStarType type, const int *list, int count);
	void SetStarList(tStarType type, const char *string);

protected:
	struct tStarRec
	{
		int id;
		tStarType type;
		int index;
	};

	struct tReindexInfo
	{
		int			var_index;
		int			comp_index;
		int			chk_index;
	};

	int m_Count, m_Capacity;
	tStarRec *m_List;

	// Search star by its id
	int Search(int id) const;

	// Refresh selection indices
	void Reindex(void);

	// Load selection
	void LoadSelection(CmpackCatSelection *sel, tStarType type);

	// Save selection
	void SaveSelection(CmpackCatSelection *sel, tStarType type) const;
};


//
// File information
//
class CFileInfo
{
public:
	// Constructors
	CFileInfo();
	CFileInfo(const gchar *filepath);
	CFileInfo(const CFileInfo &orig);

	// Destructor
	virtual ~CFileInfo();

	// Assignment operator
	const CFileInfo &operator=(const CFileInfo &orig);

	// Clear data
	void Clear(void);

	// Is the data valid?
	bool Valid() const
	{ return m_FullPath!=NULL; }

	// File path
	const gchar *FullPath() const
	{ return m_FullPath; }

	// File name
	const gchar *Name() const
	{ return (m_FullPath ? g_basename(m_FullPath) : NULL); }

	// Load from project file
	void Load(GKeyFile *cfg, const gchar *section, const gchar *prefix);

	// Save to project file
	void Save(GKeyFile *cfg, const gchar *section, const gchar *prefix);

	// Comparison operators
	friend bool operator==(const CFileInfo &a, const CFileInfo &b);
	friend bool operator!=(const CFileInfo &a, const CFileInfo &b);

private:
	gchar		*m_FullPath;
	time_t		m_MTime;
};

//
// CCD frame info
//
class CImageInfo
{
public:
	// Constructor(s)
	CImageInfo();
	CImageInfo(const CImageInfo &orig);
	
	// Destructor
	virtual ~CImageInfo();

	// Copy data from a file
	void Set(CmpackCcdFile *file);

	// assignment operator
	const CImageInfo &operator=(const CImageInfo &orig);

	// Clear data
	void Clear();

	// CCD image size
	inline int Width(void) const
	{ return m_Width; }
	inline int Height(void) const
	{ return m_Height; }

	// File format identifier (magic string)
	inline const char *Magic(void) const
	{ return m_Magic; }

	// Filter name
	inline const char *Filter(void) const
	{ return m_Filter; }

	// Object's name
	inline const char *Object(void) const
	{ return m_Object; }

	// Object's coordinates
	bool ObjectCoordinates(double *ra, double *dec) const;

	// Observer's name
	inline const char *Observer(void) const
	{ return m_Observer; }

	// Obsever's coordinates
	bool ObserverCoordinates(double *lon, double *lat) const;

	// Date of observation
	inline double JD(void) const
	{ return m_JD; }

	// Exposure duration
	inline double ExpTime(void) const
	{ return m_ExpTime; }

	// CCD temperature
	inline double CCDTemp(void) const
	{ return m_CCDTemp; }

	// Number of accumulated frames
	inline int Frames(void) const
	{ return m_Frames; }

protected:
	int		m_Width, m_Height;
	char	*m_Magic, *m_Filter, *m_Object, *m_Observer;
	double	m_JD, m_ExpTime, m_CCDTemp;
	bool	m_ObjCoordsValid, m_ObsCoordsValid;
	double	m_ObjRA, m_ObjDec, m_ObsLon, m_ObsLat;
	int		m_Frames;
};


//
// Photometry file info
//
class CChartInfo
{
public:
	// Constructor(s)
	CChartInfo();
	CChartInfo(const CChartInfo &orig);
	
	// Destructor
	virtual ~CChartInfo();

	// Copy data from a file
	void Set(CmpackPhtFile *file);
	void Set(CmpackCatFile *file);

	// assignment operator
	const CChartInfo &operator=(const CChartInfo &orig);

	// Clear data
	void Clear();

	// Chart size
	inline int Width(void) const
	{ return m_Width; }
	inline int Height(void) const
	{ return m_Height; }

	// Filter name
	inline const char *Filter(void) const
	{ return m_Filter; }

	// Date of observation
	inline double JD(void) const
	{ return m_JD; }

	// Exposure duration
	inline double ExpTime(void) const
	{ return m_ExpTime; }

	// Number of stars
	inline int Stars(void) const
	{ return m_Stars; }

protected:
	int		m_Width, m_Height, m_Stars;
	char	*m_Filter;
	double	m_JD, m_ExpTime;
};


//
// Table info
//
class CTableInfo
{
public:
	// Constructor
	CTableInfo();
	CTableInfo(const CTableInfo &orig);

	// Destructor
	virtual ~CTableInfo();

	// Clear data
	void Clear();

	// Copy data from context
	void Set(CmpackTable *tab);

	// assignment operator
	const CTableInfo &operator=(const CTableInfo &orig);

	// Get filter name
	const char *Filter(void) const
	{ return m_Filter; }

	// Get aperture index
	int Aperture(void) const
	{ return m_Aperture; }

protected:
	char		*m_Filter;
	int			m_Aperture;
};


//
// Read-all file info
//
class CReadAllInfo
{
public:
	// Constructor
	CReadAllInfo();
	CReadAllInfo(const CReadAllInfo &orig);

	// Destructor
	virtual ~CReadAllInfo();

	// Clear data
	void Clear();

	// Copy data from context
	void Set(CmpackAllFile *tab);

	// assignment operator
	const CReadAllInfo &operator=(const CReadAllInfo &orig);

	// Get filter name
	const char *Filter(void) const
	{ return m_Filter; }

	// Get aperture index
	int Aperture(void) const
	{ return m_Aperture; }

	// Get heliocentric correction flag
	bool HelCor(void) const
	{ return m_HelCor; }

protected:
	char		*m_Filter;
	int			m_Aperture;
	bool		m_HelCor;
};


//
// Object coordinates
//
class CObjectCoords:public CCBObject
{
public:
	// Constructor(s)
	CObjectCoords();
	CObjectCoords(const CObjectCoords &orig);
	
	// Destructor
	virtual ~CObjectCoords();

	// assignment operator
	const CObjectCoords &operator=(const CObjectCoords &orig);

	// Load selection from a CCD file
	void Load(CmpackCcdFile *file);

	// Load selection from a catalogue file
	void Load(CmpackCatFile *file);

	// Set object's designation
	void SetName(const char *name);

	// Set object's coordinates
	void SetRA(const char *objra);
	void SetDec(const char *objdec);

	// Clear data
	void Clear();

	// Object's designation
	const char *Name(void) const
	{ return m_Name; }

	// Object's coordinates
	const char *RA(void) const
	{ return m_RA; }
	const char *Dec(void) const
	{ return m_Dec; }

protected:
	char	*m_Name;
	char	*m_RA, *m_Dec;
};

//
// Observer coordinates
//
class CObserverCoords:public CCBObject
{
public:
	// Constructor(s)
	CObserverCoords();
	CObserverCoords(const CObserverCoords &orig);
	
	// Destructor
	virtual ~CObserverCoords();

	// assignment operator
	const CObserverCoords &operator=(const CObserverCoords &orig);

	// Load selection from a CCD file
	void Load(CmpackCcdFile *file);

	// Set location name
	void SetName(const char *name);

	// Set observer's coordinates
	void SetLon(const char *lon);
	void SetLat(const char *lat);

	// Clear data
	void Clear();

	// Location name
	const char *Name(void) const
	{ return m_Name; }

	// Observer's coordinates
	const char *Lon(void) const
	{ return m_Lon; }
	const char *Lat(void) const
	{ return m_Lat; }

protected:
	char	*m_Name;
	char	*m_Lon, *m_Lat;
};

#endif
